kfold <- function(targets, k = 10, equal = F, repetitions = 1){
  targets <- as.character(targets)
  table <- table(targets)
  # INPUT: targets as vector
  # OUTPUT: list of indexnumbers for folds (in dataset)
  # para. targets: input
  # para. k: number of folds
  # para. equal: class distribution is 50/50
  # repetition: how much repetitions (Output = rep*kx)
  knums <- list()
  rep <- list()

  for(i in 1:repetitions){
    c <- list()
    # get equal samplesizes
    for(name in names(table)) {
      if(equal == T){
        c[[name]] <- split(which(targets == name)[sample(1:min(table))], f = c(1:k))
      } else {
        c[[name]] <- split(which(targets == name)[sample(1:table[name])], f = c(1:k))
      }
    }

  for(j in 1:k){
      knums[[j]] <- as.numeric(unlist(sapply(names(table), function(name) c[[name]][[j]])))
  }

  rep[[i]] <- knums
  }

  # returns the 1 fold indexes for testing (1/k)
  return(rep)
}


preprocess <- function(texts,
                       decapitalize = T,
                       del_stopwords = T,
                       list_stopwords = c(''),
                       del_numbers = T,
                       del_symbols = T,
                       stemming = T,
                       lang_stemming = 'english',
                       umlaut_normalization = T
){
  if(all(list_stopwords != ''))
    del_stopwords <- F

  # decapitalize
  if(decapitalize){
    texts <- tolower(texts)
  }
  # del numbers/ symbols
  del_numbers_RegEx <- ifelse(del_numbers, '[:digit:]', '')
  del_symbols_RegEx <- ifelse(del_symbols, '[:punct:]', '')
  texts <- gsub(paste0("[[:cntrl:]", del_numbers_RegEx, del_symbols_RegEx, "]"), " ", texts)
  texts <- trimws(gsub("[[:blank:]]+", " ", texts))


  # del stop words
  if(del_stopwords){
    pattern <- paste0('\\b',list_stopwords, '\\b', collapse="|")
    texts <- gsub(pattern, '', texts, ignore.case = T)
    texts <- trimws(gsub("[[:space:]]{2,}", " ", texts))
  }

  # umlaut normalization
  if(umlaut_normalization){
    texts <- gsub('ä', 'a',
                  gsub('ü', 'u',
                       gsub('ö', 'o',
                            gsub('ß', 'ss',
                                 gsub('Ä', 'A',
                                      gsub('Ü', 'U',
                                           gsub('Ö', 'O', texts)
                                      ))))))

  }

  ###print(lang_stemming)
  # stemming
  if(stemming){
    texts <- sapply(strsplit(texts, " "), function(r) paste0(SnowballC::wordStem(r, language = lang_stemming), collapse = " "))
  }
  return(trimws(texts))
}

find_centroid <- function(words, matrix){
  #search words occour in matrix
  collect = c()
  for(word in words){
    if(word %in% rownames(matrix)){
      collect <- c(collect, word)
    }
  }
  if(length(collect) > 1){
    return(colMeans(matrix[collect,]))

  }else if(length(collect) == 1){
    return(matrix[collect,])

  }else{
    warning("Couldn't find words in corpus!")
    return(rep(NA,ncol(matrix)))
  }
}

entropy <- function(vec){
  if(any(vec > 1)){
    return(print('Numbers bigger then one. Only insert probabilities'))
  }
  return(sum(unlist(lapply(vec, function(x) ifelse(x == 0, 0, -x*log2(x))))))
}


evaluation_multi <- function(probs, testset){
  # evaluates the results in form of probabilities
  # INPUT: a dataframe with colums of probabilities for every given class
  # and testset as vector
  # probs:
  #             one          two        three
  # [1,] 0.31677478 0.3016395087  0.381585712
  # [2,] 0.08153822 0.0124325851  0.906029191
  # [3,] 0.79317329 0.1747493563  0.032077352

  output <- list()
  rlist = list()
  order = colnames(probs)
  classes <- as.character(colnames(probs))

  winner <- function(p) {
    # returns the class with highest probability
    winner <- colnames(probs)[which(p == max(p))]
    return(ifelse(length(winner) > 1, sample(winner,1), winner))
  }

  predict <- apply(probs, 1, function(x) winner(x))
  df <- cbind(probs,  data.frame(predict = as.factor(predict), testset = as.factor(testset)))
  # evaluate
  # fill with levels - considers answers which not occours in the confusion matrix
  levels <- unique(c(levels(df$testset), levels(df$predict)))
  if(length(colnames(probs)) > length(levels)){
    levels <- colnames(probs)
  }
  if(!all(classes %in% levels))
    levels <- c(levels, classes[!classes %in% levels])
  levels(df$testset) <- c(levels(df$testset),levels[!levels %in% levels(df$testset)])
  levels(df$predict) <- c(levels(df$predict),levels[!levels %in% levels(df$predict)])
  # confusion matrix
  c_mat <- table(df$predict, df$testset)[order, order]
  mat <- c_mat/sum(c_mat)

  observed <- mat/sum(mat)
  chance <- outer(rowSums(mat), colSums(mat))/sum(outer(rowSums(mat), colSums(mat)))

  weights <- mat - mat
  for(i in 1:nrow(mat)) {
    for(j in 1:ncol(mat)) {
      weights[i,j] <- ((i-j)^2)/(nrow(mat)-1)^2
    }
  }

  qO = 1-sum(diag(observed))
  qE = 1-sum(diag(chance))
  kappa_gen <- kappa <- 1-(qO/qE)

  wO = sum(weights*observed)
  wE = sum(weights*chance)

  weighted_kappa = 1-(wO/wE)
  wO2 = sum(weights^2*observed)
  wE2 = sum(weights^2*chance)

  expectedAccuracy = sum(diag(chance))
  observedAccuracy = sum(diag(mat))
  # kappa = (observedAccuracy - expectedAccuracy)/(1 - expectedAccuracy)

  extract_individual <- function(positive){
    list <- list()
    negative = 'negative'
    mat_pos <- table(
      factor(ifelse(df$predict == positive, positive, 'negative'), levels = c(positive, 'negative')),
      factor(ifelse(df$testset == positive, positive, 'negative'), levels = c(positive, 'negative'))
    )

    expectedAccuracy = max(colSums(mat_pos)/sum(colSums(mat_pos)))
    observedAccuracy = sum(diag(mat_pos))/ sum(mat_pos)
    kappa = (observedAccuracy - expectedAccuracy)/(1 - expectedAccuracy)

    list[['expectedAccuracy']] <- expectedAccuracy
    list[['observedAccuracy']] <- observedAccuracy
    list[['Kappa']] <- kappa

    list[['TP']] = mat_pos[positive,positive]
    list[['TN']] = mat_pos[negative,negative]
    list[['FP']] = mat_pos[negative,positive]
    list[['FN']] = mat_pos[positive,negative]

    list[['TPR']] = (list[['TP']]/(list[['TP']]+list[['FN']])) #Sensitivity/ Recall
    list[['TNR']] = (list[['TN']]/(list[['TN']]+list[['FP']])) #Specifity/ Selectivity
    list[['PPV']] = (list[['TP']]/(list[['TP']]+list[['FP']])) # Precision
    list[['F1']] = 2*((list[['PPV']]*list[['TPR']])/(list[['PPV']]+list[['TPR']]))

    list[['TP']] <- NULL
    list[['TN']] <- NULL
    list[['FP']] <- NULL
    list[['FN']] <- NULL

    for(i in 1:length(list)){
      if(is.nan(list[[i]]))
        list[[i]] <- 0
    }

    return(list)
  }

  results <- as.data.frame(do.call(rbind.data.frame, lapply(order, function(lvl) extract_individual(lvl))))
  results <- rbind(results, colMeans(results))
  results$weighted_kappa <- c(rep(NA, nrow(results)-1), weighted_kappa)
  results$levels <- c(levels, 'Total')

  # results[which(results$levels == 'Total'), 'expectedAccuracy'] <- expectedAccuracy
  results[which(results$levels == 'Total'), 'observedAccuracy'] <- observedAccuracy
  results[which(results$levels == 'Total'), 'Kappa'] <- kappa
  rownames(results) <- NULL
  results$Kappa <- rep(kappa_gen, nrow(results))
  results$expectedAccuracy <- expectedAccuracy
  results$TNR[nrow(results)] <- NA
  results$TPR[nrow(results)] <- NA
  results$PPV[nrow(results)] <- NA
  results$F1[nrow(results)] <- NA
  #reorder
  results <- results[,c(1:3,8,4:7,9)]
  output[['results']] <- results
  # output[['confusion matrix']] <- mat
  return(output)
}


get_bayes <- function(mat){
  p.cluster <- sum(apply(mat, 2, function(col) (col[1]/sum(mat[1,]))*(sum(col)/sum(mat))))
  sapply(1:length(colnames(mat)), function(i) {
    class <- colnames(mat)[i]
    p.class <- sum(mat[,i])/sum(mat)
    p.clust.class <-   mat[1,i]/sum(mat[1,])
    p.class.clust <- (p.clust.class*p.class)/p.cluster
    setNames(p.class.clust, paste0(class))
  })
}

classify <- function(text,
                     data, clust_info,
                     corpus_space,
                     dist_measure,
                     clust_method,
                     clustnum){
  #clustnum <- nrow(clust_info)
  data <- na.omit(data)
  data$mancode <- as.factor(data$mancode)
  suggest <- list()
  suggest$text_processed <- text

  # calculate vector of the answer
  if(any(unlist(suggest['text_processed']) %in% row.names(corpus_space))){
    centroid_answer <- as.numeric(find_centroid(unlist(suggest['text_processed']), corpus_space))
    centroid_answer <- as.data.frame(matrix(centroid_answer, nrow = 1, ncol = length(centroid_answer)))
    names(centroid_answer) <- paste0('center_d', 1:length(centroid_answer))
    suggest$centroid_answer <- centroid_answer
  } else # None of the tokens is covered by the semantic space
    return(F)

  # calculate distance between all clustercentroids and the vector of the answer
  clust_info$ansVec <- 0
  for(i in 1:nrow(clust_info)){
    d <- proxy::dist(x = suggest$centroid_answer,
                     y = clust_info[i, grep('center_d', names(clust_info))] , method = dist_measure)
    clust_info$ansVec[i] <- d
  }
  # order them (important for knn!)
  clust_info <- clust_info[order(abs(clust_info$ansVec)),]

  # get the distribution of the cluster (or the k-answers)
  if(clust_method == 'knn'){
    # stores k clustnr
    # takes the next k-clusters, where every point represents a own cluster
    clustnr <- clust_info$cluster[1:clustnum]
    distribClust <- table(data$mancode[which(data$cluster %in% clustnr)])

  } else {
    # stores one clustnr
    # nearest cluster
    clustnr <- clust_info$cluster[1]
    distribClust <- table(data$mancode[which(data$cluster == clustnr)])
  }

  # distribOthers <- table(data$mancode[which(!(data$cluster %in% clustnr))])
  # add classes from the empty (text) answers
  # empty_answers <- table(data$mancode[which(is.na(data$text))])
  # if(length(empty_answers) > 0){
  #   for(name in names(empty_answers)){
  #     distribOthers[name] <-  distribOthers[name] +  empty_answers[name]
  #   }
  # }

  prob <- distribClust/sum(distribClust)
  # prob <- get_bayes(rbind(distribClust, distribOthers))

  code <- names(prob)[which(prob == max(prob))][1]
  code <- ifelse(length(code) == 1, code, sample(code ,1))

  # creates a subspace with answers inside the nearest cluster
  if(clust_method == 'knn'){
    #data$dist_to_center <- clust_info$ansVec
    subspace <- na.omit(data[data$cluster %in% clustnr,])
    #acos(min(1, sum(x*y)/(sqrt(sum(x^2))*sqrt(sum(y^2)))))

    #acos(min(1, sum(suggest$centroid_answer*data[nr, grep('lsa_d', names(data))])/(sqrt(sum(suggest$centroid_answer^2))*sqrt(sum(data[nr, grep('lsa_d', names(data))]^2)))))

    # calculate distance to answer
  } else {
    subspace <- na.omit(data[data$cluster == clustnr,])
  }

  # order subspace in depending on its distance to the clustercentroid
  # subspace <- subspace[order(abs(subspace$angle)),c('text','mancode', 'angle')]
  subspace <- subspace[order(abs(subspace$dist_to_center)),c('text', 'text_processed','mancode', 'dist_to_center')]
  # rownames(subspace) <- NULL
  #print(head(subspace))

  # add information to output list
  suggest$prob <- prob
  suggest$code <- code
  suggest$clustnr <- clustnr
  suggest$dist_to_center <- clust_info$ansVec[1]
  suggest$simil_answers <- subspace


  # #print(suggest)
  return(suggest)
}

set_confvar <- function(var, value, project, update = T) {
  write(value, file = paste0('projects/', project, '/config/', var))
  if(update)
    update_proj_conf(project)
}

read_confvar <- function(var, project) {
  readLines(paste0('projects/', project, '/config/', var))
}

delete_confvar <- function(var, project, update = T) {
  if(file.exists(paste0('projects/', project, '/config/', var))) {
    file.remove(file = paste0('projects/', project, '/config/', var))
    if(update)
      update_proj_conf(project)
  }
}

update_proj_conf <- function(project) {
  if(!dir.exists(paste0('projects/', project)))
    warning(paste0("Project `", project, "` does not exist."))

  proj_conf <- c()
  for(config_f in list.files(paste0('projects/', project, '/config/'))) {
    proj_conf <- append(proj_conf, scan(file = paste0('projects/', project, '/config/', config_f), what = character(), quiet = T))
    names(proj_conf)[length(proj_conf)] <- config_f
  }

  assign("proj_conf", proj_conf, envir = .GlobalEnv)
}

add_br_to_string <- function(text, max_char = 40, cut_at = 120){
  # max_char = maximal character per line
  # cut_at = maximal characters for the whole text
  words <- strsplit(text, ' ')[[1]]
  # decrease maximal string length
  last_word <- max(which((cumsum(nchar(words)+1) < cut_at) == T))
  if(last_word != length(words)){
    words <- c(words[1:last_word], '...')
  }

  sent = words[1]
  llen <- nchar(sent)
  rows = 0

  if(length(words) > 1){
    for(i in 2:length(words)){
      if((llen + nchar(words[i])) <= max_char){
        sent <- paste0(sent, ' ', words[i])
        llen <- llen + nchar(words[i]) + 1
        rows = rows + 1
      } else {
        sent <- paste0(sent, ' ', '<br>', words[i])
        llen <- nchar(words[i])
        rows = rows + 1

      }
    }
  }
  return(sent)
}

restore_combine <- function(raw_dataframe, omitted_data, ids){
  # restore the omitted dataframe with NAs - to the length (nrow) of the raw_dataframe
  if(!(is.data.frame(raw_dataframe) || is.data.frame(omitted_data))){
    stop('Only accept data frames!')
  } else {
    # creates a dataframe with NAs and
    df = data.frame(matrix(vector(), nrow(raw_dataframe), length(names(omitted_data)),
                           dimnames=list(c(), names(omitted_data))), stringsAsFactors = F)
    # replace the important rows  with values
    # rows without NAs
    # idx <- which(apply(raw_dataframe, 1, function(x) !any(is.na(x))))
    df[ids,] <- omitted_data[,names(omitted_data)]

    # if values already in dataframe -> replace them
    if(all(colnames(df) %in% raw_dataframe)){
      raw_dataframe[,names(df)] <- df
      # if values not in dataframe -> combine
    } else {
      raw_dataframe <- cbind(raw_dataframe, df)
    }
    return(raw_dataframe)
  }
}

copyDir <- function(from, to) { # Adopted from R.utils::copyDirectory()
  files <- list.files(from, all.files = T, full.names = FALSE)
  files <- files[!basename(files) %in% c(".", "..")]
  files <- file.path(from, files)

  dir.create(to)

  copiedFiles <- c()
  for (file in files) {
    basename <- basename(file)

    if (!file.info(file)["isdir"][1,1]) {
      if (file.copy(from=file, to=file.path(to, basename))) {
        copiedFiles <- c(copiedFiles, file)
      }
    } else if (file.info(file)["isdir"][1,1]) {
      copiedFiles <- c(copiedFiles,
                         copyDir(file, to = file.path(to, basename)))
    }
  }
  invisible(copiedFiles)
}

information_value <- function(texts) {
  # takes texts/ responses as input
  # returns values for information strength
  # sum of relative frequency
  texts <- preprocess(texts, del_stopwords = F, del_numbers = F, stemming = F, umlaut_normalization = F)
  words <- as.character(unlist(sapply(texts, function(x) strsplit(x, ' ')[[1]])))
  words <- words[!nchar(words) == 0]
  freqs <- table(words)/max(table(words))
  info_value <- as.numeric(
    sapply(tolower(texts), function(x) {
      tokens <- strsplit(x, ' ')[[1]]
      tokens <- tokens[!nchar(tokens) == 0]
      tokens <- freqs[tokens]
      sum(tokens)
      # sum(sapply(tokens, function(y) freqs[y]))
    }))
  return(info_value)
}

hunspell4reco <- function(raw, hdict, freq, do = F) {
  if(!do) {
    return(raw)
  } else {
    names(raw) <- 1:length(raw) # required for proper hunspell()-handling
    improper <- hunspell::hunspell(raw, dict = hdict)
    candidates <- sapply(improper, function(x) hunspell::hunspell_suggest(x, dict = hdict)) # Find replacement candidates
    rep_cache <- character()

    # Identify the most probable (frequent) replacement and do so (#ToDo: Include context for some languages)
    invisible(sapply(1:length(raw), function(r) { # Walk through all responses; r = index
      if(length(improper[[r]])>0) {
        sapply(1:length(improper[[r]]), function(m) { # Walk through all mistakes; m = index
          replacement <- NULL
          if(improper[[r]][m] %in% names(rep_cache)) {
            replacement <- rep_cache[improper[[r]][m]] # This word had been replaced before; since we currently work w/o contextualization, we take the cached replacement for speeding things up -> Change when contextualization is included
          } else if(length(candidates[[r]])>0) { # Ok, let's find the best among the candidates if there are any
            s <- candidates[[r]][[m]]
            s <- setNames((length(s):1)/sum(length(s):1), s)
            l <- freq[names(s),"freq"]
            l[is.na(l)] <- min(freq$freq)
            s <- s*l / sum(s*l)
            replacement <- names(s)[which.max(s)[1]]
            rep_cache[improper[[r]][m]] <- replacement
          }
          if(!is.null(replacement)) {
            raw[r] <<- gsub(paste0("\\b", improper[[r]][m], "\\b"), replacement, raw[r])
          }
        })
      }
    }))

    return(raw)
  }
}

fishersZ <- function(value, reverse = F) {
  if(!reverse) {
    value <- ifelse(value == 1, 1-2.16e-16, value)
    return(value = 0.5*log((1+value)/(1-value)))
  } else {
    return((exp(2*value)-1)/(exp(2*value)+1))
  }
}

# Unregisters all parallel backends
unregister <- function() {
  env <- foreach:::.foreachGlobals
  rm(list=ls(name=env), pos=env)
}

