# Load Dependencies
suppressMessages(library(shiny)) # Most packages unfortunately do not acknowledge suppressPackageStartupMessages()
suppressMessages(library(shinythemes)) # Thus, we use the German crowbar =)
suppressMessages(library(shinyjs))
suppressMessages(library(plotly))
suppressMessages(library(DT))
suppressMessages(library(wordcloud))
suppressMessages(library(kknn))
suppressMessages(library(proxy))
suppressMessages(library(SnowballC))
suppressMessages(library(stopwords))
suppressMessages(library(stringr))
suppressMessages(library(shinycssloaders))
suppressMessages(library(colourpicker))
suppressMessages(library(hunspell))
suppressMessages(library(doSNOW))
suppressMessages(library(foreach))
suppressMessages(library(Rtsne))
suppressMessages(library(circular))
suppressMessages(library(grid))

# Manage Package Calls
pkgname <- "shinyReCoR"
if(! pkgname %in% (.packages())) { # Load the package if not yet done
  assign("suppressrecoappstartcommandmessage", T, envir = .GlobalEnv)
  library(pkgname, character.only = T)
}

# Working Directory and Shiny Options
setwd(system.file(package = pkgname))
options(shiny.maxRequestSize = 100*1024^2)

# Register Resource Paths
shiny::addResourcePath("corpus_vsm", system.file("corpus_vsm/", package = "shinyReCoR"))
shiny::addResourcePath("projects", system.file("projects/", package = "shinyReCoR"))
shiny::addResourcePath("response_data", system.file("response_data/", package = "shinyReCoR"))
shiny::addResourcePath("www", system.file("www/", package = "shinyReCoR"))

# Set Up Languages
languages <- list(  "English" = "_en",
                    # "Dutch" = "_nl",
                    "German" = "_de"
                  )
lang_codes <- list( "_en" = list(stopwords = 'en', stemming = 'english', spellchecker = 'en_US'),
                    "_nl" = list(stopwords = 'nl', stemming = 'dutch', spellchecker = ''),
                    "_de" = list(stopwords = 'de', stemming = 'german', spellchecker = 'de_DE'))

preprocessing_samples <- c("_de"="Das ist ein Beispieltext, um den Prozess der Vorverarbeitung zu verdeutlichen. Abhängig von der Analyse können Sie die Bearbeitungsschritte individuell bestimmen und Änderungen vornehmen. Typischerweise werden dabei Interpunktion und Zahlen (wie 9) entfernt und es werden nur sogenannte autosemantischen Wörter beibehalten; also solche, die für sich selbst eine Bedeutung haben.",
                           "_en"="This is a sample text to illustrate the process of preprocessing. Depending on the analysis, you can determine the processing steps individually. Typically, punctuation is removed and only so-called autosemantic words are retained; that is, words that have a meaning in themselves.",
                           "_es"="Este es un texto de muestra para ilustrar el proceso de preprocesamiento. Dependiendo del análisis, se pueden determinar los pasos de procesamiento individualmente. Normalmente, se elimina la puntuación y sólo se conservan las llamadas palabras autosemánticas, es decir, las palabras que tienen un significado en sí mismas.",
                           "_fr"="Voici un exemple de texte pour illustrer le processus de prétraitement. En fonction de l'analyse, vous pouvez déterminer les étapes de traitement individuellement. Généralement, la ponctuation est supprimée et seuls les mots dits autosémantiques sont conservés, c'est-à-dire les mots qui ont un sens en eux-mêmes.",
                           "_nl"="Dit is een voorbeeldtekst om het proces van voorbewerking te illustreren. Afhankelijk van de analyse kunt u de verwerkingsstappen afzonderlijk bepalen en wijzigingen aanbrengen. Typisch worden interpunctie en getallen (zoals 9) verwijderd en alleen zogenaamde autosemantische woorden worden behouden; d.w.z. woorden die op zichzelf een betekenis hebben.")
# preprocessing_text_notimportedyet <- c("_de"="Bitte zunächst die Daten in Schritt 1 importieren.",
#                                        "_en"="Please import the data in step 1 first.",
#                                        "_es"="Por favor, importa primero los datos del paso 1.",
#                                        "_fr"="Veuillez d'abord importer les données de l'étape 1.")
preprocessing_text_nomoreclasssamples <- c("_de"="Es sind keine weiteren Antworten dieser Klasse im Datensatz vorhanden.",
                                       "_en"="There are no more responses of this class in the data.",
                                       "_es"="No hay otras respuestas de esta clase en el registro.",
                                       "_fr"="Il n'y a pas d'autres réponses de cette classe dans le enregistrement.",
                                       "_nl"="Er zijn geen andere reacties van deze klasse in de gegevensverzameling.")
reco.srv.url <- "www.reco.science" # ReCo's URL

#To DO: erweitern
color_scheme <- c("#F29400","#6DA5D5", "#646567", "#37A42C", "#D40050")

# Corpora
corpus_vsms <- as.list(c(list.files("corpus_vsm", pattern = "_corpusspace"), paste0("#dev#/", list.files("corpus_vsm/#dev#", pattern = "_corpusspace"))))
names(corpus_vsms) <- sapply(strsplit(unlist(corpus_vsms), "_"), function(x) x[1])
if(any(names(corpus_vsms)=="#dev#/")) {
  corpus_vsms[[which(names(corpus_vsms)=="#dev#/")]] <- NULL
}

# Project List
projects <- as.list(list.files('projects')) # TODO: Make sure the project folder contains the required config file
if(length(projects)>0) { # Load projects that are already in the projects folder
  names(projects) <- paste0("Project `", as.character(projects), "`")
  current_project_i <- file.exists(paste0(sort(Sys.glob("projects/*/config/")), "recent_project"))
  current_project <- if(any(current_project_i)) { projects[current_project_i][1] } else { projects[which.max(file.info(Sys.glob("projects/*"))[,"atime"])] }
  names(projects)[names(projects)=="Project `unnamed`"] <- "Unnamed Project"
  # TODO: select default language
} else { # No projects in the project folder. Therefore, create the first one.
  projects <- list("Unnamed Project"="unnamed")
  current_project <- "unnamed"
  dir.create(paste0("projects/", current_project))
  dir.create(paste0("projects/", current_project, "/config"))
  write(languages[[1]], file = paste0('projects/', current_project, '/config/lang')) # TODO: Some mechanism for deciding for a reasonable language.
  write("", file = paste0('projects/', current_project, '/config/recent_project'))
  saveRDS(stopwords::stopwords(lang_codes[[1]]$stopwords), file = paste0('projects/', current_project, '/stopwords.RDS'))
}

# Read Project Configuration and load into global proj_conf object
update_proj_conf(current_project)

# *** USER INTERFACE *** ----
ui <- tagList(
  navbarPage(
    windowTitle = "ReCo",
    title = div(a(href = "https://www.reco.science", target = "_blank", id = "ReCo-title", title = "Website of ReCo (Automatic Text Response Coder)", tags$img(src = "www/ReCo-Logo-Appklein.png", style = "height:36px;margin-top:-7px;"))),
    id = 'ReCo',
    selected = "RESPONSE ANALYSIS",
    theme = shinytheme("yeti"),
    header = tagList(
      # LOGO ----
      tags$div(HTML("provided by&nbsp;"), style = "position:absolute;top:5px;right:5px;z-index:1000;color:#FFFFFF;font-size:10px;", tags$a(href = "https://tba.dipf.de/en", target="_blank", tags$img(src = "www/TBA-Logo.png", style = "height:36px;vertical-align:bottom;", title = "Website of the Centre for Technology-Based Assessment at DIPF", alt = "Logo of TBA")), HTML("<span style=\"writing-mode: vertical-rl;transform:rotate(180deg);\"><a href=\"https://www.dipf.de/en/\" target=\"_blank\" style=\"color:#FFFFFF;\" title=\"Website of DIPF | Leibniz Institute for Research and Information in Education\">@DIPF</a></span>")),
      # INTEGRATION OF shinyJS ----
      shinyjs::useShinyjs(),
      # THEME STYLING AND ADDITIONAL JS ----
      tags$head(
        tags$link(rel = "stylesheet", type = "text/css", href = "www/style.css"),
        tags$script("src"="www/script.js")
      ),
    ),
    # I) SETUP ----
    tabPanel("SETUP & CORPUS",
             column(2,
                    titlePanel("Parallel Processing"),
                    checkboxInput("do_parallel", "Enable Parallel Processing", ifelse(!is.na(proj_conf["do_pp"]), proj_conf["do_pp"]=="T", F)),
                    sliderInput("parallel_cores", label = HTML("Number of cores you want the app to use: <img src=\"www/info.svg\" style=\"height:1em;margin-bottom:4px;\" title=\"", parallel::detectCores(), "CPU cores available, but we recommend leaving one core for the rest of your system.\">"), min = 1, max = parallel::detectCores(), step = 1, ticks = F, value = ifelse(!is.na(proj_conf["pp_cores"]), as.numeric(proj_conf["pp_cores"]), (parallel::detectCores()-1)), width = "100%"),
             ),
             column(9,
                    titlePanel(div("Download Prebuilt Semantic Spaces", style = "margin-left:30px;")),
                    div("ReCo provides prebuilt semantic spaces. As their file size can be rather large, you can here manage and download them. The list of available prebuilt spaces will grow over time.", style = "margin-left:30px;"),
                    div(withSpinner(uiOutput("prebuilt_vsms_status"), size = 0.5, proxy.height = "40px"), style = "margin-top:10px;"),
                    wellPanel(DT::dataTableOutput("prebuilt_vsms"), style = "overflow-x:auto; max-width:100%;;border:none;margin-left:30px;overflow-y:visible;padding:0px;"),
                    div(textInput("prebuiltvsms_download_filename", "", width = "100%"), style = "visibility:hidden;"),
                    div(textInput("prebuiltvsms_download_name", "", width = "100%"), style = "visibility:hidden;"),
                    div(textInput("prebuiltvsms_download_mode", "", width = "100%"), style = "visibility:hidden;")
             )),
    # II) PROJECT MANAGEMENT ----
    navbarMenu("PROJECT MANAGEMENT",
               tabPanel('New Project',
                        column(2),
                        column(7,
                               titlePanel("New Project"),
                               br(),
                               tagAppendAttributes(
                                 textInput("project_name", h4("Project Name")),
                                 `data-proxy-click` = "do_createProject"
                               ),
                               uiOutput("projectnameHint", class = "reco_warn"),
                               selectInput("select_language", label = h4("Select Language"),
                                           choices = languages),
                               actionButton("do_createProject", "Create Project"),
                               br(),
                               uiOutput("newProjectText", class = "reco_success", style = "margin-top:10px;")
                               )
                        ),

               tabPanel('Delete Project',
                        column(2),
                        column(7,
                               titlePanel("Delete Project"),
                               br(),
                               uiOutput("delete_project_select"),
                               actionButton('delete_project', 'Delete'),
                               uiOutput('delFolderText', class = "reco_err", style = "margin-top:15px;")
                        )
               ),

               "-----",
               tabPanel('Project Settings',
                        column(2),
                        column(7,
                               titlePanel("Settings of the Currently Selected Project"),
                               br(),
                               uiOutput("change_project_name_ui"),
                               uiOutput("changeprojectnameHint", class = "reco_warn"),
                               disabled(selectInput("change_project_language", label = h4("Project Language"),
                                           choices = languages,
                                           selected = proj_conf["lang"])),
                               actionButton("do_changeProjectSettings", "Change Settings"),
                               br(),
                               uiOutput("changeProjectText", class = "reco_success", style = "margin-top:10px;")
                        )
               ),

               tabPanel('Code Labels',
                        column(3),
                        column(3,
                               titlePanel("Code Labels"),
                               br(),
                               selectInput("label_select", label = "Select Code", choices = list()),
                               actionButton('label_reset', HTML('<img src="www/trash-2.svg" style="height:1em;margin-bottom:7px;margin-right:5px;">Reset All Code Labels'), class = "decent_button", style = "margin-bottom:33px;padding-top:5px;"),
                               uiOutput('missing_codes'),
                               hr(),
                               selectInput('label_plot', h3('Exemplary Visualization', style = "margin-top:0px;"), choices = list('Pie Chart' = 'pie_chart',
                                                                                                   'Random Scatter Plot' = 'random_scatter'),
                                           selected = 'pie_chart'),
                               plotlyOutput('class_distribution')
                        ),
                        column(3,
                               div(textInput("label_rename", label = "Code Label", value = "", placeholder = "Enter the Code's Label Here"), style = "margin-top:86px;"),
                               div(colourpicker::colourInput("label_color", label = "Color", showColour = 'text', value = '')),
                               hr(),
                               div(h3("Code Information")),
                               uiOutput('class_information')
                        ),
                        column(3)
               )
    ),

    # III) RESPONSE ANALYSIS ----
    tabPanel("RESPONSE ANALYSIS", id = "RA",
             # Action- & Info Sidebarpanel ----
             sidebarPanel(
               width = 2,
               # Function Buttons
               div(
                 actionButton("btn_save", "Save", class = "btn-primary btn-reco-icon", style = "background-image:url(www/save.svg);margin-bottom:6px;", alt = "Floppy Disk", title="Save Current Session"),
                 actionButton("btn_xpert", "Expert", class = "btn-primary btn-reco-icon", style = "background-image:url(www/key.svg);margin-bottom:6px;", alt = "Key", title = "Switch to Expert Mode"),
                 style = "text-align:center;margin-top:-14px;"
               ),

               # Project Selection
               div(selectInput("current_project_sel", label = "",
                               choices = list()), style = "margin-bottom:-12px;"),
               div(actionButton("btn_edit", HTML("Edit"), class = "btn-primary btn-reco-icon", style = "background-image:url(www/edit.svg);margin-bottom:6px;background-size:17px 17px;background-position:4px center;font-size:0.9em;padding:3px 5px 3px 24px;", alt = "Pen", title = "Edit Project Settings", onclick = "customHref('Project Settings');"), style = "text-align:center;"),
               #---

               # Information Text
               h4("Information", style="margin-top:15px;"),
               uiOutput("txt_RA_info", style = "font-size:12px;")
             ),

             # Main Panel ----
             mainPanel(
               width = 10,
               tabsetPanel(id = "tabs_RA",
                           # Panel #1 Data Import ----
                           tabPanel(title = "1 Data", value = "tabp_RA_1",
                                    column(5,
                                            # Data Import
                                           div(span("or:", style="font-style:italic;"), actionButton("importdemodata", "Import Demo Data Set", class="decent_button", style="font-size:1em;font-weight:600;background-color:#FFFFFF;margin-left:2px;"), HTML("<a href=\"#\" onclick=\"customHref('tabp_INFO');\" title=\"Click to view details on the demo dataset\"><img src=\"www/info.svg\" style=\"height:1em;margin-left:2px;\" alt=\"info\"></a>"), style = "float:right;margin-top:29px;font-size:0.8em;"),
                                           titlePanel("Data Import"),
                                           htmlOutput("importfile_name", container = div, class = "highlighted", style = "font-size:0.8em;"),
                                           fileInput("importfile", "Choose the CSV file with your Response Data:",
                                                     multiple = FALSE,
                                                     accept = c("text/csv",
                                                                "text/comma-separated-values,text/plain",
                                                                ".csv")),

                                           # Horizontal line
                                           tags$hr(style="margin-top:-5px;"),
                                           # Input: Select column names for text and manual code
                                           selectInput('text_column', label = HTML('Name of Column Containing <span style="font-weight:bold;">Text Reponses</span>:'), choices = list(), width="100%", selectize = T),
                                           # htmlOutput("textcolumn_warning", container = div), # TODO: Add more obvious warning when original data column is overwritten
                                           selectInput('mancode_column', label = HTML('Name of Column Containing <span style="font-weight:bold;">Manual Codes</span>:'), choices = list(), width = "100%", selectize = T),
                                           # htmlOutput("textcolumn_warning", container = div), # TODO: Add more obvious warning when original data column is overwritten

                                           # Horizontal line
                                           tags$hr(style="margin-top:21px;"),
                                           h4("CSV Properties"),
                                           selectInput("encoding", label = HTML("<span style=\"font-weight:bold;\">Encoding</span>"),
                                                       choices = list("Automatic" = "GUESS",
                                                                      "UTF-8" = "UTF-8",
                                                                      "UTF-8-BOM" = "UTF-8-BOM",
                                                                      "Latin-1" = "latin1"),
                                                       selected = "GUESS",
                                                       width = 120),
                                           # Input: Checkbox if file has header
                                           checkboxInput("header", HTML("<span style=\"font-weight:bold;\">Header</span>: First row contains column names"), TRUE),
                                           # Input: Select separator
                                           radioButtons("sep", HTML("<span style=\"font-weight:bold;\">Column Separator</span>"),
                                                        choices = c("Comma" = ",",
                                                                    "Semicolon" = ";",
                                                                    "Tab" = "\t"),
                                                        selected = ","),

                                           # Input: Select quotes
                                           radioButtons("quote", HTML("<span style=\"font-weight:bold;\">Quote</span>"),
                                                        choices = c(None = "",
                                                                    "Double Quote" = '"',
                                                                    "Single Quote" = "'"),
                                                        selected = '"'),

                                           # Input: Select Missing values
                                           textInput("missing_chars", label = HTML("<span style=\"font-weight:bold;\">Missing Values</span> <span style=\"font-size:0.8em;\">(other than empty; separate multiple by comma)</span>"), placeholder = "String indicating a missing value", value = "NA"),

                                           # Horizontal line
                                           tags$hr(),

                                    ),
                                    column(6,
                                           # Input: Select number of rows to display
                                           div(HTML("&nbsp;"), style="margin-top:64px;"),

                                           htmlOutput("data_summary", style = "font-size: 0.8em;margin-bottom:-7px;"),
                                           withSpinner(DT::dataTableOutput("import_preview"), color = '#006699')
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('next_RA_1', class = "btn-navig", HTML("&raquo;"), style = "margin-left:39px;")
                                    )
                           ),
                           # Panel #2.1 Preprocessing ----
                           tabPanel("2.1 Preprocessing", value = "tabp_RA_2.1",
                                    column(5,
                                           titlePanel("Preprocessing Pipeline"),
                                           # Input: Select separator
                                           tags$br(),
                                           checkboxInput("decapitalize", "Decapitalize", (ifelse(is.na(proj_conf["pp_decapitalize"]), T, proj_conf["pp_decapitalize"]=="TRUE"))),
                                           checkboxInput("umlaut_normalization", "Character Normalization (e.g., Umlauts)", (ifelse(is.na(proj_conf["umlaut_normalization"]), T, proj_conf["umlaut_normalization"]=="TRUE"))),
                                           checkboxInput("del_numbers", "Remove Digits", (ifelse(is.na(proj_conf["pp_del_numbers"]), T, proj_conf["pp_del_numbers"]=="TRUE"))),
                                           checkboxInput("del_symbols", "Remove Symbols & Punctuation", (ifelse(is.na(proj_conf["pp_del_symbols"]), T, proj_conf["pp_del_symbols"]=="TRUE"))),
                                           tags$br(),
                                           checkboxInput("spelling_correction", HTML("Spelling Correction <img src='www/info.svg' title='Takes substantially more time. Consider enabling parallel processing in the Setup & Corpus Tab when activated.' style='height:1em;margin-bottom:4px;'>"), (ifelse((is.na(proj_conf["pp_spelling_correction"]) || nchar(lang_codes[[proj_conf["lang"]]]$spellchecker)==0), F, proj_conf["pp_spelling_correction"]=="TRUE"))),
                                           checkboxInput("del_stopwords", "Remove Stop Words", (ifelse(is.na(proj_conf["pp_del_stopwords"]), T, proj_conf["pp_del_stopwords"]=="TRUE"))),
                                           checkboxInput("stemming", "Stemming", (ifelse(is.na(proj_conf["pp_stemming"]), T, proj_conf["pp_stemming"]=="TRUE"))),
                                           actionButton('do_modal_change_stopwords', 'Change List of Stop Words', class = 'decent_button'),
                                           tags$hr(),
                                           actionButton('do_preprocess', 'Start Preprocessing'),
                                           actionButton('do_modal_preprocess', 'View Preprocessing Results'),
                                           withSpinner(uiOutput('preprocessing_status_display', class = "highlighted", style = "margin-top:5px;font-size:0.8em;"), color = "#006699")
                                    ),
                                    column(6,
                                           selectInput("text_preprocessing", label = h4("Preview", style = "margin-top:55px;"), ""),
                                           tags$hr(style = "margin-top:-2px;"),
                                           tags$h4("Before Preprocessing"),
                                           textOutput("exampletext_pre"),
                                           tags$hr(),
                                           tags$h4("After Preprocessing"),
                                           textOutput("exampletext_post")
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_2.1', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_2.1', class = "btn-navig", HTML("&raquo;"))
                                    )

                           ),

                           # Panel #2.2 Word Predictiveness ----
                           tabPanel("2.2 Word Predictiveness", value = "tabp_RA_2.2",
                                    column(5,
                                           titlePanel("Word Predictiveness and Frequency"),
                                           withSpinner(DT::dataTableOutput("frequency_tableDT"), color = '#006699'),
                                           tags$br(),
                                           downloadButton("download_predictiveness", "Download Table")

                                    ),

                                    column(6,
                                           div(
                                            selectInput("select_order_cloud", label = "",
                                                       width = 200,
                                                       choices = c('Highlight Most Predictive' = 1,
                                                                   'Highlight Most Frequent' = 5),
                                                       selected = 1),
                                            style = "margin-top:45px;float:right;"
                                           ),
                                           div(
                                             sliderInput("wordcloud_minfreq",
                                                         label = div(style="position:relative;font-size:.75em;text-align:right;top:39px;left:-48px;",
                                                                     "min. value"
                                                         ),
                                                         min = 1, max = 100, step = 1, value = 3, width = 125, ticks = F,
                                             ),
                                             style = "float:right;margin-right:15px;margin-top:41px;"
                                           ),
                                           withSpinner(plotOutput("wordcloud", height = "600px"), color = '#006699')
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_2.2', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_2.2', class = "btn-navig", HTML("&raquo;"))
                                    )
                           ),



                           # Panel #3 Text Corpus ----
                           tabPanel("3 Text Corpus", value = "tabp_RA_3",
                                    column(5,
                                           titlePanel("Create or Select the Text Corpus"),
                                           h4("Create a New Text Corpus"),
                                           tags$p(HTML("In the current app version, only prebuilt corpora are available. In the app's next version, you can create your own text corpus here in this section. <div class=\"highlighted\" style=\"margin-top:15px;\"><img src=\"www/info.svg\" style=\"height:1em;margin-bottom:7px;margin-right:3px;\">Please note that you can download more prebuilt spaces in the <a href=\"#\" onclick=\"customHref('SETUP & CORPUS');\" style=\"font-weight:bold;\">Setup & Corpus Tab</a>.</div>."), style = "margin-top:15px;font-size:0.9em;")
                                    ),
                                    column(6,
                                           h4("OR: Select a Prebuilt Corpus and Semantic Space", style = "margin-top:67px;"),
                                           # uiOutput("select_corpus_space_container"),
                                           DT::dataTableOutput("available_vsms_table"),
                                           br(),
                                           uiOutput('textcorpus_status_display', style = "font-size:0.9em;"),
                                           uiOutput('words_covered', style = "font-size:0.8em;"),
                                           br(),
                                           withSpinner(DT::dataTableOutput("wordlist_uncovered"), color = '#006699')

                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_3', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_3', class = "btn-navig", HTML("&raquo;"))
                                           )

                           ),
                           # Panel #4 Semantic Space ----
                           tabPanel("4 Semantic Space", value = "tabp_RA_4",
                                    column(5,
                                           titlePanel("Semantic Space"),
                                           tags$h4("Compute the Text Corpus's Semantic Space"),
                                           tags$p("In the current app version, only prebuilt corpora and accordingly pre-computed semantic spaces are available. In a later app version, you can compute the semantic space of the selected corpus here in this section.", style = "font-size:0.9em;"),
                                           tags$hr(),
                                           h4("Explore the Text Corpus's Semantic Space"),
                                           tagAppendAttributes(
                                             textInput("word1", label = 'Word 1'),
                                             `data-proxy-click` = "do_wordDistances"
                                           ),
                                           tagAppendAttributes(
                                             textInput("word2", label = 'Word 2'),
                                             `data-proxy-click` = "do_wordDistances"
                                           ),
                                           selectInput("distanceWords", label = h5("Similarity Measure"),
                                                       width = 200,
                                                       choices = c('Cosine' = 'cosine',
                                                                   'Euclidean' = 'euclidean'),
                                                       selected = 'cosine'),
                                           actionButton('do_wordDistances', 'Compare Words', style = "margin-bottom:10px;"),
                                           tags$br(),
                                           uiOutput('show_distanceWords'),
                                           tags$hr(),
                                           tags$div(actionButton("show_semspacelist", "Show Complete List of Tokens", class = "decent_button"), style = "text-align:right;")),


                                    column(6,
                                           h4("Low-Dimensional Representation of the Corpus Semantics", style = "margin-top:67px;"),
                                           tags$p(paste0("Please be aware that the figure below represents a low-dimensional reduction of the hyperdimensional space; typically, 300 dimensions. For visualization purposes, the figure below shows the end point of the response vectors reduced by a dimensionality reduction technique, thus omitting a large share of information."), style = "font-size:0.9em;margin-top:-1px;"),
                                           tags$hr(style="margin-bottom:-5px;"),
                                           div(selectInput("dimensions_corpusspace", label = "",
                                                       width = 200,
                                                       choices = c('Display Interactive 2D Figure' = 2,
                                                                   'Display Interactive 3D Figure' = 3),
                                                       selected = 3), style = "float:left;"),
                                           div(selectInput("dimred_technique_semspace", label = "",
                                                           choices = c("PCA" = "pca",
                                                                       "t-SNE" = "tsne"),
                                                           width = 80,
                                                           selected = "tsne"), style = "float:right;margin-left:10px;margin-right:10px;"),
                                           div(
                                             sliderInput("tsne_sat_semspace",
                                                         label = div(style="width:250px;margin-bottom:-15px;",
                                                                     div(style="float:left;margin-bottom:-15px;font-size:.75em;", "accurate"),
                                                                     div(style="float:right;margin-bottom:-15px;font-size:.75em;", "fast")
                                                         ),
                                                         min = 0, max = 1, step = .1, value = .8, width = 250, ticks = F),
                                             style = "float:right;"
                                           ),

                                           div(withSpinner(plotlyOutput('corpusspace_plotly'), color = '#006699'),style = "clear:both;"),
                                           tags$div("log-frequency", style = "float:right;font-size:0.8em;color:#AFAFAF;position:relative;z-index:1;margin-top:-370px;left:20px;")),

                                    column(1,
                                           br(),
                                           actionButton('prev_RA_4', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_4', class = "btn-navig", HTML("&raquo;"))
                                           )
                                    ),


                           # Panel #5 Response Semantics ----
                           tabPanel("5 Response Semantics", value = "tabp_RA_5",
                                    column(5,
                                           titlePanel("Semantic Response Vectors"),
                                           actionButton('do_responsevectors', 'Compute Response Vectors', style = "margin-top:25px;"),
                                           uiOutput("respvecs_status_display", class = "highlighted", style = "font-size:0.8em;margin-top:3px;"),
                                           uiOutput("info_max_ans", style = "font-size:0.8em;"),
                                           tags$hr(),
                                           tags$h4(id = "respvectors_dt_title", "High-Dimensional Response Vectors"),
                                           wellPanel(DT::dataTableOutput("vsm_dims"), style = "overflow:auto;max-height:500px;width:100%;border-style:none;")
                                    ),
                                    column(6,
                                           h4("Low-Dimensional Representation of the Response Semantics", style = "margin-top:67px;"),
                                           tags$p("Please be aware that the figure below represents a low-dimensional reduction of the hyperdimensional space; typically, 300 dimensions. For visualization purposes, the figure below shows the end point of the response vectors reduced by a dimensionality reduction technique, thus omitting a large share of information.", style = "font-size:0.9em;margin-top:-1px;"),
                                           tags$hr(style="margin-bottom:-5px;"),
                                           div(selectInput("dimensions_respvectors", label = "",
                                                       choices = c('Display Interactive 2D Figure' = 2,
                                                                   'Display Interactive 3D Figure' = 3),
                                                       width = 220,
                                                       selected = 3), style = "float:left;"),
                                           div(selectInput("dimred_technique_respvectors", label = "",
                                                       choices = c("PCA" = "pca",
                                                                   "t-SNE" = "tsne"),
                                                       width = 80,
                                                       selected = "tsne"), style = "float:right;margin-left:10px;margin-right:10px;"),
                                           div(
                                             sliderInput("tsne_sat_respvectors",
                                                         label = div(style="width:250px;margin-bottom:-15px;",
                                                                     div(style="float:left;margin-bottom:-15px;font-size:.75em;", "accurate"),
                                                                     div(style="float:right;margin-bottom:-15px;font-size:.75em;", "fast")
                                                                     ),
                                                         min = 0, max = 1, step = .1, value = .8, width = 250, ticks = F),
                                               style = "float:right;"
                                             ),
                                           div(withSpinner(plotlyOutput('respvectors_plotly'), color = '#006699'), style = "clear:both;")
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_5', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_5', class = "btn-navig", HTML("&raquo;"))
                                    )
                           ),

                           # Panel #6 Response Types ----
                           tabPanel("6 Response Types", value = "tabp_RA_6",
                                    column(5,
                                           titlePanel("Response Types"),
                                           tags$h4("Create Response Clusters by Their Semantic Similarity"),
                                           tags$p("Currently, you need to specify the number of clusters manually. The optimal number depends on the heterogeneity and the number of responses. You are free to experiment with the number. In a later version of the app, it will determine the optimal number of clusters empirically.", style = "font-size:0.9em;"),
                                           tags$hr(),
                                           # Distance Matrix
                                           selectInput("distance", label = h5("Choose Distance Measure"),
                                                       width = 200,
                                                       choices = c('Cosine' = 'cosine',
                                                                   'Euclidean' = 'euclidean'),
                                                       selected = ifelse(!is.na(proj_conf["distMeasure"]), proj_conf["distMeasure"], "cosine")),
                                           # Cluster
                                           selectInput("clustering", label = h5("Choose Clustering Method"),
                                                       width = 200,
                                                       choices = c('Hierarchichal Clustering' = 'hclust',
                                                                   # 'k-Nearest Neighbors' = 'knn',
                                                                   'k-Means' = 'kmeans'),
                                                       selected = ifelse(!is.na(proj_conf["clustMethod"]), proj_conf["clustMethod"], "cosine")),
                                           tags$br(),
                                           uiOutput("clustnum_slider"),
                                           tags$br(),
                                           actionButton("do_clust", "Build Cluster Model"),
                                           uiOutput("respclusters_status_display", class = "highlighted", style = "margin-top:3px;font-size:0.8em;"),
                                           tags$br(),
                                           tags$br(),
                                           downloadButton("download_cluster", "Download Data (incl. Classification)")
                                    ),
                                    column(6,
                                           tags$h4("Cluster Table", style = "margin-top:66px;"),
                                           tags$p("This table provides information on the distribution of classes and distances to the centroid within clusters. Entropy is a measure for cluster purity. With an entropy of 0 there is only one class inside the cluster, with an entropy of 1 the classes are equally distributed. Weighted entropy relates cluster size to entropy.", style = "font-size:0.9em;"),
                                           tags$hr(),
                                           DT::dataTableOutput('clustertable'),
                                           textOutput('knn_info'),
                                           tags$br(),
                                           tags$br(),
                                           tags$br(),
                                           tags$div(textInput("cluster_detail_nr", "", width = "100%"), style = "visibility:hidden;")
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_6', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_6', class = "btn-navig", HTML("&raquo;"))
                                    )
                           ),

                           # Panel #7 Evaluation ----
                           tabPanel("7 Evaluation", value = "tabp_RA_7",
                                    column(5,
                                           titlePanel("Evaluation"),
                                           tags$h4(HTML("Stratified <span style=\"font-style:italic;\">k</span>-Fold Cross Validation")),
                                           tags$hr(),
                                           numericInput('folds', 'Number of Folds', value = 10, min = 5, max = 20, step = 1,
                                                        width = "25%"),
                                           numericInput('repetitions', 'Number of Repetitions', value = 1, min = 1, max = 100, step = 1,
                                                        width = "25%"),
                                           actionButton("do_fold", "Start Cross-Validation"),
                                           tags$hr(),
                                           selectInput('label_select_results', 'Filter Results for Certain Code Levels', "", width = 240),
                                           uiOutput('kfold_results_UI'),
                                           # tags$hr(),
                                           # actionButton("do_modal_eval_defs", "Show Definitions")
                                           #tags$div(actionButton("do_modal_eval_defs", "Show Definitions", class = "decent_button"), style = "text-align:right;")),

                                    ),
                                    column(6,
                                           tags$h4("Results by Folds", id = "cv_res_h4", style = "margin-top:66px;"),
                                           #tags$p("Inside the table you can see the results of the individual folds. You can see the aggregated values on the left side.", style = "font-size:0.9em;"),
                                           tags$hr(id = "cv_res_hr"),
                                           # checkboxInput('add_expAcc', 'Add Expected eccuracy', value = FALSE, width = NULL),
                                           # checkboxInput('add_obsAcc', 'Add Observed Accuracy', value = FALSE, width = NULL),
                                           # checkboxInput('add_AUC', 'Add AUC', value = FALSE, width = NULL),

                                           # plotlyOutput('kfold_plot'),
                                           DT::dataTableOutput('kfold_table'),
                                           # br(),
                                           # br(),
                                           # DT::dataTableOutput('kfold_results')
                                           # actionButton('do_modal_evaluation', 'Show Results'),
                                           #downloadButton('download_preprocess', 'Download')
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_7', class = "btn-navig", HTML("&laquo;")),
                                           actionButton('next_RA_7', class = "btn-navig", HTML("&raquo;"))
                                    )

                           ),

                           # Panel #8 Classifier ----
                           tabPanel("8 Classifier", value = "tabp_RA_8",
                                    column(5,
                                           titlePanel("Classifier"),
                                           tags$h4("Explore the Classifier"),
                                           textAreaInput('text_classifier', label = NULL, placeholder = "Type your response here ...", width = "100%", height = "3.5em"),
                                           actionButton('do_classify', 'Classify Response'),
                                           tags$br(),
                                           tags$hr(id = "classifier_res_hr1"),
                                           uiOutput('resultClass'),

                                           tags$br(),
                                           uiOutput('original_text'),

                                           uiOutput('processed_text'),
                                           tags$br(),
                                           uiOutput('cluster_info'),
                                           uiOutput('similar_sents'),
                                           tags$hr(id = "classifier_res_hr2"),
                                           actionButton('do_modal_classifier_answers', 'Show Cluster Details')

                                    ),
                                    column(6,
                                           h4("Low-Dimensional Representation of the Response Semantics", style = "margin-top:66px;"),
                                           p("For visualization purposes, the dimensions of the semantic space are reduced here.", style = "font-size:0.9em;"),
                                           hr(style="margin-bottom:-5px;"),
                                           div(selectInput("dimensions_respvectors_classifier", label = "", width = 220,
                                                       choices = c('No Figure' = 1,
                                                                   'Display Interactive 2D Figure' = 2,
                                                                   'Display Interactive 3D Figure' = 3),
                                                       selected = 3), style = "float:left;"),
                                           div(checkboxInput("show_cluster", label = "Highlight Cluster", value = F), style = "float:left;margin-left:15px;margin-top:21px;"),
                                           div(selectInput("dimred_technique_classifier", label = "",
                                                           choices = c("PCA" = "pca",
                                                                       "t-SNE" = "tsne"),
                                                           width = 80,
                                                           selected = "tsne"), style = "float:right;margin-left:10px;margin-right:10px;"),
                                           div(sliderInput("tsne_sat_classifier",
                                                         label = div(style="width:200px;margin-bottom:-15px;",
                                                                     div(style="float:left;margin-bottom:-15px;font-size:.75em;", "accurate"),
                                                                     div(style="float:right;margin-bottom:-15px;font-size:.75em;", "fast")
                                                         ),
                                                         min = 0, max = 1, step = .1, value = .8, width = 200, ticks = F),
                                             style = "float:right;"
                                           ),
                                           div(withSpinner(plotlyOutput('classifier_plotly'), color = '#006699'), style = "clear:both;")
                                    ),
                                    column(1,
                                           br(),
                                           actionButton('prev_RA_8', class = "btn-navig", HTML("&laquo;"))
                                    )
                           )
               )
             )
    ),

    # IV) INFO ----
    tabPanel("INFO", value = "tabp_INFO",
             column(3,
                    ),
             column(6,
                    tags$div(HTML("<a href=\"#\" onclick=\"customHref('RESPONSE ANALYSIS');\"><img src=\"www/corner-left-up.svg\" style=\"height:1em;margin-right:5px;margin-bottom:3px;\" alt=\"up arrow\">Back to Response Analysis</a>"), style="font-size:0.8em;float:right;margin-top:33px;"),
                    titlePanel("Information on ReCo"),
                    tags$h3("Tutorial on How To"),
                    tags$p("ReCo's methodology adopts techniques from different disciplines and can get complex in places. Therefore, together with the release, we have written a paper that illuminates the methodological background and can serve as a tutorial, that leads you through the process step by step:"),
                    tags$hr(),
                    tags$p(HTML("Andersen, N., & Zehner, F. (2021). shinyReCoR: A Shiny Application for Automatically Coding Text Responses Using R. <span style=\"font-style:italic;\">Psych</span>, <span style=\"font-style:italic;\">3</span>(3), 422-446. <a href=\"https://doi.org/10.3390/psych3030030\" target=\"_blank\">doi: 10.3390/psych3030030</a> <span class=\"decent\">(Open Access)</span>"), style = "text-align:center;"),
                    tags$hr(),
                    tags$p("In addition, each step of the RESPONSE ANALYSIS is accompanied by a small informative text on the left that should give you a rough idea about what a particular step in the procedure is about."),
                    tags$h3("Demo Data Set"),
                    tags$p(HTML("So, what is ReCo about and how can we demonstrate its capabilities? ReCo has been designed for automatically scoring text responses from the <a href=\"https://www.oecd.org/pisa/\" target=\"_blank\">PISA studies</a>. There, students have to write responses to some questions (e.g., asking about the gist of a text). By the way, ReCo's scope goes beyond automatic scoring or coding, but this is a different story we will cover another time.")),
                    tags$p(HTML("The shinyReCoR package comes with a demo data set (in German as well as English). This is nice and handy, but it also has a flaw. Text responses are hard to simulate and the data we are typically working with are strictly confidential. Therefore, we used real empirical data from the PISA studies (for showing you the diversity of real-world data) and semi-automatically transformed the responses into responses for a different question by replacing single words (but retaining concept relations, typos, etc.). The disadvantage here is that this is only possible for rather simple questions with simple responses. While the demo data thus only contains short responses with rather low language diversity and complexity, we could already show that the ReCo method works with more complex responses pretty well (<a href=\"https://doi.org/10.1177/0013164415590022\" target=\"_blank\">Zehner, Sälzer, & Goldhammer, 2016</a>; <a href=\"https://www.pedocs.de/volltexte/2018/14979/pdf/0013164415590022_A.pdf\" target=\"_blank\" class=\"decent\">[Full Text]</a>).")),
                    tags$p("So, finally, how does the test question look like for which the responses in the demo data are formulated? It is a prototypical item for the PISA reading question, but is not part of the PISA studies:"),
                    tags$img(src = "www/SampleItem.PNG", style = "width:100%;"),
                    tags$h3("Contact"),
                    tags$p(HTML("Finally, should you be interested in contacting us, we are happy to get in contact:<br/><div style=\"text-align:center;font-weight:bold;font-size:1.3em;\"><a href=\"mailto:info@reco.science\">info@reco.science</a></div><div style=\"margin-top:15px;\">You can also find more information on the team behind ReCo at</div><div style=\"text-align:center;font-weight:bold;font-size:1.3em;\"><a href=\"https://www.reco.science\ target=\"_blank\">www.reco.science</a></div>"))
                    ),
             column(3,
                    )
             ),

    # V) HOW TO CITE ----
    tabPanel("HOW TO CITE",
             column(3),
             column(6,
                    titlePanel("How to Cite"),
                    tags$div("If you use ReCo for your scientific work, we would be happy if you cited either the ReCo method or the shiny app."),
                    tags$h3(HTML("Shiny App <span style=\"font-style:italic;\">shinyReCoR</span>"), style = "text-align:center;"),
                    tags$p(HTML("Andersen, N., & Zehner, F. (2021). shinyReCoR: A Shiny Application for Automatically Coding Text Responses Using R. <span style=\"font-style:italic;\">Psych</span>, <span style=\"font-style:italic;\">3</span>(3), 422--446. <a href=\"https://doi.org/10.3390/psych3030030\" target=\"_blank\">doi: 10.3390/psych3030030</a> <span class=\"decent\">(Open Access)</span>"), style = "text-align:center;font-size:1.1em;"),
                    tags$h3("Method", style = "text-align:center;"),
                    tags$p(HTML("Zehner, F., Sälzer, C. & Goldhammer, F. (2016). Automatic coding of short text responses via clustering in educational assessment. <span style=\"font-style:italic;\">Educational and Psychological Measurement</span>, <span style=\"font-style:italic;\">76</span>(2), 280--303. <a href=\"https://doi.org/10.1177/0013164415590022\" target=\"_blank\">doi:10.1177/0013164415590022</a> <a href=\"https://www.pedocs.de/volltexte/2018/14979/pdf/0013164415590022_A.pdf\" target=\"_blank\" class=\"decent\">[Full Text]</a>"), style = "text-align:center;font-size:1.1em;")
                    ),
             column(3)
             )
  )
)

# *** SERVER **** ----
server <- function(input, output, session) {
  # $ PROJECT SELECTION ----
  updateSelectInput(session = session, inputId = "current_project_sel", choices = projects, selected = current_project) # Set up Project Drop Down at Start Up
  observeEvent(input$current_project_sel, { # When user switches to another project
    if(nchar(input$current_project_sel)==0)
      return()

    # Update Project Name
    v$current_project <- input$current_project_sel

    # Update FileImport
    reset("importfile")

    # Reset Object v (except current_project and projects-list)
    # 0 SETUP
    v$prebuilt_vsms <- data.frame()
    # 1 Data
    v$data <- data.frame()
    v$import_filename <- F
    v$codes <- data.frame()
    # 2 Preprocessing/Predictiveness
    v$show_pp_examples <- list()
    v$classifier_input <- ''

    v$pp_spelling_correction <- F
    v$pp_freq <- data.frame()
    v$dictionary <- NULL
    v$pp_decapitalize <- T
    v$umlaut_normalization <- T
    v$pp_del_numbers <- T
    v$pp_del_symbols <- T
    v$pp_del_stopwords <- T
    v$pp_stemming <- T
    v$stopwords <- c()
    v$preprocessing_status <- "p"
    # 3 Text Corpus
    v$selected_vsm <- c("file"="", "name"="")
    v$vsm <- data.frame()
    # 4 Semantic Space
    # 5 Response Semantics
    # 6 Response Types
    v$clust_info <- data.frame()
    v$respvector_status <- "p"
    v$respclusters_status <- "p"
    v$cluster_status <- "p"
    v$resp_vectors <- data.frame()
    v$resp_clusters <- data.frame()
    v$clustnum = 100
    v$clustering = 'hclust'
    v$distance = 'cosine'
    # 7 Evaluation
    v$eval_res <- data.frame()
    # 8 Classifier
    v$classifier_input <- ''


    # Update Project Configuration
    update_proj_conf(v$current_project)

    # Load Associated Object (From File System)
    load_project_files()

    # Remove Recent Flag From All other Projects and Flag Current Project Instead
    unlink("projects/*/config/recent_project")
    set_confvar("recent_project", "T", v$current_project)

    # Update Preprocessing Selection
    updateCheckboxInput(session = session, inputId = "decapitalize", value = ifelse(is.na(as.logical(proj_conf["pp_decapitalize"])), T, proj_conf["pp_decapitalize"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "umlaut_normalization", value = ifelse(as.logical(is.na(proj_conf["umlaut_normalization"])), T, proj_conf["umlaut_normalization"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "del_numbers", value = ifelse(as.logical(is.na(proj_conf["pp_del_numbers"])), T, proj_conf["pp_del_numbers"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "del_symbols", value = ifelse(as.logical(is.na(proj_conf["pp_del_symbols"])), T, proj_conf["pp_del_symbols"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "spelling_correction", value = ifelse(as.logical(is.na(proj_conf["pp_spelling_correction"]) || (nchar(lang_codes[[proj_conf["lang"]]]$spellchecker)==0)), F, proj_conf["pp_spelling_correction"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "del_stopwords", value = ifelse(as.logical(is.na(proj_conf["pp_del_stopwords"])), T, proj_conf["pp_del_stopwords"]=="TRUE"))
    updateCheckboxInput(session = session, inputId = "stemming", value = ifelse(as.logical(is.na(proj_conf["pp_stemming"])), T, proj_conf["pp_stemming"]=="TRUE"))

    # Update Various Interface Objects
    updateSelectInput(session = session, inputId = "select_corpus_space", selected = proj_conf["corpus"])
    updateTextInput(session = session, inputId = "text_classifier", value = "")
    updateTextInput(session = session, inputId = "word1", value = "")
    updateTextInput(session = session, inputId = "word2", value = "")

    # Update Project Settings Language
    updateSelectInput(session = session, inputId = "change_project_language", selected = proj_conf["lang"])

    # Hide Interface Objects that are Only Uncovered Later On
    hide("label_select_results")
    hide("cv_res_h4")
    hide("cv_res_hr")
    hide("do_modal_classifier_answers")
    hide("classifier_res_hr1")
    hide("classifier_res_hr2")

    # Disable/Enable Interface Objects that Should Not be Available for Certain Languages or Back
    if(nchar(lang_codes[[proj_conf["lang"]]]$spellchecker)==0) {
      shinyjs::disable(id = "spelling_correction")
    } else {
      shinyjs::enable(id = "spelling_correction")
    }
  })

  # $ REACTIVE VALUES ----
  v <- reactiveValues(# ATTENTION! Make sure that changes in object v are reflected in the project selection update! (see above)
                      "projects" = projects,
                      "data" = data.frame(),
                      "vsm" = data.frame(),
                      "resp_vectors" = data.frame(),
                      "resp_clusters" = data.frame(),
                      "current_project" = current_project,
                      "preprocessing_status" = "p", # 'a' for active, 'd' for done, 'p' for pending
                      "import_filename" = F,
                      "respvector_status" = "p", # 'a' for active, 'd' for done, 'p' for pending
                      "respclusters_status" = "p", # 'a' for active, 'd' for done, 'p' for pending
                      "clust_info" = data.frame(),
                      "cluster_status" = "p",
                      "eval_res" = data.frame(),
                      "stopwords" = c(),
                      "codes" = data.frame(),
                      "show_pp_examples" = list(),
                      "pp_spelling_correction" = F,
                      "selected_vsm" = c("name"="", "file"=""),
                      "dictionary" = NULL,
                      "pp_freq" = data.frame(),
                      "pp_decapitalize" = T,
                      "umlaut_normalization" = T,
                      "pp_del_numbers" = T,
                      "pp_del_symbols" = T,
                      "pp_del_stopwords" = T,
                      "pp_stemming" = T,
                      "classifier_input" = "",
                      "prebuilt_vsms" = data.frame()
                      )

  # $ LOAD STORED RESULTS (if available) & CHECK REQUIREMENTS ----
  observeEvent(input$tabs_RA, {
    if(nchar(v$current_project)==0)
      return()

    # REQUIREMENT CHECKS
    check_workflow_reqs()
  })

   # Load Project Files ----
  load_project_files <- function() {
    # Show Loading Status
    disp_loading_status <- F
    if(!is.na(proj_conf["import_filename"]) && nchar(proj_conf["import_filename"])>0) {
      disp_loading_status <- T
      showModal(
        modalDialog(
          h3(HTML("Loading Project"), style = "color:#006699;margin-top:5px;"),
          div("", id = "load_disp", style = "color:#000000;font-size:0.8em;"),
          size = "s", footer = NULL
        )
      )
    }

    # Load Import Filename
    if(!is.null(proj_conf["import_filename"]) && !is.na(proj_conf["import_filename"]) && nchar(proj_conf["import_filename"])>0 && (isFALSE(v$import_filename) || v$import_filename!=proj_conf["import_filename"])) {
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading import file</div>"))
      v$import_filename <- proj_conf["import_filename"]
    }

    # Load Locally Available Semantic Spaces
    html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>initializing local semantic spaces</div>"))
    v$prebuilt_vsms <- readRDS("corpus_vsm/meta_info.RDS")

    # Load Text Corpus
    if(!is.null(proj_conf["corpus"]) && !is.na(proj_conf["corpus"]) && nchar(proj_conf["corpus"])>0 && (nchar(v$selected_vsm["file"])==0 || v$selected_vsm["file"]!=proj_conf["corpus"])) {
      if(file.exists(paste0('corpus_vsm/', proj_conf["corpus"]))) {
        runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
        html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading semantic space</div>"), add = T)
        v$vsm <- readRDS(paste0('corpus_vsm/', proj_conf["corpus"]))
      } else {
        v$vsm <- data.frame()
      }
      available_vsms <- available_vsms()
      v$selected_vsm[c("file", "name")] <- c(proj_conf["corpus"], available_vsms$Name[available_vsms$File.Name==proj_conf["corpus"]])
    }

    # Load Data
    if('data.RDS' %in% list.files(paste0('projects/', v$current_project)) && nrow(v$data) == 0) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading data</div>"), add = T)
      v$data <- readRDS(paste0('projects/', v$current_project, '/data.RDS'))
    }

    # Update Column Selector
    if(is.data.frame(v$data) && nrow(v$data)>0) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>initializing column selector</div>"), add = T)
      cols <- colnames(v$data)
      if(!is.na(proj_conf["textcol_generated"])) # The "text" column in the data has been artificially generated -> omit it from selection ("It's not there in the imported data set.")
        cols <- cols[cols!="text"]
      if(!is.na(proj_conf["mancodecol_generated"])) # The "mancode" column in the data has been artificially generated -> omit it from selection
        cols <- cols[cols!="mancode"]
      updateSelectInput(session, inputId = 'text_column',
                        choices = cols, selected = ifelse(!is.na(proj_conf["text_columnname"]),
                                                          proj_conf["text_columnname"],
                                                          ifelse("text" %in% colnames(v$data),
                                                                 "text",
                                                                 ifelse(length(cols)==0,
                                                                        "",
                                                                        cols[min(which(cols != "mancode"))]
                                                                 )
                                                          )
                        )
      )
      updateSelectInput(session, inputId = 'mancode_column',
                        choices = cols, selected = ifelse(!is.na(proj_conf["mancode_columnname"]),
                                                          proj_conf["mancode_columnname"],
                                                          ifelse("mancode" %in% colnames(v$data),
                                                                 "mancode",
                                                                 ifelse(length(cols)==0,
                                                                        "",
                                                                        cols[max(which(cols != "text"))]
                                                                 )
                                                          )
                        )
      )
    }

    # Load Codes / Labels
    if('codes.RDS' %in% list.files(paste0('projects/', v$current_project)) && nrow(v$codes) == 0) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading code labels</div>"), add = T)
      v$codes <- readRDS(paste0('projects/', v$current_project, '/codes.RDS'))
    }
    update_pp_examples() # generate pp examples

    # Load Cluster Information
    if('clust_info.RDS' %in% list.files(paste0('projects/', v$current_project)) && nrow(v$clust_info) == 0) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading response types</div>"), add = T)
      v$clust_info <- readRDS(paste0('projects/', v$current_project, '/clust_info.RDS'))
    }

    # Set Preprocessing Status
    if('text_processed' %in% names(v$data)) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading preprocessing pipeline</div>"), add = T)
      v$preprocessing_status <- "d"

      v$pp_decapitalize <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_decapitalize')))
      v$umlaut_normalization <- as.logical(readLines(paste0('projects/', v$current_project, '/config/umlaut_normalization')))
      v$pp_del_numbers <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_del_numbers')))
      v$pp_del_symbols <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_del_symbols')))
      v$pp_del_stopwords <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_del_stopwords')))
      v$pp_stemming <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_stemming')))
      v$pp_spelling_correction <- as.logical(readLines(paste0('projects/', v$current_project, '/config/pp_spelling_correction')))
    }

    # Set Clustering Status
    if('cluster' %in% names(v$data)) {
      v$cluster_status <- "d"
    }

    # Load Cross-Fold Evaluation Results
    if(file.exists(paste0("projects/", v$current_project, "/kfold_cv.RDS")) && nrow(v$eval_res)==0) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading evaluation results</div>"), add = T)
      v$eval_res <- readRDS(paste0("projects/", v$current_project, "/kfold_cv.RDS"))
    }

    # Load Stopword List
    if(disp_loading_status) {
      runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
      html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>loading stop words</div>"), add = T)
    }
    v$stopwords <- readRDS(paste0("projects/", v$current_project, "/stopwords.RDS"))

 ####!!!!!!
    # Load Language Model for Spell Checker
    if(nrow(v$pp_freq) == 0) {
      if(disp_loading_status) {
        runjs("document.querySelectorAll('.active_loading')[0].classList.remove('active_loading');")
        html("load_disp", HTML("<div class='active_loading'><span class=\"loading_prefix\">~$</span>initializing spell checker</div>"), add = T)
      }
      l <- lang_codes[[input$select_language]]$spellchecker
      if(l!="")
        v$freq <- readRDS(paste0('frequencies/', l,'.RDS'))
      else
        v$freq <- data.frame()
    }

    # Disable Code Labels Interface if There Are No Codes
    if(!is.data.frame(v$codes) || nrow(v$codes)==0) {
      shinyjs::disable(selector = '.navbar-nav a[data-value="Code Labels"')
    } else {
      shinyjs::enable(selector = '.navbar-nav a[data-value="Code Labels"')
      updateSelectInput(session = session, inputId = "label_select", choices = v$codes$mancode)
    }

    # Clean Up Code Labels Interface
    updateSelectInput(session = session, inputId = "label_select", choices = v$codes$mancode)

    # Remove Loading Dialog
    if(disp_loading_status) {
      removeModal()
    }
  }

  # Check Workflow Requirements
  check_workflow_reqs <- function() {
    tab_str <- strsplit(input$tabs_RA, "_")[[1]]
    if(length(tab_str)!=3)
      return()
    step_nr <- as.numeric(tab_str[length(tab_str)])
    if(step_nr>1 && !step_nr%in%3:4 && (!is.data.frame(v$data) || nrow(v$data)==0)) { # Data Import available at almost all steps after first
      warnDialogue("Please import data first (see Step 1).")
    } else if(step_nr>1 && !step_nr%in%3:4 && !"text" %in% names(v$data)) { # No text column identified
      warnDialogue(HTML("No text column has been identified in your data. Please go to <span style=\"font-style:italic;\">Step 1 Data</span> and make sure to import data with a column called <span style=\"font-style:italic;\">text</span> or use the column selector to tell ReCo what column contains the texts."))
    } else if(step_nr>=2.2 && !step_nr%in%3:4 && !"text_processed" %in% colnames(v$data)) { # From 2.2 on, we need preprocessed data
      warnDialogue("Please preprocess the data first (see Step 2.1).")
    } else if(step_nr>=4 && (!is.data.frame(v$vsm) || nrow(v$vsm)==0))  { # For Semantic Space and later, we need a loaded VSM
      warnDialogue("Please load a semantic space first (see Step 3).")
    } else if(step_nr>=6 && !'vsm_d1' %in% names(v$data)) { # For Computing Response Types, we need response vectors
      warnDialogue("Please compute the semantic response vectors first (see Step 5).")
    } else if(step_nr==7 && !"cluster" %in% names(v$data)) { # For carrying out CV Evaluation and using the classifier, we need the cluster model
      warnDialogue("Please identify response types first (see Step 6).")
    }
  }

  # II) SETUP ----
  # Switch for Enabling and Disabling Parallel Processing
  observeEvent({ input$do_parallel
                  input$parallel_cores
                  pi}, {
    set_confvar("do_pp", ifelse(input$do_parallel, "T", "F"), v$current_project)
    set_confvar("pp_cores", input$parallel_cores, v$current_project)
  })

  # Check and Display Whether meta_info.RDS is up-to-date (compared to server version)
  output$prebuilt_vsms_status <- renderUI({
    req(v$prebuilt_vsms) # For updated UI after successful download

    output <- ''
    last_updated_srv <- last_updated_loc <- 0
    tryCatch({
        # Check whether meta_info.RDS is up to date from a timestamp
        last_updated_srv <- data.table::fread("https://www.reco.science/vsms/lastupdate", sep = ";", header = F, showProgress = F) %>%
          as.character %>%
          as.POSIXct(format = "%Y-%m-%d %H:%M:%S")

        last_updated_loc <- data.table::fread("corpus_vsm/lastupdate", sep = ";", header = F) %>%
          as.character %>%
          as.POSIXct(format = "%Y-%m-%d %H:%M:%S")

      },
      error = function(e) {
        safeError(e)
        output <<- e
      }
    )
    if(last_updated_srv==0) { # No connection
      output <- paste0("<div style=\"font-size:0.8em;text-align:right;font-style:italic;\" class=\"reco_warn\"><img src=\"www/alert-octagon.svg\" style=\"height:1em;margin-bottom:4px;margin-right:4px;\" alt=\"Connection Failed\" class=\"reco_warn\">Could not connect to <a href=\"https://", reco.srv.url, "\" target=\"_blank\">", reco.srv.url, "</a>. Thus, <span style=\"font-weight:bold;\">we do not know</span> whether your list of spaces is up to date. Either your internet connection failed or our server did not respond.</div>")
    } else if(last_updated_loc>=last_updated_srv) { # List is up to date
      output <- paste0("<div style=\"font-size:0.8em;text-align:right;font-style:italic;\" class=\"decent\"><img src=\"www/check-circle.svg\" style=\"height:1em;margin-bottom:4px;margin-right:4px;\" alt=\"Your local list of available semantic spaces from ", reco.srv.url, " is up to date.\" class=\"decent\">The table of available prebuilt spaces below is up-to-date.</div>")
    } else { # Update required
      output <- paste0("<div class=\"reco_warn\" style=\"text-align:center;margin-top:15px;\"><img src=\"www/alert-octagon.svg\" style=\"height:1em;margin-bottom:4px;margin-right:7px;\" alt=\"Your local list of available semantic spaces from ", reco.srv.url, " is out of date.\">Your list of prebuilt semantic spaces is out of date.", div(actionButton("update_meta_info", "Press to Update", class = "significant_button"), style = "text-align:center;margin-bottom:10px;margin-top:5px;"), "</div>")
    }

    return(HTML(output))
  })

  # Update meta_info.RDS
  observeEvent(input$update_meta_info, {
    tryCatch({
      # Disable the Download Button
      disable("update_meta_info")

      # Try to download the latest meta_info.RDS
      if(download.file(paste0("https://", reco.srv.url, "/vsms/meta_info.RDS"), destfile = "corpus_vsm/meta_info.RDS", method = "auto", mode = "wb", quiet = T)!=0) # beware that `mode="wb"` is decisive for Windows; otherwise Windows users will have \n converted to \r\n and the file will be corrupted
        warning(paste0("Could not download the requested file. What about the connection between your machine and the ReCo server? If you continue to have problems, contact us at info@reco.science or check whether you can download https://", reco.srv.url, "/vsms/meta_info.RDS manually and move it to your package folder: ", system.file("corpus_vsm/", package = pkgname)))
      # ... and also update the timestamp file
      write(paste(as.POSIXct(Sys.time()), as.POSIXlt(Sys.time())$zone), file = "corpus_vsm/lastupdate")

      # Read the New meta_info
      new_meta_info <- readRDS("corpus_vsm/meta_info.RDS")

      # ... and integrate it with the local one (download date in particular)
      new_meta_info$Download.Date <- sapply(new_meta_info$File.Name, function(fn) { x <- v$prebuilt_vsms$Download.Date[v$prebuilt_vsms$File.Name==fn]; if(length(x)==0) "" else x }) %>% as.character
      v$prebuilt_vsms <- new_meta_info

      # Store the Integrated one to Disk
      saveRDS(v$prebuilt_vsms, file = "corpus_vsm/meta_info.RDS")

      # Enable the Download Button
      enable("update_meta_info")
      }, error = function(e) {
        safeError(e)
        enable("update_meta_info") # Enable the Download Button
      }
    )
  })

  # Display Available Prebuilt Semantic Spaces
  output$prebuilt_vsms <- DT::renderDataTable({
    req(v$prebuilt_vsms)
    if(!is.data.frame(v$prebuilt_vsms) || nrow(v$prebuilt_vsms)==0)
      return(DT::datatable(data.frame()))

    # Identify Already Successfully Downloaded Files
    downloaded <- apply(v$prebuilt_vsms[,c("File.Name", "Hash")], 1,
                        function(prebvsm) {
                          if(file.exists(paste0("corpus_vsm/", prebvsm[1]))) { # VSM File Found on the Local Drive
                            if(tools::md5sum(paste0("corpus_vsm/", prebvsm[1]))==prebvsm[2]) # File's Hash is fine too
                              return(c("available", "<img src=\"www/check-circle.svg\" style=\"height:1.5em;margin-right:5px;\" alt=\"checkmark\" title=\"You have downloaded this space. It is ready to use.\" onclick=\"prevent_dialog_once++;\"><img src=\"www/trash-2.svg\" style=\"height:1.5em;cursor:pointer;\" alt=\"trash\"  title=\"Click to remove file.\">"))
                            else # File's hash doesn't look good (e.g., partial download). Should be removed.
                              return(c("corrupt", "<img src=\"www/cloud-lightning.svg\" style=\"height:1.5em;cursor:pointer;\" alt=\"corrupt\"  title=\"This file is corrupt. It is not usable. Click to clean up.\">"))
                          } else # Has not been downloaded yet
                            return(c("cand", "<img src=\"www/download.svg\" style=\"height:1.5em;cursor:pointer;\" alt=\"download\"  title=\"Click to download file.\">"))
                        }) %>% as.data.frame

    # Render Data Table
    DT::datatable(v$prebuilt_vsms %>% select(-matches("Hash|Download\\.Date")) %>% mutate("Action"=as.character(downloaded[2,]), .before = Name) %>% mutate("mode"=as.character(downloaded[1,])), filter = "top", rownames = F,
                  colnames = c("Status & Action", "Name", "Language", "Stemmed", "Lemmatized", "Method", "#Dims", "#Types", "Source", "File Name", "File Size", "Details", "Mode"),
                  escape = -1,
                  options = list(order = list(list(2, "asc"), list(1, "asc")),
                                 columnDefs = list(list(width = "12%", targets = 0, orderable = F, className = "dt-head-center dt-body-center prebuiltdownload-control"),
                                                   list(width = "15%", targets = 1),
                                                   list(width = "5%", targets = 2),
                                                   list(width = "5%", targets = 3),
                                                   list(width = "5%", targets = 4),
                                                   list(width = "10%", targets = 5),
                                                   list(width = "5%", targets = 6),
                                                   list(width = "5%", targets = 7),
                                                   list(width = "10%", targets = 8),
                                                   list(visible = F, targets = 9), # Hide File Name (but keep it in the table for JS)
                                                   list(width = "8%", targets = 10, render = JS(
                                                     "function(data, type, row, meta) {",
                                                     "var filesize = parseInt(data);",
                                                     "const units_map = ['b', 'Kb', 'Mb', 'Gb', 'Tb', 'Pb'];",
                                                     "var power = (filesize<=0) ? 0 : Math.min(Math.round(Math.log(filesize) / Math.log(1024)), units_map.length - 1);",
                                                     "var unit = (power==0) ? 'bytes' : units_map[power];",
                                                     "return Math.round((filesize/Math.pow(1024, power))*10)/10 + ' ' + unit;",
                                                     "}")),
                                                   list(width = "20%", targets = 11),
                                                   list(visible = F, targets = 12) # Hide Mode (but keep it in the table for JS)
                                            ),
                                 lengthMenu = c(10, 20, 50),
                                 dom = "iptrl"
                  ),
                  select = "none",
                  callback = JS("
                      prevent_dialog_once = 0;
                      table.on('click', 'td.prebuiltdownload-control', function() {
                        if(prevent_dialog_once>0) {
                          prevent_dialog_once--;
                        } else {
                          var td = $(this), row = table.row(td.closest('tr'));
                          Shiny.setInputValue('prebuiltvsms_download_filename', '');
                          Shiny.setInputValue('prebuiltvsms_download_filename', row.data()[9]);
                          Shiny.setInputValue('prebuiltvsms_download_mode', '');
                          Shiny.setInputValue('prebuiltvsms_download_mode', row.data()[12]);
                          Shiny.setInputValue('prebuiltvsms_download_name', '');
                          Shiny.setInputValue('prebuiltvsms_download_name', row.data()[1]);
                        }
                      })
                                  ")
                ) %>%
        DT::formatRound(columns = 7:8, digits = 0) %>%
        DT::formatStyle(columns = 2, target = "cell", color = "#006699")
  })

  # Suggest Action for Prebuilt VSMs
  observeEvent(input$prebuiltvsms_download_name, {
    if(!is.character(input$prebuiltvsms_download_name) || nchar(input$prebuiltvsms_download_name)==0)
      return()

    # Dialog's Icon
    icon <-
      if(input$prebuiltvsms_download_mode=="available") "trash-2"
      else if(input$prebuiltvsms_download_mode=="cand") "download"
      else if(input$prebuiltvsms_download_mode=="corrupt") "cloud-lightning"

    # Dialog's Question
    question <-
      if(input$prebuiltvsms_download_mode=="available") "Are you sure you want to delete this semantic space?<div style=\"font-size:0.8em;margin-bottom:10px;margin-top:5px;color:#99c1d6;\">You will no longer be able to use projects that involved this semantic space.</div>"
      else if(input$prebuiltvsms_download_mode=="cand") paste0("Are you sure you want to download this semantic space?<div style=\"font-size:0.8em;margin-bottom:10px;margin-top:5px;color:#99c1d6;\">This might take some time. Consider the displayed file size and your internet connection.</div>", ifelse(!grepl("64", Sys.getenv("R_ARCH")), paste0("<div class=\"reco_warn\" style=\"margin:10px 0px 5px 0px;font-weight:bold;\">IMPORTANT: It seems like your R is running on 32bit rather than 64bit mode. It is possible that files larger than 2GB get truncated.<br/><span style=\"font-weight:normal;font-size:0.8em;\">If this is the case (your file will be shown to be corrupt), try to download the file via your browser <a href=\"https://", reco.srv.url, "/vsms/", input$prebuiltvsms_download_filename, "\" target=\"_blank\" style=\"text-decoration:underline;\">here</a> and place it in the package folder: ", system.file("corpus_vsm/", package = pkgname), "</span></div>"), ""))
      else if(input$prebuiltvsms_download_mode=="corrupt") paste0("The file associated with this space seems corrupt.<br />Clean up and download again.<div style=\"font-size:0.8em;margin-bottom:10px;margin-top:5px;color:#99c1d6;\">The file's md5-sum is not the expected one. This might indicate an incomplete download.</div><div style=\"margin:20px 0px 5px 0px;font-weight:bold;\">IMPORTANT: If you continue to have this problem, consider to download the file manually.<br/><span style=\"font-weight:normal;font-size:0.8em;\">For this, try to download the file via your browser <a href=\"https://", reco.srv.url, "/vsms/", input$prebuiltvsms_download_filename, "\" target=\"_blank\" style=\"text-decoration:underline;\">here</a> and place it in the package folder:<br/>", system.file("corpus_vsm/", package = pkgname), "</span></div>")

    # Button ID
    button_id <-
      if(input$prebuiltvsms_download_mode%in%c("available", "corrupt")) "delete_prebuiltvsm"
      else if(input$prebuiltvsms_download_mode=="cand") "download_prebuiltvsm"

    # Button Text
    button_text <-
      if(input$prebuiltvsms_download_mode=="available") "Yes, Delete!"
      else if(input$prebuiltvsms_download_mode=="cand") "Yes, Download."
      else if(input$prebuiltvsms_download_mode=="corrupt") "Yes, Clean Up."

    # Suggest Action
    showModal(
      modalDialog(
        h3(paste0("Semantic Space `", input$prebuiltvsms_download_name, "`")),
        h5(img(src = paste0("www/", icon, ".svg"), style = "height:2em;float:left;margin-right:10px;margin-bottom:10px;"), HTML(question), style = "color:#006699;"),
        div(actionButton(button_id, button_text, class = "significant_button"), HTML("<div id=\"download_loader\" class=\"lds-ellipsis\" style=\"display:none;\"><div></div><div></div><div></div><div></div></div>"), align = "center"),
        size = "m", easyClose = T,
      )
    )
  })

  # Delete Prebuilt VSM
  observeEvent(input$delete_prebuiltvsm, {
    req(input$delete_prebuiltvsm)

    # Make Sure a File is Chosen and Exists
    if(nchar(input$prebuiltvsms_download_filename)==0 || !file.exists(paste0("corpus_vsm/", input$prebuiltvsms_download_filename)))
      return()

    tryCatch({
      if(!file.remove(paste0("corpus_vsm/", input$prebuiltvsms_download_filename))) { # Try to Remove the File on the File System
        v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- paste0("deletionfailed", Sys.time())
        removeModal()
      } else { # Download worked
        if(v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"]=="") # Update Reactive Object, but make sure the value is there to trigger an update (in case of manual downloads)
          v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- "manual"
        v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- "" # Update Reactive Object Indeed
        saveRDS(v$prebuilt_vsms, "corpus_vsm/meta_info.RDS") # Update meta_info on the File System

        # Empty Input Values
        updateTextInput(session = session, inputId = "prebuiltvsms_download_filename", value = "")
        updateTextInput(session = session, inputId = "prebuiltvsms_download_mode", value = "")
        updateTextInput(session = session, inputId = "prebuiltvsms_download_name", value = "")

        # Update Drop-Down Menu for Text Corpus / Space Selection
        corpus_vsms <- as.list(c(list.files("corpus_vsm", pattern = "_corpusspace"), paste0("#dev#/", list.files("corpus_vsm/#dev#", pattern = "_corpusspace"))))
        names(corpus_vsms) <- sapply(strsplit(unlist(corpus_vsms), "_"), function(x) x[1])
        if(any(names(corpus_vsms)=="#dev#/")) {
          corpus_vsms[[which(names(corpus_vsms)=="#dev#/")]] <- NULL
        }

        # Close Modal Dialog
        removeModal()
      }
    }, warning = function(w) {
      v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- paste0("deletionfailed", Sys.time())
      showModal(modalDialog(HTML("If you had cancelled a download before (or it did not work) and tried to clean it now, it is possible that you need to restart R (not only the app) as R might have an open connection to the file you are trying to delete. Which is why this does not work.<div class=\"reco_err\" style=\"margin-top:15px;\">tldr; Long story short, restart R, come back, and try again =) ...</div><div class=\"reco_warn\" style=\"margin-top:15px;\">R's Warning Message:</div><div>", w$message, "</div>"), title = HTML("<span class=\"reco_err\">Error: Cannot Remove File</span>"), size = "m", easyClose = T))
    }, error = function(e) {
      safeError(e)
      removeModal()
      updateTextInput(session = session, inputId = "prebuiltvsms_download_filename", value = "")
      updateTextInput(session = session, inputId = "prebuiltvsms_download_mode", value = "")
      updateTextInput(session = session, inputId = "prebuiltvsms_download_name", value = "")
    })
  })

  # Download Prebuilt VSM
  observeEvent(input$download_prebuiltvsm, {
    req(input$download_prebuiltvsm)

    # Make Sure a File is Chosen and Exists
    if(nchar(input$prebuiltvsms_download_filename)==0 || !input$prebuiltvsms_download_filename%in%v$prebuilt_vsms$File.Name)
      return()

    tryCatch({
      # Disable Download Button & Show Loader
      disable("download_prebuiltvsm")
      show("download_loader")

      # Try to Download File
      if(withProgress(download.file(paste0("https://", reco.srv.url, "/vsms/", input$prebuiltvsms_download_filename), destfile = paste0("corpus_vsm/", input$prebuiltvsms_download_filename), method = "auto", mode = "wb", quiet = T), value = 0, message = paste0("Downloading ", input$prebuiltvsms_download_name), detail = HTML("You cannot perform other actions meanwhile. Progress is not displayed."))!=0) {
        v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- paste0("downloadfailed", Sys.time())
        warning(paste0("Could not download the requested file. What about the connection between your machine and the ReCo server? If you continue to have problems, contact us at info@reco.science or check whether you can download https://", reco.srv.url, "/vsms/", input$prebuiltvsms_download_filename, " manually and move it to your package folder: ", system.file("corpus_vsm/", package = pkgname)))
      } else { # Download should have been successful
        v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- paste(as.POSIXct(Sys.time()), as.POSIXlt(Sys.time())$zone) # Update Reactive Object Indeed
        saveRDS(v$prebuilt_vsms, "corpus_vsm/meta_info.RDS") # Update meta_info on the File System

        # Empty Input Values
        updateTextInput(session = session, inputId = "prebuiltvsms_download_filename", value = "")
        updateTextInput(session = session, inputId = "prebuiltvsms_download_mode", value = "")
        updateTextInput(session = session, inputId = "prebuiltvsms_download_name", value = "")

        # Update Drop-Down Menu for Text Corpus / Space Selection
        corpus_vsms <- as.list(c(list.files("corpus_vsm", pattern = "_corpusspace"), paste0("#dev#/", list.files("corpus_vsm/#dev#", pattern = "_corpusspace"))))
        names(corpus_vsms) <- sapply(strsplit(unlist(corpus_vsms), "_"), function(x) x[1])
        if(any(names(corpus_vsms)=="#dev#/")) {
          corpus_vsms[[which(names(corpus_vsms)=="#dev#/")]] <- NULL
        }

        # Re-enable Download Button & Hide Loader
        enable("download_prebuiltvsm")
        hide("download_loader")

        # Close Modal Dialog
        removeModal()
      }
    }, warning = function(w) {
      v$prebuilt_vsms[which(v$prebuilt_vsms$File.Name==input$prebuiltvsms_download_filename),"Download.Date"] <- paste0("downloadfailed", Sys.time())
      showModal(modalDialog(HTML(paste0("Could not download the requested file. What about the connection between your machine and the <a href=\"https://", reco.srv.url, "\" target=\"_blank\">ReCo Server</a>? If you can reach other websites but not ReCo's, our server might be temporarily down.<br/><br/>If you continue to have problems, contact us at <a href=\"mailto:info@reco.science\">info@reco.science</a> or check whether you can download <a href=\"https://", reco.srv.url, "/vsms/", input$prebuiltvsms_download_filename, "\" target=\"_blank\">", input$prebuiltvsms_download_filename, "</a> manually and move it to shinyReCoR's package folder: <span class=\"font-family:Courier New;monospace;\">", system.file("corpus_vsm/", package = pkgname), "</span><div style=\"margin-top:15px;\"><div class=\"reco_warn\">R's Warning Message:</div>", w$message, "</div>")), title = HTML("<span class=\"reco_warn\">Warning: Download Failed (Internet Connection?)</span>"), size = "m", easyClose = T))
    }, error = function(e) {
      safeError(e)
      # Re-enable Download Button & Hide Loader
      enable("download_prebuiltvsm")
      hide("download_loader")
      # removeModal()
      showModal(modalDialog(paste0("There was an error: ", e), title = "Error", easyClose = T, size = "s"))
    }
    )
  })

  # II) PROJECT MANAGEMENT ----
  # CV Evaluation
  observe({
    updateSelectInput(session, inputId = 'label_select_results',
                      choices = c('Total', v$codes$mancode), selected = 'Total')
  })

  # create New Projects
  observeEvent(input$do_createProject, {
    if(projectname_msg()=="") {
      # Create Folders
      dir.create(paste0('projects/', input$project_name))
      dir.create(paste0('projects/', input$project_name, '/config/'))

      # Create Language Config
      write(input$select_language, file = paste0('projects/', input$project_name, '/config/lang'))

      # Append Project List
      v$projects[[paste0("Project `", input$project_name, "`")]] <- input$project_name

      # Set New Project as Current Project
      v$current_project <- input$project_name
      set_confvar("remove_app_popup", 1, v$current_project)

      # Update Project Drop Down List
      updateSelectInput(session = session, inputId = "current_project_sel", choices = v$projects, selected = v$current_project)

      # Save Stopword List
      saveRDS(stopwords::stopwords(lang_codes[[input$select_language]]$stopwords),
              file = paste0('projects/', v$current_project, '/stopwords.RDS'))

      # Empty Project Name Text Field
      updateTextInput(session = session, inputId = "project_name", value = "")

      # Disable Code Labels--Interface (no data yet)
      shinyjs::disable(selector = '.navbar-nav a[data-value="Code Labels"')

      # Flag New Project
      proj_conf["new"] <<- "TRUE"
    }
  })

  # Produce Warning Messages for Project Name
  projectname_msg <- eventReactive(input$do_createProject, {
    # reset active tab to upload section
    updateTabsetPanel(session, "tabs_RA", selected = "tabp_RA_1")
    # check insert name
    if(nchar(input$project_name)==0) {
      msg <- "Please enter a project name."
    } else if(input$project_name %in% v$projects) {
      msg <- "A project with this name already exists."
    } else if(!grepl("^[a-z0-9_\\-]+$", input$project_name, ignore.case = T)) {
      msg <- "The name can only include characters A-Z, digits, - and _."
    } else {
      msg <- ""
    }
    msg
  })

  # Output Warning Messages for Project Name
  output$projectnameHint <- renderUI({
    HTML(projectname_msg())
  })

  output$newProjectText <- renderUI({
    if(projectname_msg()=="") {
      HTML(paste0("Project <span style=\"font-style:italic;\">", v$current_project , "</span> successfully created.<br />Click <a href=\"#\" onclick=\"customHref('RESPONSE ANALYSIS');\">here</a> to start your analysis."))
    }
  })


  # Project Settings
  observeEvent(input$do_changeProjectSettings, {
    if(changeprojectname_msg()=="") {
      # Move Project to New Folder Name
      copyDir(paste0("projects/", v$current_project), paste0("projects/", input$change_project_name))
      unlink(paste0("projects/", v$current_project), recursive = T)

      # Update Projects Object
      v$projects[[ifelse(v$current_project=="unnamed", "Unnamed Project", paste0("Project `", v$current_project, "`"))]] <- input$change_project_name
      names(v$projects)[unlist(v$projects)==input$change_project_name] <- if(input$change_project_name=="unnamed") { "Unnamed Project" } else { paste0("Project `", input$change_project_name, "`")}

      # Update Project Name in Current Project Variable
      v$current_project <- input$change_project_name

      # Update Project Drop Down List
      updateSelectInput(session = session, inputId = "current_project_sel", choices = v$projects, selected = v$current_project)
    }
  })
  output$change_project_name_ui <- renderUI({
    tagAppendAttributes(
      textInput("change_project_name", h4("Project Name"), value = v$current_project),
      `data-proxy-click` = "do_changeProjectSettings"
    )
  })
  # Produce Warning Messages for Project Name
  changeprojectname_msg <- eventReactive(input$do_changeProjectSettings, {
    if(nchar(input$change_project_name)==0) {
      msg <- "Please enter a project name."
    } else if(input$change_project_name %in% v$projects) {
      msg <- "A project with this name already exists."
    } else if(!grepl("^[a-z0-9_\\-]+$", input$change_project_name, ignore.case = T)) {
      msg <- "The name can only include characters A-Z, digits, - and _."
    } else {
      msg <- ""
    }
    msg
  })

  output$changeprojectnameHint <- renderUI({
    HTML(changeprojectname_msg())
  })

  output$changeProjectText <- renderUI({
    if(changeprojectname_msg()=="") {
      HTML(paste0("Project settings updated.<br />Click <a href=\"#\" onclick=\"customHref('RESPONSE ANALYSIS');\">here</a> to continue your analysis."))
    }
  })

  # Delete Projects
  output$delete_project_select <- renderUI({
    selectInput("del_project", label = "",
                choices = v$projects,
                selected = v$current_project)
  })

  observeEvent(input$delete_project,{
    unlink(paste0('projects/', input$del_project), recursive = T) # Remove folder and files
    if(input$del_project!="unnamed") {
      v$projects[[paste0("Project `", input$del_project, "`")]] <- NULL # Remove project from list object
    } else {
      v$projects[["Unnamed Project"]] <- NULL
    }

    if(length(v$projects)>=1) { # At least one project left
      if(v$current_project==input$del_project) # Change current_project to first project in the list if necessary
        v$current_project <- v$projects[[1]]
    } else { # No project left -> create a new unnamed one
      v$projects <- list("Unnamed Project"="unnamed")
      v$current_project <- "unnamed"

      dir.create(paste0("projects/", v$current_project))
      dir.create(paste0("projects/", v$current_project, "/config"))
      write(languages[[1]], file = paste0("projects/", v$current_project, "/config/lang"))
      write("", file = paste0("projects/", v$current_project, "/config/recent_project"))
      saveRDS(stopwords::stopwords(lang_codes[[1]]$stopwords), file = paste0('projects/', v$current_project, '/stopwords.RDS'))
    }

    # Update Project Drop Down List
    updateSelectInput(session = session, inputId = "current_project_sel", choices = v$projects, selected = v$current_project)
  })

  output$delFolderText <- eventReactive(input$delete_project,{
    HTML(paste0('Project <span class="font-style:italic;">', input$del_project, '</span> deleted.'))
  })

  # Code Labels ---
  # Put the currently selected label into the respective text field
  observeEvent(input$label_select, {
    req(input$label_select)
    labeli <- which(as.numeric(input$label_select)==v$codes$mancode)
    colourpicker::updateColourInput(session = session, inputId = "label_color", value = v$codes$colors[labeli]) # Change color value in colorpicker
    updateTextInput(session = session, inputId = "label_rename", value = v$codes$names[labeli]) # Change label name in respective text field
  })

  # Reset Code Labels
  observeEvent(input$label_reset, {
    # First, reset the colors
    mcs <-  sort(unique(v$data$mancode))
    add_c <- ifelse(length(mcs)-length(color_scheme) > 0, length(mcs)-length(color_scheme), 0)
    v$codes$colors <- c(color_scheme, rep(color_scheme[length(color_scheme)],add_c))[1:length(mcs)]
    v$codes <- v$codes[order(mcs, decreasing = F),]

    # Next, reset the labels
    v$codes$names <- ""

    # Update the interface
    labeli <- which(as.numeric(input$label_select)==v$codes$mancode)
    colourpicker::updateColourInput(session = session, inputId = "label_color", value = v$codes$colors[labeli])
    updateTextInput(session = session, inputId = "label_rename", value = v$codes$names[labeli])

    # Update the preprocessing examples
    update_pp_examples()

    # And save the reset version
    saveRDS(v$codes, file = paste0('projects/', v$current_project, "/codes.RDS"))
  })

  # Change Label Name
  observeEvent(input$label_rename, {
    req(input$label_select)

    labeli <- which(input$label_select == v$codes$mancode)
    if(v$codes$names[labeli]==input$label_rename) { # Don't do anything if value has not changed
      return()
    }

    if(grepl("<|>", input$label_rename)) { # Bad idea: possibly HTML in label name
      updateTextInput(session = session, inputId = "label_rename", value = gsub(">", "&gt;", gsub("<", "&lt;", input$label_rename)))
    }

    v$codes$names[labeli] <- input$label_rename
    if(anyDuplicated(v$codes$names) && nchar(v$codes$names[anyDuplicated(v$codes$names)])>0) { # Hint that labels have been chosen twice
      warnDialogue("Please note that you have chosen the same label for at least two different codes.")
    }
    if(any(v$codes$names %in% v$codes$mancode)) { # Hint that some label(s) equal(s) a code
      warnDialogue("Please note that at least one label equals a code value, which might be quite confusing for your interpretation.")
    }

    update_pp_examples()
    saveRDS(v$codes, file = paste0('projects/', v$current_project, "/codes.RDS"))
  })

  # Change Label Color
  observeEvent(input$label_color, {
    req(input$label_select)

    labeli <- which(input$label_select == v$codes$mancode)
    if(v$codes$colors[labeli]==input$label_color) { # Don't do anything if value has not changed
      return()
    }

    if(input$label_color %in% v$codes$colors) { # Color has already been chosen for another code
      warnDialogue("The selected color has already been chosen for another code. Please select another color.")
      colourpicker::updateColourInput(session = session, inputId = "label_color", value = v$codes$colors[labeli]) # Change to previous color
    }
    if(input$label_color=="#FFFFFF") { # Chose white
      warnDialogue("You have chosen white as the code's color, which might not be the best of all ideas.")
    }

    # Change value in data frame and on disk
    v$codes$colors[labeli] <- input$label_color
    saveRDS(v$codes, file = paste0('projects/', v$current_project, "/codes.RDS"))
  })

  output$missing_codes <- renderUI({
    if(any(is.na(v$data$mancode))) {
      text <- paste0('Your data contains ', sum(is.na(v$data$mancode)), ' ', '(', round(sum(is.na(v$data$mancode))/nrow(data)*100,2), '%) ', 'responses without codes!')
    } else {
      text <- ''
    }
    HTML(text)
  })
  # Plot for Indicating Plot Look & Feel Due To Label & Color Selection
  output$class_distribution <- renderPlotly({
    if(input$label_plot == 'pie_chart') {
      plot_ly(type='pie',
              labels = ifelse(v$codes$names == '', v$codes$mancode, v$codes$names),
              values = unlist(sapply(v$codes$mancode, function(x) sum(na.omit(v$data$mancode) == x))),
              title = 'Code Distribution',
              textposition = 'outside',
              textinfo='label+percent', insidetextorientation='radial',
              marker = list(colors =  v$codes$colors,
                            line = list(color = '#FFFFFF', width = 2)),
      )

    } else if(input$label_plot == 'random_scatter'){
      if(nrow(v$data) > 1500) {
        distr <- round(table(v$data$mancode)/nrow(v$data),2)*1500
      } else {
        distr <- table(v$data$mancode)
      }
        x = runif(sum(distr), 0, 1)
        y = runif(sum(distr), 0, 1)
        color= as.factor(unlist(sapply(1:length(distr), function(i) rep(names(distr)[i], distr[i]))))

        ax <- list(
          title = "",
          zeroline = FALSE,
          showline = FALSE,
          showticklabels = FALSE,
          showgrid = FALSE
        )

        color <- unlist(sapply(1:length(distr), function(i) rep(names(distr)[i], distr[i])))
        color <- as.factor(sapply(color, function(x) {
          label <- v$codes$names[which(v$codes$mancode == x)]
          ifelse(label != '', label, x)
        }))
      # create random dots
      plot_ly(x=runif(sum(distr), 0, 1),
              y= runif(sum(distr), 0, 1),
              color = color,
              colors = colorRamp(v$codes$colors),
              opacity = .85) %>%
        layout(xaxis = ax, yaxis = ax)
    }
  })

  # Code Label Information
  output$class_information <- renderUI({
    req(input$label_select)

    infos <- list()
    # extract information
    ids <- which(v$data$mancode == input$label_select)
    infos[['Manual Code']] <- input$label_select
    infos[['Label']] <- v$codes$names[which(v$codes$mancode == input$label_select)]
    infos[['Color']] <- paste0('<div id="label_color_disp" style="padding-left:5px;color:#FFFFFF;background-color:', v$codes$colors[which(v$codes$mancode == input$label_select)], '">', v$codes$colors[which(v$codes$mancode == input$label_select)], '</div>')

    infos[['n']] <- paste0(length(ids), '<span class="decent" style="font-size:0.8em;margin-left:5px;">(', round(length(ids)/nrow(v$data)*100, 1), '%)</span>')
    missings <- sum(is.na(v$data$text))
    miss_to_label <- round(missings/length(ids)*100, 1)
    miss_to_all <- round(missings/nrow(v$data)*100, 1)
    infos[['Missings']] <- paste0(missings, '<span class="decent" style="font-size:0.8em;margin-left:5px;">(', miss_to_label, '% of code, ', miss_to_all, '% of all)</span>')

    texts <- as.character(na.omit(v$data$text[ids]))
    # texts <- lapply(texts, function(x) preprocess(x,
                                                  # del_stopwords = F,
                                                  # stemming = F,
                                                  # lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming,
                                                  # del_symbols = T))
    texts <- lapply(texts, function(x) strsplit(x, ' ')[[1]])
    infos[['Average Text Length']] <- paste0(round(mean(unlist(lapply(texts, function(x) length(x)))), 1), ' words per response')
    infos[['']] <- paste0(round(mean(unlist(sapply(unlist(texts), function(x) nchar(x)))), 1), ' characters per word')

    # print information
    text <- paste0(sapply(1:length(infos), function(x) paste0('<tr><td><span class=\"highlighted\" style=\"font-style:italic; font-size:1.3em\">',
                                                                names(infos)[x], ifelse(nchar(names(infos)[x])>0, ':', ''), '   &nbsp',
                                                                '</span></td>',
                                                                '<td><span class=\"highlighted\" style=\"color:#000000; font-size:1.3em\">',
                                                                ifelse(as.character(infos[x])=="", "-", as.character(infos[x])),
                                                                '</span></td></tr>'
    )), collapse = ''
    )
    text <- paste0('<table id="Descriptive Information">', text, '</table>')
    texts <- as.character(na.omit(v$data$text[ids]))
    values <- information_value(texts)
    best_text <- texts[which(values == max(values))[1]]
    worst_text <- texts[which(values == min(values))[1]]
    best_text <- paste0('<br><br><span class=\"highlighted\" style=\"font-style:italic; font-size:1.1em\">', 'Most Informative Response', '</span><br>',
                        '<span class=\"highlighted\" style=\"color:#000000; font-size:1.1em\">', best_text,'</span><br><br>')

    worst_text <- paste0('<br><span class=\"highlighted\" style=\"font-style:italic; font-size:1.1em\">', 'Least Informative Response', '</span><br>',
                        '<span class=\"highlighted\" style=\"color:#000000; font-size:1.1em\">', worst_text,'</span>')

    text <- paste0(text, best_text, worst_text)
    HTML(text)
    })

  # III) RESPONSE ANALYSIS ----
  # 1 Data Import ----
  # Function for handling data import to a given path
  import_file <- function(path, demo = F) {
    tryCatch(
      {
        t1 <- Sys.time()
        miss_chars <- sapply(strsplit(input$missing_chars, ","), trimws)
        if(!demo) { # Import user file
          if(input$encoding == 'GUESS'){
            df <- read.csv(input$importfile$datapath,
                           header = input$header,
                           sep = input$sep,
                           quote = input$quote,
                           na.strings = c("", miss_chars),
                           stringsAsFactors = F)

          } else {
            df <- read.csv(input$importfile$datapath,
                           header = input$header,
                           sep = input$sep,
                           quote = input$quote,
                           encoding = input$encoding,
                           na.strings = c("", miss_chars),
                           stringsAsFactors = F)

          }
          fname <- basename(input$importfile$name)
        } else { # Import demo data
          df <- read.csv(path,
                         header = T,
                         sep = ";",
                         quote = "\"",
                         na.strings = c("NA"),
                         stringsAsFactors = F)
          fname <- basename(path)
        }
        #print(Sys.time()-t1)

        # Delete recent project files
        # delete files in project file
        sapply(dir(paste0('projects/',v$current_project)), function(file) {
          if(!file %in% c('config', 'stopwords.RDS'))
            file.remove(paste0(paste0('projects/',v$current_project, '/', file)))
        })
        # delete files in project/config file
        sapply(dir(paste0('projects/',v$current_project, '/config')), function(file) {
          if(!file %in% c('lang', 'remove_app_popup', 'recent_project'))
            # file.remove(paste0(paste0('projects/',v$current_project, '/config/', file)))
            delete_confvar(file, v$current_project)
        })

        # Reset v
        # TODO: init v via function
        v$data <- data.frame()
        v$vsm <- data.frame()
        v$resp_vectors <- data.frame()
        v$resp_clusters <- data.frame()
        v$preprocessing_status <- "p"
        v$import_filename <- F
        v$selected_vsm <- c("file"="", "name"="")
        v$respvector_status <- "p"
        v$respclusters_status <- "p"
        v$clust_info <- data.frame()
        v$cluster_status <- "p"
        v$eval_res <- data.frame()
        v$stopwords <- readRDS(paste0("projects/", v$current_project, "/stopwords.RDS"))
        v$codes <- data.frame()
        v$show_pp_examples <- list()
        v$classifier_input <- ""

        v$pp_spelling_correction <- F
        v$pp_freq <- data.frame()
        v$dictionary <- NULL
        v$pp_decapitalize <- T
        v$umlaut_normalization <- T
        v$pp_del_numbers <- T
        v$pp_del_symbols <- T
        v$pp_del_stopwords <- T
        v$pp_stemming <- T


        # Store file name
        set_confvar("import_filename", fname, v$current_project)
        v$import_filename <- proj_conf["import_filename"]

        # Reset config variables (just in case they were set by a previous import)
        delete_confvar("text_columnname", v$current_project)
        delete_confvar("mancode_columnname", v$current_project)
        delete_confvar("textcol_generated", v$current_project)
        delete_confvar("mancodecol_generated", v$current_project)

        preselect_text <- names(df)[which(unlist(sapply(1:ncol(df), function(x) is.character(df[1:100,x]))))][1]
        preselect_mc <- names(df)[which(unlist(sapply(1:ncol(df), function(x) length(table(df[1:100,x])) < 50)))][1]
        # select the important columns (text and mancode) (select them automatically if they occur in names)
        updateSelectInput(session, inputId = 'text_column',
                          choices = names(df), selected = ifelse(!is.na(proj_conf["text_columnname"]),
                                                                 proj_conf["text_columnname"],
                                                                 ifelse("text" %in% colnames(df),
                                                                        "text",
                                                                        ifelse(length(preselect_text) != 0, preselect_text,
                                                                               names(df)[min(which(names(df) != "mancode"))])
                                                                        #names(df)[min(which(names(df) != "mancode"))]
                                                                        )
                                                                 )
                         )
        updateSelectInput(session, inputId = 'mancode_column',
                          choices = names(df), selected = ifelse(!is.na(proj_conf["mancode_columnname"]),
                                                                 proj_conf["mancode_columnname"],
                                                                 ifelse("mancode" %in% colnames(df),
                                                                        "mancode",
                                                                        ifelse(length(preselect_mc) != 0, preselect_mc,
                                                                               names(df)[min(which(names(df) != "mancode"))])
                                                                        )
                                                                        # names(df)[max(which(names(df) != "text"))])
                                                                )
                        )

        # If import was successful, handle empty responses
        if(is.data.frame(df) && all(c("mancode", "text") %in% colnames(df))) {
          empty_answers <- table(df$mancode[is.na(df$text)])
          saveRDS(empty_answers, file = paste0('projects/', v$current_project, "/empty_answers.RDS"))
          df$text[nchar(trimws(df$text))==0] <- NA # Tweak empty responses to missing responses
        } else {
          saveRDS(0, file = paste0('projects/', v$current_project, "/empty_answers.RDS"))
        }

        # Handle color codes
        mcs <-  sort(unique(df$mancode))
        add_c <- ifelse(length(mcs)-length(color_scheme) > 0, length(mcs)-length(color_scheme), 0)
        colors <- c(color_scheme, rep(color_scheme[length(color_scheme)],add_c))[1:length(mcs)]
        names <- rep('', length(mcs))
        v$codes <- data.frame(mancode = mcs, colors, names)

        saveRDS(v$codes, file = paste0('projects/', v$current_project, "/codes.RDS"))

        update_pp_examples() # Generate preprocessing examples

        # Disable Code Labels Interface if There Are No Codes and Enable Otherwise
        if(!is.data.frame(v$codes) || nrow(v$codes)==0) {
          shinyjs::disable(selector = '.navbar-nav a[data-value="Code Labels"')
        } else {
          shinyjs::enable(selector = '.navbar-nav a[data-value="Code Labels"')
          updateSelectInput(session = session, inputId = "label_select", choices = v$codes$mancode)
        }
      },
      error = function(e) {
        safeError(e)
      }
    )


    # Assign the imported data to the non-user reactive variable v$data
    if(!is.null(df) && is.data.frame(df)) {
      saveRDS(df, file = paste0('projects/', v$current_project, "/data.RDS"))
      v$data <- df
    } else {
      warning("File not properly imported. Please check the import settings (e.g., separator char). Only CSV files allowed here.")
    }
  }

  # Column Selector Observer
  observeEvent({
    input$text_column
    input$mancode_column
    pi # Arbitrary constant
    }, {
      req(v$data, input$text_column, input$mancode_column)

      if(ncol(v$data)<2) # This only makes sense if more than 1 column is available
        return()
      if(input$text_column==input$mancode_column) { # Moreover, we might experience problems if the same column is selected for both -> thus, prevent this
        warning("Please do not choose the same column to both contain text responses as well as the manual code.")
        return()
      }

      data_changed <- F
      # TEXT: A column name different than the default "text" / previously selected column is selected
      if((input$text_column!="text" || (!is.na(proj_conf["text_columnname"]) && proj_conf["text_columnname"]!="text")) && (is.na(proj_conf["text_columnname"]) || proj_conf["text_columnname"]!=input$text_column)) {
        if("text" %in% colnames(v$data) && is.na(proj_conf["textcol_generated"])) { # Original data contained text column, but it's not selected so that the selected column overwrites the original data
          warning("You selected the text response's column name to deviate from the default name ('text'). However, the imported data already contains a column named `text`. This will now overwrite the original column internally. Beware of this when you export your data back from here or you want to change back to selecting this as your text response column. Your original column will have been lost.")
        }
        if(is.na(proj_conf["textcol_generated"]) && (!"text" %in% colnames(v$data))) { # This will be an artificially generated column as this none-"text"-named column *now* is added to v$data
          set_confvar("textcol_generated", "T", v$current_project)
        }
        if(!is.na(proj_conf["textcol_generated"]) && input$text_column=="text") { # Changed back from none-"text"-named column to "text" -> note in config
          delete_confvar("textcol_generated", v$current_project)
        }
        v$data$text <- v$data[,input$text_column] # Overwrite or add text column in/to v$data
        data_changed <- T
      }
      set_confvar("text_columnname", input$text_column, v$current_project)

      # MANCODE: A column name different than the default "mancode" / previously selected column is selected
      if((input$mancode_column!="mancode" || (!is.na(proj_conf["mancode_columnname"]) && proj_conf["mancode_columnname"]!="mancode")) && (is.na(proj_conf["mancode_columnname"]) || proj_conf["mancode_columnname"]!=input$mancode_column)) {
        if("mancode" %in% colnames(v$data) && is.na(proj_conf["mancodecol_generated"])) { # Original data contained mancode column, but it's not selected so that the selected column overwrites the original data
          warning("You selected the manual codes' column name to deviate from the default name ('mancode'). However, the imported data already contains a column named `mancode`. This will now overwrite the original column internally. Beware of this when you export your data back from here or you wawnt to change back to selecting this as your manual codes' column. Your original column will have been lost.")
        }
        if(is.na(proj_conf["mancodecol_generated"]) && (!"mancode" %in% colnames(v$data))) { # This will be an artificially generated column as this none-"mancode"-named column now is added to v$data
          set_confvar("mancodecol_generated", "T", v$current_project)
        }
        if(!is.na(proj_conf["mancodecol_generated"]) && input$mancode_column=="mancode") { # Changed back from none-"mancode"-named column to "mancode" -> note in config
          delete_confvar("mancodecol_generated", v$current_project)
        }
        v$data$mancode <- v$data[,input$mancode_column] # Overwrite or add mancode column in/to v$data
        data_changed <- T
      }
      set_confvar("mancode_columnname", input$mancode_column, v$current_project)

      # If v$data has been changed, store the updated object in the project's file system and update the code object
      if(data_changed) {
        saveRDS(v$data, file = paste0('projects/', v$current_project, '/data.RDS'))
        update_pp_examples() # Update the list of PP examples

        # Handle color codes
        mcs <-  sort(unique(v$data$mancode))
        add_c <- ifelse(length(mcs)-length(color_scheme) > 0, length(mcs)-length(color_scheme), 0)
        colors <- c(color_scheme, rep(color_scheme[length(color_scheme)],add_c))[1:length(mcs)]
        names <- rep('', length(mcs))
        v$codes <- data.frame(mancode = mcs, colors, names)
        saveRDS(v$codes, file = paste0('projects/', v$current_project, "/codes.RDS"))

        update_pp_examples() # Generate preprocessing examples

        # Disable / Enable Code Labels--Interface (no data yet)
        if(!is.data.frame(v$codes) || nrow(v$codes)==0) {
          shinyjs::disable(selector = '.navbar-nav a[data-value="Code Labels"')
        } else {
          shinyjs::enable(selector = '.navbar-nav a[data-value="Code Labels"')
          updateSelectInput(session = session, inputId = "label_select", choices = v$codes$mancode)
        }
      }
    })

  # Trigger import when interface is Used (Browse)
  observeEvent({
    input$importfile
    input$header
    input$sep
    input$quote
    input$encoding
    input$missing_chars
    pi # Arbitrary constant
    }, {
    req(input$importfile)

    # Try to read imported file
    import_file(input$importfile$datapath)
  })

  # Demo Data Set Button
  observeEvent(input$importdemodata, {
    # import demo dataset
    # create name
    dataname <- paste0('ReCo_Q00', proj_conf["lang"])
    import_file(system.file(paste0("response_data/", dataname, ifelse(grepl("UTF-8|utf8", Sys.getlocale()), "-UTF8", "-ANSI"), ".csv"), package = pkgname), demo = T)
    reset("importfile")
  })

  # If applicable, display filename of imported data set
  output$importfile_name <- renderUI({
    if(is.character(v$import_filename) && nchar(v$import_filename)>0)
      return(HTML(paste0("<img src=\"www/check-circle.svg\" style=\"height:1em;margin-right:3px;margin-bottom:3px;\"><span style=\"font-weight:bold;\">", v$import_filename, "</span> imported", ifelse(substr(v$import_filename, 1, nchar("ReCo_Q00_"))!="ReCo_Q00_", "", ": Description of the <span class=\"reco_warn\" style=\"font-weight:bold;\">Demo Data</span> available at <a href=\"#\" onclick=\"customHref('tabp_INFO');\" style=\"font-weight:bold;\">INFO</a>."))))
    return("")
  })

  # show Imported Data
  output$import_preview <-  DT::renderDataTable({
    if(!"text" %in% colnames(v$data) || length(unique(v$data$mancode)) > 50)
      return()

    v$data %>%  select(any_of(c("id", "text", "mancode"))) %>%
      DT::datatable(colnames = c("ID", "Response", "Manual Code")[ifelse("id" %in% names(v$data), 1, 2):3], rownames = !("id" %in% names(v$data)), filter = "top", #extensions = "Buttons",
                    options = list(columnDefs = list(
                                     list(width = "7%", targets = 0),
                                     list(width = "78%", targets = 1),
                                     list(width = "15%", targets = 2)),
                                   dom = "iptrlB",
                                   buttons =
                                     list(list(
                                       extend = 'collection',
                                       buttons = c('copy'),
                                       text = '<img src="www/copy.svg" style="height:1.5em;margin-left:-3px;margin-right:-4px;margin-bottom:2px;">'),
                                          list(
                                       extend = 'collection',
                                       buttons = c('csv', 'excel', 'pdf'),
                                       text = '<img src="www/saveb.svg" style="height:1.5em;margin-left:-3px;margin-right:-4px;margin-bottom:2px;">'
                                     ))
                                   ),

                    )
  },server = T)

  # Data Import Summary
  output$data_summary <- renderUI({
    req(v$data)
    if(nrow(v$data)==0)
      return()

    if(file.exists(paste0('projects/', v$current_project, "/empty_answers.RDS"))) {
      empty_answers <- readRDS(file = paste0('projects/', v$current_project, "/empty_answers.RDS"))
    } else {
      empty_answers <- 0
    }
    summary_text <- ""
    if(!all(c("text", "mancode") %in% colnames(v$data))) {
      summary_text <- paste0(summary_text, "<div class=\"reco_warn\">WARNING: Columns <span style=\"font-style:italic;\">text</span> and <span style=\"font-style:italic;\">mancode</span> have not been detected. Please select the columns in your data which contain the texts and the manual codes.</div>")
    }
    summary_text <- paste0(summary_text, "<img src=\"www/bar-chart-2.svg\" style=\"height:2.7em;margin-right:4px;float:left;\"><div>Data contains <span style=\"font-weight:bold;\">", nrow(v$data), " row", ifelse(nrow(v$data)!=1, "s", ""), "</span> and <span style=\"font-weight:bold;\">", ncol(v$data), " column", ifelse(ncol(v$data)!=1, "s", ""), ifelse(ncol(v$data)>0, paste0("<img src=\"www/info.svg\" style=\"height:1em;margin-left:2px;margin-bottom:4px;\" title=\"", paste0(colnames(v$data), collapse = ", "), ifelse(!is.na(proj_conf["textcol_generated"]) || !is.na(proj_conf["mancodecol_generated"]), paste0(" (added by the app: ", paste0(c("text", "mancode")[c(!is.na(proj_conf["textcol_generated"]), !is.na(proj_conf["mancodecol_generated"]))], collapse = " and "), ")"), ""), "\">"), ""), "</span>")
    if("text" %in% colnames(v$data) && is.character(v$data$text)) {
      summary_text <- paste0(summary_text, ", excluding <span style=\"font-weight:bold;\">", sum(empty_answers), " empty/missing</span> responses")
    }
    summary_text <- paste0(summary_text, ".</div>")
    if(is.numeric(v$data$mancode)) {
      rcd_text <- ""
      for(lvl in sort(na.omit(unique(v$data$mancode)))) {
        if(nchar(lvl)>0) {
          rcd_text <- paste0(rcd_text, "<span style=\"font-style:italic;margin-right:3px;\">", lvl, "</span> (", round((sum(v$data$mancode==lvl, na.rm = T)*100)/sum(!is.na(v$data$mancode)), 1), "%), ")
        }
      }
      summary_text <- paste0(summary_text, "<div>Response Codes: ", rcd_text, "and <span style=\"font-style:italic;\">", sum(is.na(v$data$mancode)), " code", ifelse(sum(is.na(v$data$mancode))!=1, "s", ""), " missing</span></div>")
    }
    HTML(paste0(hr(), summary_text, hr()))
  })

  # Next Button
  observeEvent(input$next_RA_1, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_2.1")
  })


  # 2.1 Preprocessing ----
  observeEvent(v$show_pp_examples,{
    updateSelectInput(session, inputId = 'text_preprocessing',
                      choices = v$show_pp_examples,
                      selected = 1)
  })

  # function is not outsourced because of problems with umlauts and the individual stopwordlist v$stopwords
  preprocess <- function(texts,
                         decapitalize = T,
                         del_stopwords = T,
                         list_stopwords = c(''),
                         del_numbers = T,
                         del_symbols = T,
                         stemming = T,
                         lang_stemming = 'english',
                         umlaut_normalization = T,
                         pp = F
  ){
    # decapitalize
    if(decapitalize){
      texts <- tolower(texts)
    }
    # del numbers/ symbols
    del_numbers_RegEx <- ifelse(del_numbers, '[:digit:]', '')
    del_symbols_RegEx <- ifelse(del_symbols, '[:punct:]', '')
    texts <- gsub(paste0("[[:cntrl:]", del_numbers_RegEx, del_symbols_RegEx, "]"), " ", texts)
    texts <- trimws(gsub("[[:blank:]]+", " ", texts))


    # delete stop words
    if(del_stopwords){
      pattern <- paste0('\\b',list_stopwords, '\\b', collapse="|") # Create regexp pattern

      if(!pp || input$parallel_cores==1) { # Sequential Processing
        texts <- gsub(pattern, '', texts, ignore.case = T)
      } else { # Parallel Processing
        "%doreco%" <- ifelse(input$do_parallel, `%dopar%`, `%do%`)
        processor_cluster <- snow::makeCluster(input$parallel_cores, type="SOCK")
        on.exit(snow::stopCluster(processor_cluster))
        registerDoSNOW(processor_cluster)

        textssplit <- split(texts, cut(seq_along(texts), max(2, length(processor_cluster)), labels = F))
        texts <- foreach(batch = textssplit, .combine = c) %doreco% {
          gsub(pattern, "", batch, ignore.case = T)
        }

        unregister()
      }

      # Remove double spaces that might have resulted from deleting stop words
      texts <- trimws(gsub("[[:space:]]{2,}", " ", texts))
    }

    # umlaut normalization
    if(umlaut_normalization){
      texts <- gsub('ä', 'a',
                    gsub('ü', 'u',
                         gsub('ö', 'o',
                              gsub('ß', 'ss',
                                   gsub('Ä', 'A',
                                        gsub('Ü', 'U',
                                             gsub('Ö', 'O', texts)
                                        ))))))

    }

    # stemming
    if(stemming){
      texts <- sapply(strsplit(texts, " "), function(r) paste0(SnowballC::wordStem(r, language = lang_stemming), collapse = " "))
    }

    return(trimws(texts))
  }

  # Update Preprocessing Examples List
  update_pp_examples <- function() {
    v$show_pp_examples <- list()
    v$show_pp_examples[['Sample Text']] <- 1
    if(nrow(v$codes)>0) {
      mcl <- if((v$codes$names %>% paste0(., collapse = "") %>% nchar)==0) paste0("Code = ", v$codes$mancode) else v$codes$names
      for(mc in mcl){
        for(i in 1:3) {
          v$show_pp_examples[[paste0('Random Response #', as.character(i), ' (', as.character(mc), ')')]] <- length(v$show_pp_examples) + 1
        }
      }
    }
  }

  # Prepare 3 Sample Texts per Code Level for Demonstrating the Selected Preprocessing Pipeline
  preprocess_samples <- reactive({
    xmpls <- as.character(preprocessing_samples[proj_conf["lang"]])
    if(!is.null(v$data) && is.data.frame(v$data) && nrow(v$data)>0) {
      for(mc in v$codes$mancode) {
        idxs <- which(v$data$mancode == mc & !is.na(v$data$text) & nchar(v$data$text) > 0 & !is.na(v$data$mancode)) # select indices
        idxs <- sample(idxs, ifelse(length(idxs) < 3, length(idxs), 3)) # sample indices
        texts <- c(v$data$text[idxs], rep(as.character(preprocessing_text_nomoreclasssamples[proj_conf["lang"]]), 3-length(idxs)))
        xmpls <- c(xmpls, texts)
      }
    }
    xmpls
  })

  # Display Text Samples Before Preprocessing
  output$exampletext_pre <- renderText({
    preprocess_samples()[as.numeric(input$text_preprocessing)]
  })

  # Display Text Samples After Preprocessing
  output$exampletext_post <- renderText({
    # get the texts
    text <- preprocess_samples()[as.numeric(input$text_preprocessing)]
    l <- lang_codes[[proj_conf["lang"]]]$spellchecker

    if(l!="") { # spelling correction
      dictionary <- hunspell::dictionary(paste0("/dictionaries/", l))
      text <- hunspell4reco(text, hdict = dictionary, v$freq, do = input$spelling_correction)
    }
    # # preprocess + return
    text <- preprocess(texts = text,
                       decapitalize = input$decapitalize,
                       del_stopwords = input$del_stopwords,
                       list_stopwords = v$stopwords,
                       umlaut_normalization = input$umlaut_normalization,
                       del_numbers = input$del_numbers,
                       del_symbols = input$del_symbols,
                       stemming = input$stemming,
                       lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming
    )
    text
  })

  # Display Current Preprocessing Status
  output$preprocessing_status_display <- renderUI({
    HTML(
      switch(which(c('p', 'a', 'd') %in% v$preprocessing_status),
                   "Please press <span style=\"font-style:italic;\">Start Preprocessing</span> for applying the selected techniques to the data.",
                   "Preprocessing currently active ...", # Doesn't work bc the server is busy when this is true
                   "<img src=\"www/check-circle.svg\" style=\"height:1em;margin-right:3px;margin-bottom:6px;\">Preprocessing results ready"
                   )
    )
  })

  # Preprocess All Text Responses from the Data Set
  observeEvent(input$do_preprocess, {
    if(is.null(v$data) || !is.data.frame(v$data) || nrow(v$data)==0) { # All necessary data there?
      warnDialogue("Please import data first (see Step 1).")
    } else {
      preprocess_texts()
    }
  })

  # Button for Deleting Later Results bc an Earlier Process Shall be Re-Run: Pre-Processing
  observeEvent(input$preprocess_over, {
    # Reset status of this and the following steps
    v$preprocessing_status <- "p"
    v$respvector_status <- "p"
    v$respclusters_status <- "p"
    v$cluster_status <- "p"

    # Delete outcome data of this and the following steps
    v$data <- v$data[,-1*(which(names(v$data) == 'text_processed'):ncol(v$data))]
    v$clust_info <- data.frame()
    unlink(paste0('projects/', v$current_project, '/clust_info.RDS'))
    removeModal()
  })

  preprocess_texts <- function() {
    if(!"text" %in% names(v$data)) {
      warnDialogue("Please import data in Step 1 and make sure to choose your text (response) column.")
      return()
    }

    # if preprocessed and further data is already in df, warn user and allow to delete outcomes of later steps
    do_pp <- T
    if('vsm_d1' %in% names(v$data) ){
      showModal(
        modalDialog(
          h3(img(src = "www/alert-octagon.svg", style = "height:1em;margin-bottom:10px;"), "Warning", class = "reco_warn"),
          div("You are about to re-run this process. The results of the further steps you already have performed depend on this process. Press the button below to delete your previously computed results.", class = "reco_warn", style = "margin-bottom:15px;"),
          actionButton('preprocess_over', HTML('<img src="www/trash-2.svg" style="height:1em;margin-bottom:7px;margin-right:5px;">Delete All Outcomes From This Step On'), style = "color:#006699;"),
          easyClose = T
        )
      )
      do_pp <- F
    } else if('text_processed' %in% names(v$data)) { # VSM dimensions not available, but text_processed: simply delete and go ahead
      v$data <- v$data[,-1*(which(names(v$data) == 'text_processed'):ncol(v$data))]
    }

    if(do_pp) {
      v$preprocessing_status <- 'a' # Set flag for active preprocessing
      # preprocess the text and store them in a vector
      # texts <- na.omit(v$data)$text

      ids <- which(!is.na(v$data$text))
      texts <- v$data$text[ids]
      l <- lang_codes[[proj_conf["lang"]]]$spellchecker

      # Spelling Correction
      do_sc <- isolate(input$spelling_correction)
      if(do_sc) {
        freq <- isolate(v$freq)

        if(input$do_parallel) { # if parallel processing is enabled
          processor_cluster <- snow::makeCluster(input$parallel_cores, type="SOCK")
          on.exit(snow::stopCluster(processor_cluster))
          registerDoSNOW(processor_cluster)

          "%doreco%" <- `%dopar%`
          textssplit <- split(texts, cut(seq_along(texts), max(2, length(processor_cluster)), labels = F))

          texts <- foreach(batch = textssplit, .packages=c('hunspell','shinyReCoR'), .combine = "c") %doreco% {
            hunspell4reco(batch, hunspell::dictionary(paste0("/dictionaries/", l)), freq, do_sc)
          }

          unregister()
        } else { # Sequential processing
          texts <- hunspell4reco(texts, hunspell::dictionary(paste0("/dictionaries/", l)), freq, do_sc)
        }
      }

      text_processed <- preprocess(texts = texts,# ignore rows with NAs
                                   decapitalize = input$decapitalize,
                                   umlaut_normalization = input$umlaut_normalization,
                                   del_numbers = input$del_numbers,
                                   del_symbols = input$del_symbols,
                                   del_stopwords = input$del_stopwords,
                                   list_stopwords = v$stopwords,
                                   stemming = input$stemming,
                                   lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming,
                                   pp = input$do_parallel
      )

      # deletes the column when already in data
      if('text_processed' %in% names(v$data)){
        v$data[,which(names(data) == 'text_processed')] <- NULL
      }
      # combine dataset with new data
      v$data <- restore_combine(v$data, data.frame(text_processed, stringsAsFactors = F), ids)
      saveRDS(v$data, paste0('projects/', v$current_project, '/data.RDS'))

      # save configs
      v$pp_decapitalize <- input$decapitalize
      v$umlaut_normalization <- input$umlaut_normalization
      v$pp_del_numbers <- input$del_numbers
      v$pp_del_symbols <- input$del_symbols
      v$pp_del_stopwords <- input$del_stopwords
      v$pp_stemming <- input$stemming
      v$pp_spelling_correction <- input$spelling_correction

      set_confvar("pp_decapitalize", input$decapitalize, v$current_project)
      set_confvar("umlaut_normalization", input$umlaut_normalization, v$current_project)
      set_confvar("pp_del_numbers", input$del_numbers, v$current_project)
      set_confvar("pp_del_symbols", input$del_symbols, v$current_project)
      set_confvar("pp_del_stopwords", input$del_stopwords, v$current_project)
      set_confvar("pp_stemming", input$stemming, v$current_project)
      set_confvar("pp_spelling_correction", input$spelling_correction, v$current_project)

      v$preprocessing_status <- 'd' # Set flag for completed preprocessing
    }
  }

  # Show All Preprocessed Responses in a Modal Dialogue
  observeEvent(input$do_modal_preprocess, {
    if(is.null(v$data) || !is.data.frame(v$data) || !all(c("text", "mancode") %in% colnames(v$data))) { # All necessary data there?
      warnDialogue("Please import data first (see Step 1).")
    } else {
      if(!"text_processed" %in% colnames(v$data)){ # If text has not been preprocessed yet, trigger it now.
        preprocess_texts()
      }
      # Show preprocessing results
      showModal(
        modalDialog(
          h4("Preprocessing Results"),
          DT::renderDataTable(DT::datatable(na.omit(v$data)[,c("text", "text_processed", "mancode")], colnames = c("Response", "Preprocessed Response", "Manual Code"), rownames = F, filter = "top", options = list(columnDefs = list(list(width = "48%", targets = 0), list(width = "40%", targets = 1), list(width = "12%", targets = 2)), dom = "liptr"))),
          easyClose = T, size = "l", footer = modalButton("Close"))
      )
    }
  })

  # Change Stopwords
  observeEvent(input$do_modal_change_stopwords, {
      # Modal for Change Stopword List
      showModal(
        modalDialog(
          h4("Stop Words"),
          renderUI({
            HTML("<span class=\"highlighted\" style=\"font-size:0.8em\">You can change the language in the project settings.</span>")
          }),

          tags$hr(),
          wellPanel(renderUI({
            HTML(paste0(v$stopwords, collapse = ', '))
            }), style = "overflow-y:auto; max-height: 200px;border-style:none;"),

          tags$hr(),
          actionButton('do_reset_stopwords', 'Reset Complete List of Stop Words'),

          tags$hr(),
          h4("Remove Stop Words"),
          renderUI({
            HTML("<span class=\"highlighted\" style=\"font-size:0.8em\">You can type in or select multiple words to remove them from the list.</span>")
          }),
          selectInput("del_stopword_list", "Type in or Select Stop Words", v$stopwords, multiple = TRUE),
          actionButton('do_del_stopwords', 'Remove Stop Words'),

          tags$hr(),
          h4("Add Stop Words"),
          renderUI({
            HTML("<span class=\"highlighted\" style=\"font-size:0.8em\">You can separate multiple stop words with commas; e.g., </span><span class=\"highlighted\" style=\"font-size:0.8em; font-style:italic;\">add, stop, words, in, this, form</span>")
          }),

          textInput('input_stopwords', 'Type in Stop Words', value = "", width = NULL, placeholder = NULL),
          actionButton('do_add_stopwords', 'Add Stop Words'),
          easyClose = T, size = "l", footer = modalButton("Close"))
      )

      # reset stopword list
      observeEvent(input$do_reset_stopwords, {
        v$stopwords <- stopwords::stopwords(language = gsub('_', '', proj_conf["lang"]))
        saveRDS(v$stopwords, file = paste0('projects/', v$current_project, '/stopwords.RDS'))
      })

      # remove selected stopwords
      observeEvent(input$do_del_stopwords, {
        if(any(input$del_stopword_list %in% v$stopwords)) {
          v$stopwords <- v$stopwords[-which(v$stopwords %in% input$del_stopword_list)]
          saveRDS(v$stopwords, file = paste0('projects/', v$current_project, '/stopwords.RDS'))
        }
      })

      # add stopwords from selection
      observeEvent(input$do_add_stopwords, {
        # preprocess and tokenize input string
        input <- strsplit(gsub("[[:punct:]]", "", input$input_stopwords), ' ')[[1]]
        input <- tolower(input[input != ''])
        input <- input[!input %in% v$stopwords]
        v$stopwords <- append(v$stopwords, input)
        v$stopwords <- v$stopwords[order(v$stopwords)]
        saveRDS(v$stopwords, file = paste0('projects/', v$current_project, '/stopwords.RDS'))
      })
    })

  # Prev Button
  observeEvent(input$prev_RA_2.1, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_1")
  })
  # Next Button
  observeEvent(input$next_RA_2.1, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_2.2")
  })

  # 2.2 Word Predictiveness --------
  # Word Frequency only
  freq_table <- reactive({
    req(v$data)
    if(is.null(v$data) || !is.data.frame(v$data) || !all(c("text_processed", "mancode")%in%colnames(v$data)))
      return()

    answers <- na.omit(v$data)$text_processed
    mancode <- na.omit(v$data)$mancode
    N <- nrow(na.omit(v$data))

    # Prepare tokens in the preprocessed responses
    toks <- unique(unlist(str_split(answers, " ")))
    toks <- toks[trimws(toks)!=""]
    trm <- matrix(0, nrow = length(toks), ncol = N) # Create Term Response Matrix

    occurrences <- sapply(toks, function(tok) grep(paste0("\\b", tok, "\\b"), answers, perl = T)) # Find occurrences per token #TODO Bug: Special Characters like in "schadé" (PS428Q05) are not found
    for(i in 1:length(toks))
      trm[i,occurrences[[i]]] <- 1 # Store occurrences, but cap at 1

    # get distribution
    freq_x2 <- sapply(v$codes$mancode, function(mc) rowSums(trm[,mancode==mc, drop = F]))
    # calc chi2 for every token
    chiSq <- apply(freq_x2, 1, function(x) if(sum(x)>0) sum(sapply(x, function(y) (y-mean(x))^2/mean(x) )) else 0)
    # a <- rowSums(trm[,mancode==1]) # number of CORRECT responses in which the token OCCURS
    # b <- rowSums(trm[,mancode==0]) # number of INCORRECT responses in which the token OCCURS
    # c <- rowSums(trm[,mancode==1]==0) # number of CORRECT responses in which the token DOES NOT occur
    # d <- rowSums(trm[,mancode==0]==0) # number of INCORRECT responses in which the token DOES NOT occur
    # chiSq <- (N*(a*d-b*c)^2)/((a+c)*(b+d)*(a+b)*(c+d))

    dist <- c('1' = 22, '33' == 99)
    colnames(freq_x2) <- as.character(v$codes$mancode)
    n_mancode <- apply(freq_x2, 1, function(x) {
      dist <- x[which(x != 0)]
      paste0(
        sapply(1:length(dist), function(y) paste0('<span class="white-space:nowrap;"><span style="color:#AFAFAF;margin-right:4px;">n<sub>', names(dist)[y], '</sub>:</span>', dist[y], '</span>', collapse = ', ')),
        collapse = ', '
      )
      })

    dominant <- unlist(apply(freq_x2, 1, function(x) {
      cs <- names(x)[which(x == max(x))]
      ifelse(length(cs) > 1, paste0(cs, collapse = '/'), cs)
      }))
    # n_mancode <- apply(freq_x2, 1, function(x)  paste0(sapply(1:length(x), function(y)
    #   paste0(v$codes$mancode[y], ':', x[y])), collapse = ' | '))


    n_all <-  apply(freq_x2, 1, function(x) sum(x))

    df <- data.frame("preprocessed_tokens" = toks, "chi_square" = round(chiSq, 3), "n_mancode" = n_mancode, "dominant" = dominant, "n" = n_all)
    df
  })

  output$frequency_tableDT <- DT::renderDataTable({
    req(v$data)
    if(!is.data.frame(v$data) || !all(c("text_processed", "mancode")%in% colnames(v$data) ))
      return()

    freq_table() %>%
      DT::datatable(colnames = c("Token", "Chi<sup>2</sup>", "Code Distribution", "Dominant", "Total"), rownames = F, escape = -c(2:4), filter = "top", options = list(order = list(list(1, "desc")), columnDefs = list(list(width = "25%", targets = 0), list(width = "10%", targets = 1), list(width = "50%", targets = 2), list(width = "10%", targets = 3), list(width = "5%", targets = 3)), dom = "liptr")) %>%
      DT::formatString(columns = 1, prefix = "[", suffix = "]") %>%
      DT::formatRound(columns = 2, digits = 1) %>%
      DT::formatStyle(columns = 1, target = "cell", color = "#006699")
  })

  # download data.RDS
  output$download_predictiveness <- downloadHandler(
    filename = function() {
      paste("ReCo_",v$current_project, "_WordPredictiveness", ".csv", sep="")
    },
    content = function(file) {
      write.csv(freq_table(), file)
    }
  )

  output$wordcloud <- renderCachedPlot({
    req(v$data)
    if(!is.data.frame(v$data) || !all(c("text_processed", "mancode")%in%colnames(v$data)))
      return()

    df <- data.frame(freq_table())

    df$color <- as.character(sapply(df$dominant, function(c){
      if(nchar(c) > 1){
        color <- '#afafaf'
      } else {
        color <- v$codes$colors[as.character(v$codes$mancode) == c]
      }
      color
    }))
    df <- na.omit(df)
    wordcloud::wordcloud(words = df$preprocessed_tokens,
                         freq = df[,ifelse((input$select_order_cloud == 1), 2, 5)],
                         min.freq = input$wordcloud_minfreq,
                         max.words = 200,
                         random.order = F, rot.per = 0.2, scale = c(3.7,1.1),
                         colors = df$color, ordered.colors = T
    )

    text(0.2, .5, "[", cex = 30, col = "#0066993F")
    text(0.8, .5, "]", cex = 30, col = "#0066993F")
  },
  cacheKeyExpr = { list(input$select_order_cloud, freq_table(), input$wordcloud_minfreq) })

  # build a Dialog Frame
  modalResponsePredictiveness <- function(failed = F) {
    df <- freq_table()
    df$preprocessed_tokens <- paste0("[", df$preprocessed_tokens, "]")
    modalDialog(
      renderTable(df),
      easyClose = T, footer = modalButton("Close"))
  }

  # Prev Button
  observeEvent(input$prev_RA_2.2, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_2.1")
  })
  # Next Button
  observeEvent(input$next_RA_2.2, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_3")
  })

  # 3 Text Corpus ----
  # Merge Available VSM Files with Meta Information from Prebuilt Corpora (#ToDo: Change in the long run. Self-made spaces can have meta information too)
  available_vsms <- function() {
    # First, recognize the currently available VSM files (might have changed due to downloads compared to start-up)
    corpus_vsms <- as.list(c(list.files("corpus_vsm", pattern = "_corpusspace"), paste0("#dev#/", list.files("corpus_vsm/#dev#", pattern = "_corpusspace"))))
    names(corpus_vsms) <- sapply(strsplit(unlist(corpus_vsms), "_"), function(x) x[1])
    if(any(names(corpus_vsms)=="#dev#/")) {
      corpus_vsms[[which(names(corpus_vsms)=="#dev#/")]] <- NULL
    }

    # Process the files and meta information
    vsm_files <- corpus_vsms %>% unlist %>% as.character
    available_vsms <- v$prebuilt_vsms %>% select(-matches("Source|File\\.Size|Download\\.Date|Hash")) %>% filter(File.Name %in% vsm_files | paste0("#dev#/", File.Name) %in% vsm_files)
    unmeta_files <- setdiff(vsm_files, v$prebuilt_vsms$File.Name)
    if(length(unmeta_files)>0) {
      available_vsms[(nrow(available_vsms)+1):(nrow(available_vsms)+length(unmeta_files)),"File.Name"] <- unmeta_files
      available_vsms[is.na(available_vsms$Name),"Name"] <- unmeta_files
    }
    available_vsms[is.na(available_vsms$Details),"Details"] <- "No Info"
    available_vsms$Name <- gsub("#dev#/|_corpusspace|\\.RDS", "", available_vsms$Name)

    return(available_vsms[!base::duplicated(available_vsms$Name),])
  }

  # Observer for VSM Selection
  observeEvent(input$available_vsms_table_rows_selected, {
    available_vsms <- available_vsms()
    v$selected_vsm <- c("name"=available_vsms$Name[input$available_vsms_table_rows_selected], "file"=available_vsms$File.Name[input$available_vsms_table_rows_selected])
  })

  # Loads the prebuilt semantic space
  observeEvent(v$selected_vsm, {
    if(!is.character(v$selected_vsm) || length(v$selected_vsm)!=2 || any(names(v$selected_vsm)!=c("name", "file")) || nchar(v$selected_vsm["file"])==0 || (nchar(v$selected_vsm["file"])>0 && !file.exists(file.path("corpus_vsm", v$selected_vsm["file"])) && !file.exists(file.path("corpus_vsm", "#dev#", v$selected_vsm["file"]))))
      return()

    set_confvar("corpus", v$selected_vsm["file"], v$current_project)
    if(nchar(v$selected_vsm["file"])>0) {
      v$vsm <- readRDS(file.path("corpus_vsm", ifelse(file.exists(file.path("corpus_vsm", "#dev#", v$selected_vsm["file"])), "#dev#", ""), v$selected_vsm["file"]))
      colnames(v$vsm) <- paste0('vsm_d', 1:ncol(v$vsm))
    } else {
      v$vsm <- data.frame()
    }
  })

  # Display Available Prebuilt and Self-Made Semantic Spaces with Meta Information and Let Users Select them
  output$available_vsms_table <- DT::renderDataTable({
    req(v$prebuilt_vsms)
    if(!is.data.frame(v$prebuilt_vsms) || nrow(v$prebuilt_vsms)==0)
      return(DT::datatable(data.frame()))

    available_vsms <- available_vsms()

    # Render Data Table
    preselected_rowi <- if(v$selected_vsm["name"]=="") { F } else { which(v$selected_vsm["name"]==available_vsms$Name) }
    DT::datatable(available_vsms %>% mutate("Select" = "", .before = 1), filter = "top", rownames = F,
                  colnames = c("", "Name", "Language", "Stemmed", "Lemma", "Method", "#Dims", "#Types", "File Name", "Details"),
                  options = list(order = list(list(2, "asc"), list(1, "asc")),
                                 columnDefs = list(list(width = "6%", targets = 0, orderable = F, className = "select-checkbox"),
                                                    list(width = "26%", targets = 1, render = JS(
                                                      "function(data, type, row, meta) {",
                                                      "return type === 'display' && data.length > 16 ?",
                                                      "(row[row.length-1]=='No Info' ? '&nbsp;<div style=\"position:absolute;margin-top:-1.4em;\">' : '') + '<span title=\"' + data + '\" style=\"white-space: nowrap;\">' + data.substr(0, (row[row.length-1]=='No Info' ? 100 : 16)).replace(/\\s([^\\s]*)$/, '') + (row[row.length-1]!='No Info' ? '...<img src=\"www/info.svg\" style=\"position:absolute;height:1em;margin-bottom:4px;\" />' : '') + '</span>' + (row[row.length-1]=='No Info' ? '</div>' : '') : data;",
                                                      "}")),
                                                    list(width = "13%", targets = 2, render = JS(paste0(
                                                      "function(data, type, row, meta) {",
                                                      "return type === 'display' && data !== null && data != '", lang_codes[[proj_conf["lang"]]]$stopwords, "' ?",
                                                      "data + '<img src=\"www/cloud-lightning.svg\" style=\"position:relative;height:1em;margin-left:5px;top:-3px;\" alt=\"alert\" title=\"Does not match the current project language: ", lang_codes[[proj_conf["lang"]]]$stopwords, ".\"/>' : (data===null ? '' : data);",
                                                      "}"))),
                                                    list(width = "13%", targets = 3, render = JS(paste0(
                                                      "function(data, type, row, meta) {",
                                                      "return type === 'display' && data !== null && data != '", ifelse(input$stemming, 'TRUE', 'FALSE'), "' ?",
                                                      "data + '<img src=\"www/cloud-lightning.svg\" style=\"position:relative;height:1em;margin-left:5px;top:-3px;\" alt=\"alert\" title=\"Does not match the current project setting: stemming = ", ifelse(input$stemming, 'TRUE', 'TRUE'), "\"/>' : (data===null ? '' : data);",
                                                      "}"))),
                                                    list(width = "0%", targets = 4, visible = F), # Let's Hide Lemmatizing for now as it's irrelevant
                                                    list(width = "10%", targets = 5),
                                                    list(width = "8%", targets = 6),
                                                    list(width = "12%", targets = 7),
                                                    list(width = "0%", visible = F, targets = 8), # Hide File Name (but keep it in the table for JS)
                                                    list(width = "12%", targets = 9, className = "dt-body-center dt-head-center", render = JS(
                                                      "function(data, type, row, meta) {",
                                                      "return type === 'display' && data.length > 29 ?",
                                                      "'<img src=\"www/info.svg\" style=\"position:relative;height:1em;top:-4px;\" title=\"' + data + '\"/>' : data;",
                                                      "}"))
                                 ),
                                 lengthMenu = c(4, 10, 20, 50),
                                 pageLength = 4,
                                 dom = "iptrl"
                  ),
                  escape = -c(1, 9),
                  selection = list(mode="single", selected = preselected_rowi)#, # ToDo: Navigate to page of selected row
                  # callback = JS(ifelse(!preselected_rowi, "", paste0("table.on( 'init.dt', function () {",
                  #                   "table.row(10).show.draw(false);",
                  #                   "});")))
    ) %>%
      DT::formatRound(columns = 7:8, digits = 0) %>%
      DT::formatStyle(columns = 2, target = "cell", color = "#006699")
  })

  output$textcorpus_status_display <- renderUI({
    if(!is.data.frame(v$vsm) || nrow(v$vsm)==0)
      return()

    if("text_processed" %in% names(v$data)) {
      words_vsm <- rownames(v$vsm)
      words_resp <- unlist(sapply(na.omit(v$data$text_processed), function(x) trimws(strsplit(x, ' ')[[1]])))
      uncoverage <- round((1-(sum(words_resp %in% words_vsm)/length(words_resp)))*100, 1)
    }

    HTML(ifelse(!is.null(v$vsm) && is.data.frame(v$vsm) && nrow(v$vsm)>0 && is.character(v$selected_vsm["name"]) && nchar(v$selected_vsm["name"])>0, paste0("<div class=\"highlighted\" style=\"text-align:center;\"><img src=\"www/check-circle.svg\" style=\"height:1em;margin-right:3px;margin-bottom:6px;\">Corpus <span style=\"font-weight:bold;\">", v$selected_vsm["name"], "</span> and its precomputed semantic space loaded</div>", ifelse("text_processed" %in% names(v$data), paste0("<h4 style=\"text-align:center;\">Uncovered Response Tokens (", uncoverage, "%)</h4>"), "")), "If you want to use a prebuilt corpus, choose one from the table. <span style=\"font-size:0.7em;\">More available @<a href=\"#\" onclick=\"customHref('SETUP & CORPUS');\">SETUP</a></a>"))
  })

  # Table of Tokens that are NOT Covered by the Corpus
  output$wordlist_uncovered <- DT::renderDataTable({
    if(!is.data.frame(v$vsm) || nrow(v$vsm)==0 || !"text_processed" %in% names(v$data))
     return()

    words_vsm <- rownames(v$vsm)
    words_resp <- unlist(sapply(na.omit(v$data$text_processed), function(x) trimws(strsplit(x, ' ')[[1]])))
    words_resp <- table(words_resp[!words_resp %in% words_vsm])
    words_resp <- words_resp[order(words_resp, decreasing = T)]

    DT::datatable(data.frame('Word' = as.character(names(words_resp)), 'Total' = as.numeric(words_resp)), rownames = F, options = list(dom = "iptrl")) %>%
      DT::formatString(columns = 1, prefix = "[", suffix = "]") %>%
      DT::formatStyle(columns = 1, target = "cell", color = "#006699")
  })


  # Prev Button
  observeEvent(input$prev_RA_3, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_2.2")
  })
  # Next Button
  observeEvent(input$next_RA_3, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_4")
  })

  # 4 Semantic Space ----
  distanceWords <- eventReactive({  input$do_wordDistances
                                    input$current_project_sel
                                    pi
    }, {
      matrix <- v$vsm

      word1p <- preprocess(texts = input$word1,
                 decapitalize = as.logical(v$pp_decapitalize),
                 del_numbers = as.logical(v$pp_del_numbers),
                 del_stopwords = as.logical(v$pp_del_stopwords),
                 list_stopwords = as.character(v$stopwords),
                 del_symbols = as.logical(v$pp_del_symbols),
                 umlaut_normalization = as.logical(v$umlaut_normalization),
                 stemming = as.logical(v$pp_stemming),
                 lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming
      )

      word2p <- preprocess(texts = input$word2,
                           decapitalize = as.logical(v$pp_decapitalize),
                           del_numbers = as.logical(v$pp_del_numbers),
                           del_stopwords = as.logical(v$pp_del_stopwords),
                           list_stopwords = as.character(v$stopwords),
                           del_symbols = as.logical(v$pp_del_symbols),
                           umlaut_normalization = as.logical(v$umlaut_normalization),
                           stemming = as.logical(v$pp_stemming),
                           lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming
      )



      simil_output <- "<hr>"
      if(word1p=="" || word2p=="") { # At least one word empty
        simil_output <- paste0(simil_output, "Please enter <span class=\"highlighted\">two words</span> for comparing them.")
      } else { # Both words entered
        if(!(word1p %in% row.names(matrix))){ # Word 1 not in semspace
          simil_output <- paste0(simil_output, "Cannot find word: <span class=\"highlighted\">", input$word1, "</span> [", word1p, "]<br/>")
        }
        if(!(word2p %in% row.names(matrix))){ # Word 2 not in semspace
          simil_output <- paste0(simil_output, "Cannot find word: <span class=\"highlighted\">", input$word2, "</span> [", word2p, "]<br/>")
        }
        if((word1p %in% row.names(matrix)) && (word2p %in% row.names(matrix))){ # Go and get their similarity
          simil_output <- paste0(simil_output, "<h5 style=\"font-weight:500;margin-left:10px;\">Similarity of <span class=\"highlighted\" style=\"font-style:italic;\">[", word1p, "]</span> and <span class=\"highlighted\" style=\"font-style:italic;\">[", word2p, "]</span>:</h5><span class=\"highlighted\" style=\"padding-left:10px;font-size:1.1em;\">",
                                 round(as.numeric(proxy::simil(matrix[c(word1p, word2p),],
                                                             method = input$distanceWords,
                                                             convert_distances = T)),
                                     3), "</span>")
        }
      }
      simil_output
    })

  output$show_distanceWords <- renderUI({
    HTML(distanceWords())
  })

  # Only Display Speed-Accuracy-Tradeoff-Slider if t-SNE is selected
  observeEvent(input$dimred_technique_semspace, {
    if(input$dimred_technique_semspace == "tsne") {
      show("tsne_sat_semspace")
    } else {
      hide("tsne_sat_semspace")
    }
  })

  pca_corpus <- reactive({
    if(nrow(v$data)==0 || nrow(v$vsm)==0 || !"text_processed" %in% names(v$data))
      return()
    df <- na.omit(data.frame(v$vsm))
    # only takes words which occour in the answers
    words <- table((strsplit(paste0(na.omit(v$data$text_processed), collapse = ' '), ' ')[[1]]))
    df <- df[rownames(df) %in% names(words),]
    # reduce
    pca <- prcomp(df)

    df$freq <- words[match(rownames(df), names(words))]
    df$word <- names(words)[match(rownames(df), names(words))]
    pca <- data.frame('dim1' = pca$x[,1], 'dim2' = pca$x[,2], 'dim3' = pca$x[,3],
                      'freq'= as.numeric(df$freq), 'word' = as.character(df$word))
    pca
  })

  tsne_corpus <- reactive({
    if(!is.data.frame(v$vsm) || nrow(v$vsm)==0)
      return()
    df <- na.omit(data.frame(v$vsm))
    # only takes words which occour in the answers
    words <- table((strsplit(paste0(na.omit(v$data$text_processed), collapse = ' '), ' ')[[1]]))
    df <- df[rownames(df) %in% names(words),]

    #reduce
    reddim <- tryCatch({Rtsne::Rtsne(df, perplexity = min(((nrow(df)-1)/3), 30), dims = as.numeric(input$dimensions_corpusspace), theta = input$tsne_sat_semspace, check_duplicates = F)$Y},
                       error = function(e) { stop("Too few samples for tSNE-reduction. Please switch to *PCA* instead.\n\nOriginal message: ", e) })
    colnames(reddim) <- paste0('dim', 1:ncol(reddim))

    tsne <- data.frame(reddim,
                      'freq'= as.numeric( words[match(rownames(df), names(words))]),
                      'word' = as.character(names(words)[match(rownames(df), names(words))]))
    tsne
  })
  #TO DO: Plot dynamisch an das Layout anpassen, keine fixen Werte!
  #input$dimensions_respvectors
  output$corpusspace_plotly <- renderPlotly({
    if(nrow(v$data)==0 || nrow(v$vsm)==0 || !"text_processed" %in% names(v$data))
      return()
    # method
    if(input$dimred_technique_semspace == 'pca'){
      df <- pca_corpus()
    } else if(input$dimred_technique_semspace =='tsne') {
      df <- tsne_corpus()
    }
    # dims
    if(input$dimensions_corpusspace == 2){
      plot_ly(x=df$dim1, y=df$dim2,
              type = "scatter", mode="markers", color=log(df$freq), colors = colorRamp(c("#006699", "#f39c12")), marker = list(size = 7), text = df$word, width = 700, height = 600, hovertemplate = paste0("[%{text}]<extra></extra>"), opacity = .85)
    }else{
      plot_ly(x=df$dim1, y=df$dim2, z=df$dim3, type="scatter3d",
              mode="markers", color=log(df$freq), colors = colorRamp(c("#006699", "#f39c12")), size = 2, text = df$word,
              width = 700, height = 600, hovertemplate = paste0("[%{text}]<extra></extra>"))
    }
  })

  # Show Complete List of Tokens in Semantic Space
  modalSemanticSpaceList <- function() {
    if(nrow(v$vsm)==0)
      return()

    freq_table <- freq_table()
    row.names(freq_table) <- freq_table$preprocessed_tokens
    semspace_freq <- data.frame("preprocessed_token" = paste0("[", row.names(v$vsm), "]"), "freq_in_data" = 0, row.names = row.names(v$vsm))
    semspace_freq[intersect(freq_table$preprocessed_tokens, row.names(semspace_freq)),"freq_in_data"] <- freq_table[intersect(freq_table$preprocessed_tokens, row.names(semspace_freq)), "n"]
    modalDialog(
      h4("All (Preprocessed) Tokens in the Semantic Space", style = "margin-bottom:15px;"),
      DT::renderDataTable(DT::formatStyle(DT::datatable(semspace_freq, filter = "top", options = list(order = list(list(1, "desc"), list(0, "asc")), columnDefs = list(list(width = "80%", targets = 0), list(width = "20%", targets = 1)), dom = "liptr"), colnames = c("Preprocessed Token", "Frequency in Responses"), rownames = F), columns = 1, color = "#006699")),
      easyClose = T,
      footer = modalButton("Close")
    )
  }
  # activate Dialog
  observeEvent(input$show_semspacelist, {
    if(nrow(v$vsm)==0) {
      warnDialogue("Please load a semantic space first.")
    } else {
      showModal(modalSemanticSpaceList())
    }
  })

  # Prev Button
  observeEvent(input$prev_RA_4, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_3")
  })
  # Next Button
  observeEvent(input$next_RA_4, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_5")
  })



  # 5 Response Semantics ----
  # Check whether the response vector results are already stored in the file system
  # TODO: try without
  respvecs_in_folder <- reactive({
    ifelse('vsm_d1' %in% names(v$data), 'd', 'p')
  })

  # Display Current Response Vector Status
  output$respvecs_status_display <- renderUI({
    HTML(
      ifelse(respvecs_in_folder()=="p",
             "Please press <span style=\"font-style:italic;\">Compute Response Vectors</span> for projecting the responses into the semantic space.",
             "<img src=\"www/check-circle.svg\" style=\"height:1em;margin-right:3px;margin-bottom:4px;\">Response vectors ready")
    )
  })

  # Button for Deleting Later Results bc an Earlier Process Shall be Re-Run: Pre-Processing
  observeEvent(input$centroid_over, {
    # Reset status of this and the following steps
    v$respvector_status <- "p"
    v$respclusters_status <- "p"
    v$cluster_status <- "p"

    # Delete outcome data of this and the following steps
    v$data <- v$data[,-1*(which(names(v$data) == 'vsm_d1'):ncol(v$data))]
    v$clust_info <- data.frame()
    unlink(paste0('projects/', v$current_project, '/clust_info.RDS'))
    removeModal()
  })

  # Check whether response vectors require computation and if yes, do compute
  observeEvent(input$do_responsevectors, {
    if(nrow(v$data)==0 || !'text_processed' %in% names(v$data) || nrow(v$vsm)==0) {
      warnDialogue("Please carry out the previous steps first. If you are unsure what is required, please refer to the info texts on the left.")
      return()
    }

    # if response types and further data is already in df, warn user and allow to delete outcomes of later steps
    do_cent <- T
    if('vsm_d1' %in% names(v$data) && 'cluster' %in% names(v$data)){
      showModal(
        modalDialog(
          h3(img(src = "www/alert-octagon.svg", style = "height:1em;margin-bottom:10px;"), "Warning", class = "reco_warn"),
          div("You are about to re-run this process. The results of the further steps you already have performed depend on this process. Press the button below to delete your previously computed results.", class = "reco_warn", style = "margin-bottom:15px;"),
          actionButton('centroid_over', HTML('<img src="www/trash-2.svg" style="height:1em;margin-bottom:7px;margin-right:5px;">Delete All Outcomes From This Step On'), style = "color:#006699;"),
          easyClose = T
        )
      )
      do_cent <- F
    } else if('vsm_d1' %in% names(v$data)) { # Clustering not available, but response centroids are: simply delete and go ahead
      v$data <- v$data[,-1*(which(grepl("vsm_d", names(v$data))))]
    }

    # Carry out response centroid vector computation
    if(do_cent) {
      withProgress({
        v$respvector_status <- "a"

        ids <- which(!is.na(v$data$text))
        n <- length(ids)
        # n <- nrow(na.omit(v$data))

        if(input$do_parallel) { # if parallel processing is enabled
          processor_cluster <- snow::makeCluster(input$parallel_cores, type="SOCK")
          on.exit(snow::stopCluster(processor_cluster))
          registerDoSNOW(processor_cluster)
        }

        "%doreco%" <- ifelse(input$do_parallel, `%dopar%`, `%do%`)

        #texts <- na.omit(v$data)$text_processed
        texts <- v$data$text_processed[ids]
        v_vsm <- isolate(v$vsm)
        respvecs <- foreach(i = 1:length(texts), .combine = rbind, .packages = 'shinyReCoR') %doreco% {
          colMeans(v_vsm[strsplit(texts[i], ' ')[[1]],], na.rm = T)
        }
        if(input$do_parallel)
          unregister()

        respvecs <- as.data.frame(respvecs)
        rownames(respvecs) <- NULL
        colnames(respvecs) <- paste0("vsm_d", 1:ncol(respvecs))
        v$data <- restore_combine(v$data, respvecs, ids)
        saveRDS(v$data, file = paste0('projects/', v$current_project, '/data.RDS'))

        v$resp_vectors <- respvecs
        v$respvector_status <- "d"

      },
      message = "Computing Response Vectors", value = 0)
    }

  })

  output$vsm_dims <- DT::renderDataTable({
    if(length(grep('vsm_d', names(v$data))) == 0)
      return()
    df <- na.omit(v$data)[,c("text", paste0('vsm_d', 1:ncol(v$vsm)))]
    stext <- substr(df$text, 1, 40)
    df$text <- paste0(stext, ifelse(nchar(df$text)>40, "...", ""))
    df %>%
      DT::datatable(rownames = F, colnames = c("Response", paste0("#", c(1:ncol(v$vsm)))), options = list(searching = F, pageLength = 4, lengthMenu = c(4, 10), columnDefs = list(list(width = "8%", targets = 0)), dom = "lfiptr")) %>%
      DT::formatRound(columns = 2:(ncol(v$vsm)+1), digits = 3)
  })

  # Data Summary Display
  output$info_max_ans <- renderUI({
    if(length(grep('vsm_d', names(v$data))) == 0)
      return()

    empty_answers <- readRDS(paste0('projects/', v$current_project,'/empty_answers.RDS'))

    text <- paste0('<hr><div class="highlighted"><img src="www/bar-chart-2.svg" style="height:1em;margin-right:3px;margin-bottom:4px;">Data Summary</div>',
                   '<ul style="list-style:none;margin-left:0;padding-left:1em;text-indent:-1em;"><li>&bull; ', sum(!is.na(v$data$text) & nchar(v$data$text)>0, na.rm = T), ' non-empty responses</li><li>+ ', sum(is.na(v$data$text) | nchar(v$data$text)==0), ' empty responses</li>'#<li>+ ', (nrow(v$data) - nrow(v$resp_vectors)), ' responses without any tokens represented in the semantic space</li>'
                   )
    if(is.numeric(v$data$mancode)) {
      rcd_text <- ""
      for(lvl in sort(na.omit(unique(v$data$mancode)))) {
        if(nchar(lvl)>0) {
          rcd_text <- paste0(rcd_text, "<span style=\"font-style:italic;margin-right:3px;\">", lvl, "</span> (", round((sum(v$data$mancode==lvl, na.rm = T)*100)/sum(!is.na(v$data$mancode)), 1), "%), ")
        }
      }
      text <- paste0(text, "<li>&bull; Response Codes: ", rcd_text, "and <span style=\"font-style:italic;\">", sum(is.na(v$data$mancode)), " codes missing</span></li>")
    }

    text <- paste0(text, "<li>&bull; ", length(grep('vsm_', names(v$data))), " dimensions in the vector space model</li></ul>")

    HTML(text)
  })

  # Response Vectors: Reduce Dimensions via PCA
  pca_respvectors <- reactive({
    if(length(grep('vsm_d', names(v$data))) == 0)
      return()

    df <- na.omit(v$data)
    tokens <- sapply(df$text_processed, function(x)
      paste0('[', paste0(strsplit(x, ' ')[[1]], collapse = '], ['), ']')
    )

    text_and_tokens <- paste0(sapply(df$text, function(x)  add_br_to_string(x, 45)),
                              '<br>',
                              sapply(tokens, function(x)  add_br_to_string(x, 45)))
    pca <- prcomp(df[,grepl('vsm_d',names(df))])
    pca <- data.frame('dim1' = pca$x[,1], 'dim2' = pca$x[,2], 'dim3' = pca$x[,3],
                      'mancode'= as.character(df$mancode), 'text' = text_and_tokens)
    pca
  })

  # Response Vectors: Reduce Dimensions via t-SNE
  tsne_respvectors <- reactive({
    if(length(grep('vsm_d', names(v$data))) == 0)
      return()

    # df <- na.omit(v$data)
    df <- v$data[which(!is.na(v$data$vsm_d1) & !is.na(v$data$mancode)),]
    dups <- which(duplicated(df[,grepl('vsm_d',names(df))]))
    tokens <- sapply(df$text_processed, function(x)
      paste0('[', paste0(strsplit(x, ' ')[[1]], collapse = '], ['), ']')
    )

    dups <- which(duplicated(df[,grepl('vsm_d',names(df))]))
    reddim <- tryCatch({Rtsne::Rtsne(df[-dups,grepl('vsm_d',names(df))], perplexity = min(((nrow(df)-length(dups)-1)/3), 30), dims = as.numeric(input$dimensions_respvectors), theta = input$tsne_sat_respvectors, check_duplicates = F)$Y},
                       error = function(e) { stop("Too few samples for tSNE-reduction. Please switch to *PCA* instead.\n\nOriginal message: ", e) })
    colnames(reddim) <- paste0('dim', 1:ncol(reddim))

    # deconstruct (consider all, not only duplicates)
    pat <- sapply(1:nrow(df), function(i) paste0(df[i,grepl('vsm_d',names(df))], collapse = ' '))
    # tab_pat <- table(pat)
    reddim <- reddim[match(pat, pat[-dups]),]

    text_and_tokens <- paste0(sapply(1:nrow(df), function(i) add_br_to_string(df$text[i], 45)),
                              '<br>', # (<span style="font-style:italic;">n</span> = X)<br>', # ToDo: Add number of duplicates
                              sapply(tokens, function(x) add_br_to_string(x, 45)))

    #reddim <- Rtsne::Rtsne(df[-dups,grepl('vsm_d',names(df))], dims = as.numeric(input$dimensions_respvectors), theta = input$tsne_sat_respvectors)
    # colnames(reddim$Y) <- paste0("dim", 1:ncol(reddim$Y))
    return(cbind(reddim, data.frame('mancode'= as.character(df[,"mancode"]), 'text' = text_and_tokens)))
  })

  # Only Display Speed-Accuracy-Tradeoff-Slider if t-SNE is selected
  observeEvent(input$dimred_technique_respvectors, {
    if(input$dimred_technique_respvectors=="tsne") {
      show("tsne_sat_respvectors")
    } else {
      hide("tsne_sat_respvectors")
    }
  })

  #TO DO: Plot dynamisch an das Layout anpassen, keine fixen Werte!
  output$respvectors_plotly <- renderPlotly({
    if(length(grep('vsm_d', names(v$data))) == 0)
      return()

    df <- if(input$dimred_technique_respvectors=="pca") { pca_respvectors() } else { tsne_respvectors() }

    if(all(nchar(v$codes$names) > 0)){
      codes <- v$codes$names[match(df$mancode, as.character(v$codes$mancode))]
      v$codes <- v$codes[order(v$codes$names),]
    } else {
      codes <- as.character(df$mancode)
      v$codes <- v$codes[order(v$codes$mancode),]
    }

    if(input$dimensions_respvectors == 2){
      plot_ly(x = df[,1], y = df[,2],
              mode ="markers", color = codes, colors = colorRamp(v$codes$colors), text = df$text,
              width = 700, height = 600, hovertemplate = paste('%{text}<extra></extra>'), marker = list(size = 6, opacity = .5))
    } else {
      plot_ly(x=df[,1], y=df[,2], z=df[,3], type="scatter3d",
              mode="markers", color = codes, colors = colorRamp(v$codes$colors), size = 2, text = df$text,
              width = 700, height = 600,  hovertemplate = paste('%{text}<extra></extra>'), opacity = .6)
    }
  })

  # Prev Button
  observeEvent(input$prev_RA_5, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_4")
  })
  # Next Button
  observeEvent(input$next_RA_5, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_6")
  })



    # 6 Response Types ----
  # Generate slider for number of clusters
  output$clustnum_slider <- renderUI({
    sliderInput("clustnum", paste0(ifelse(input$clustering=="knn", "k", "Number of Clusters"), ":"),
                min = 1, max = min(800, ifelse(nrow(v$data)==0, 800, nrow(v$data))), value = ifelse(!is.na(proj_conf["clustNum"]), proj_conf["clustNum"], ifelse(input$clustering=="knn", 15, ifelse(nrow(v$data)==0, 800, min(800, length(unique(v$data$text_processed)))))), width = "100%")
  })

  # Stored response vectors in the file system?
  respclusters_in_folder <- reactive({
    ifelse(file.exists(paste0('projects/', v$current_project, '/response_clusters.RDS')), 'd', 'p')
  })

  # Display Current Clustering Status
  output$respclusters_status_display <- renderUI({
    HTML(
      switch(which(c("p", "a", "d")%in%v$cluster_status),
             "Please press <span style=\"font-style:italic;\">Cluster Response Vectors</span> for grouping the responses according to their semantics.",
             "Computation currently active ...", # Doesn't work bc the server is busy when this is true
             "<img src=\"www/check-circle.svg\" style=\"height:1em;margin-right:3px;margin-bottom:4px;\">Response clusters ready"
      )
    )
  })

  # Cluster Response Vectors
  observeEvent(input$do_clust, {
    if(nrow(v$data) == 0 || length(grep('vsm_d', names(v$data))) == 0) {
      warnDialogue("Please carry out the previous steps first. If you are unsure what is required, please refer to the info texts on the left.")
      return()
    }


    withProgress(message = paste0('Clustering'), value = 0, min = 0, max = 1, {


    # if cluster information are already in the data
    if('cluster' %in% names(v$data)){
      v$data[,which(names(v$data) %in% c('dist_to_center', 'cluster'))] <- NULL
    }

    v$cluster_status <- "a"
    # v$respclusters_status <- "a"

    # if KNN
    if(input$clustering == 'knn'){
      #df <- na.omit(v$data)
      # TODO: add autocode to knn
      ids <- which(!is.na(v$data$vsm_d1))
      df <- v$data[ids,]

      v$data <- restore_combine(v$data, data.frame("cluster" = c(1:nrow(df)), "dist_to_center" = rep(0, nrow(df))), ids)
      v$clust_info <- na.omit(cbind(v$data[grep('vsm_d', names(v$data))], cluster = v$data$cluster, n = 1, weighted_entropy = 1, dominant =v$data$mancode))
      names(v$clust_info)[grep('vsm_d', names(v$clust_info))] <- paste0('center_d', 1:length(grep('vsm_d', names(v$data))))

    } else {
      # if HCLUST or KMEANS
      # create distance matrix

      ids <- which(!is.na(v$data$vsm_d1))
      df <- v$data[ids,]
      # df <- na.omit(v$data)
      d <- proxy::dist(df[,grepl('vsm_', names(df))], method = input$distance)
      # if kmeans
      if(input$clustering == 'kmeans'){
        cluster <- kmeans(d, input$clustnum)
        cluster <- cluster$cluster
        # if hclust
      } else if(input$clustering == 'hclust'){
        cluster <- hclust(d, method = "ward.D2", members = NULL)
        cluster <- cutree(cluster, input$clustnum)
      }

      v$data <- restore_combine(v$data, as.data.frame(cluster), ids)
      # CLUSTER INFORMATION
      # create cluster centroids
      v$clust_info = data.frame(matrix(vector(), length(unique(cluster)),  length(grep('vsm_d', names(df))),
                               dimnames=list(c(), paste0('center_d',1:length(grep('vsm_d', names(df)))) )),
                               stringsAsFactors=F)

      v$clust_info$cluster <- unique(cluster)
      for(i in 1:nrow(v$clust_info)){
        setProgress(detail = paste0("Extract Response Types")) # progress bar

        idx <- which(v$data$cluster == v$clust_info$cluster[i])
        center <- apply(v$data[idx,grep('vsm_d', names(v$data))], 2, mean)
        v$clust_info[i, grep('center_d', names(v$clust_info))] <- center

        incProgress((1/(nrow(v$clust_info))/2))
      }

      # calc answer distance to cluster center for all answers in all clusters
      v$data$dist_to_center <- 0 #NA
      lsaidxr <- grep('vsm_', names(v$data))
      lsaidxc <- grep('center_d', names(v$clust_info))
      for(i in 1:nrow(v$clust_info)){
        setProgress(detail = paste0("Visualize Distances")) # progres bar

        center <- v$clust_info[i, lsaidxc]
        idx <- which(v$data$cluster == v$clust_info$cluster[i])
        v$data$dist_to_center[idx] <- sapply(idx, function(iid) {
          acos(min(1, sum(as.numeric(center)*as.numeric(v$data[iid ,lsaidxr])) / ( sqrt(sum(as.numeric(center)^2)) * sqrt(sum(as.numeric(v$data[iid ,lsaidxr])^2)) ) ))
        })
        incProgress((1/(nrow(v$clust_info))/2))
      }

      # calculate entropy for every cluster
      v$clust_info$n <- sapply(v$clust_info$cluster, function(x) length(v$data$mancode[which(v$data$cluster == x)]))
      v$clust_info$entropy <- sapply(v$clust_info$cluster, function(x) entropy(table(v$data$mancode[which(v$data$cluster == x)])/
                                                       length(v$data$mancode[which(v$data$cluster == x)])
      ))
      max_entropy <- entropy(rep(1/nrow(v$codes),nrow(v$codes)))
      v$clust_info$weighted_entropy = (max_entropy - v$clust_info$entropy)*(v$clust_info$n/sum(v$clust_info$n)*100)

      v$clust_info$dominant <- sapply(v$clust_info$cluster, function(x)
        names(table(na.omit(v$data)$mancode[which(na.omit(v$data)$cluster == x)]))
        [which(table(na.omit(v$data)$mancode[which(na.omit(v$data)$cluster == x)]
        ) == max(table(na.omit(v$data)$mancode[which(na.omit(v$data)$cluster == x)])))]
      )
      # if classes are equal
      v$clust_info$dominant <- sapply(v$clust_info$dominant, function(x) if(length(x) > 1){paste0(x, collapse = '/')}else{x})

      # assign automatic codes (AC) for the responses
      v$data$autocode <- NA
      dominant <- sapply(unique(v$data$cluster), function(c) {
        v$data$autocode[which(v$data$cluster == c)] <- v$clust_info$dominant[which(v$clust_info$cluster == c)]
      })
    }

    # save config variables
    set_confvar("distMeasure", input$distance, project =  v$current_project)
    set_confvar("clustMethod", input$clustering, project =  v$current_project)
    set_confvar("clustNum", input$clustnum, project =  v$current_project)

    v$cluster_status <- "d"
    saveRDS(v$data, file = paste0('projects/', v$current_project, '/data.RDS'))
    saveRDS(v$clust_info, file = paste0('projects/', v$current_project, '/clust_info.RDS'))

  }) #end progress bar

  })

  output$knn_info <- renderText({
    if(input$clustering != 'knn' || nrow(v$clust_info) == 0)
      return()
    'Since you have chosen knn as the cluster method, we cannot display information about the cluster distributions. KNN uses the data points themselves for categorization and does not form fixed clusters.'
  })


  # download data.RDS
  output$download_cluster <- downloadHandler(
    filename = function() {
      paste("ReCo_",v$current_project, "_data", ".csv", sep="")
    },
    content = function(file) {
      write.csv(v$data, file)
    }
  )

  # render data table with clustinfo
  output$clustertable <- DT::renderDataTable({
    if(nrow(v$clust_info) == 0 || input$clustering == 'knn')
      return()

    # get the class distribution of all texts as string variable
    clust_distr <- sapply(v$clust_info$cluster, function(x) table(v$data$mancode[which(v$data$cluster == x)]), simplify = F)
    clust_distr <- sapply(clust_distr, function(x) paste0(sapply(1:length(x), function(y) paste0('<span class="white-space:nowrap;"><span style="color:#AFAFAF;margin-right:4px;">n<sub>', names(x)[y], "</sub>:</span>", x[y], '</span>')), collapse = ', '))

    df <- data.frame(v$clust_info$cluster, v$clust_info$n, clust_distr, as.character(v$clust_info$dominant), round(v$clust_info$entropy, 3), round(v$clust_info$weighted_entropy, 3), "Details" = "<img src=\"www/list.svg\" style=\"height:1em;cursor:pointer;\">") %>%
    DT::datatable(colnames = c("Cluster", "<span style='font-style:italic;'>n</span>", "Code Distribution", "Dominant", "Entropy", "Weighted Entropy", "Details"), rownames = F,
                    escape = -c(2, 3, 7),
                    options = list(order = list(list(1, "desc")),
                                   columnDefs = list(list(width = "10%", targets = 0),
                                                     list(width = "10%", targets = 1),
                                                     list(width = "30%", targets = 2),
                                                     list(width = "20%", className = "dt-head-center dt-body-center", targets = 3),
                                                     list(width = "10%", targets = 4),
                                                     list(width = "10%", targets = 5),
                                                     list(width = "10%", targets = 6),
                                                     list(orderable = F, className = "dt-head-center dt-body-center details-control", targets = -1)),
                                   dom = "liptr"
                                ),
                    callback = JS("
                      table.on('click', 'td.details-control', function() {
                          var td = $(this), row = table.row(td.closest('tr'));
                          Shiny.setInputValue('cluster_detail_nr', '');
                          Shiny.setInputValue('cluster_detail_nr', row.data()[0]);
                      })
                                  ")
      )
    df
  })

  # Show Detailed Information on Single Clusters
  observeEvent(input$cluster_detail_nr, {
    if(!is.numeric(input$cluster_detail_nr) || nrow(v$clust_info)==0 || !input$cluster_detail_nr %in% v$data$cluster)
      return()

    # Show Cluster Detail Info
    showModal(
      modalDialog(
        h3(paste0("Details on Cluster ", input$cluster_detail_nr, ": Responses' Distance to the Cluster Centroid"), style = "margin-bottom:-5px;"),
        tags$div(renderPlot({
          par("mar"=c(2, 4.1, 1, 2.1)) # Remove Region for Title
          x <- na.omit(v$data$dist_to_center[v$data$cluster == input$cluster_detail_nr])
          x <- circular::circular(x, type = "angles", units = "radians", modulo = "asis", zero = pi, rotation = "clock")
          circular::rose.diag(x, shrink = 1, bins = 100, ticks = F, border = par("bg"), col = "#646464", tol = 0, prop = 1, axes = T, main = "") # Adjust optional parameters to fit

          for(mc in v$codes$mancode[rev(order(sapply(v$codes$mancode, function(mcc) sum(v$data$cluster == input$cluster_detail_nr & v$data$mancode == mcc, na.rm = T))))]) {
            circs <- as.numeric(na.omit(v$data$dist_to_center[v$data$cluster == input$cluster_detail_nr & v$data$mancode == mc]))
            if(is.vector(circs) && length(circs) > 0){
              circs <- circular::circular(circs, type = "angles", units = "radians", modulo = "asis", zero = pi, rotation = "clock")
              graphics::points(circs, col = paste0(v$codes$colors[v$codes$mancode == mc], "AF"), cex = 0.7, sep = .05, pch = 15+mc, stack = T, bins = 100)#bins=400
            }
          }

          grid::grid.rect(x = unit(0.55, "npc"), y = unit(0.30, "npc"),
                    width = unit(1, "npc"), height = unit(0.4, "npc"),
                    gp = gpar(col="white"))

        })
        ),
        tags$div(HTML("&nbsp;"), style = "margin-top:-220px;"),
        tags$div(DT::renderDataTable(DT::datatable(v$data %>%
                                                 filter(cluster==input$cluster_detail_nr) %>%
                                                 select(text, text_processed, mancode, dist_to_center),
                                               colnames = c("Response", "Preprocessed Response", "Manual Code", "Distance to Centroid"),
                                               rownames = F,
                                               filter = "top",
                                               options = list(
                                                 # order = list(list(3, "asc")),
                                                 columnDefs = list(
                                                    list(width = "38%", targets = 0),
                                                    list(width = "27%", targets = 1),
                                                    list(width = "15%", targets = 2),
                                                    list(width = "20%", targets = 3)
                                                    ),
                                                 dom = "lfiptr"
                                                 )
                                               ) %>%
                                   DT::formatRound(columns = c('dist_to_center'), digits = 3),
              )),
                 # style = "margin-top:-100px;"),
        tags$div("", style = "clear:both;"),
        easyClose = T, size = "l", footer = modalButton("Close")
      )
    )
  })

  # Prev Button
  observeEvent(input$prev_RA_6, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_5")
  })
  # Next Button
  observeEvent(input$next_RA_6, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_7")
  })


  # 7 Evaluation ----
  # Cross Validataion
  observeEvent(input$do_fold, {
    if(!"cluster" %in% names(v$data)) {
      warnDialogue("Please carry out earlier steps first (Step 6 required).")
      return()
    }

    withProgress(message = paste0(input$repetitions, "x ", input$folds, '-Fold Validation:'), value = 0, min = 0, max = 1, {
    # delete unlabeled data
    # takes only data wih class

    df <- v$data[which(!is.na(v$data$mancode) & !is.na(v$data$vsm_d1)),]
    df$cluster <- NULL
    df$mancode <- as.factor(df$mancode)
    #calculate empty answer distribution
    if(input$clustering == 'knn'){
      empty_answers <- round(table(df$mancode[which(is.na(df$vsm_d1))])*(nrow(v$data)-input$clustnum)/nrow(v$data))
    } else {
      empty_answers <- round(table(df$mancode[which(is.na(df$vsm_d1))])*(1-(1/input$folds)))
    }


    # df <- na.omit(df)
    # builds a large distance matrix (including NAs)
    d <- proxy::dist(df[,grepl('vsm_', names(df))], method = input$distance)
    d_mat <- as.matrix(d)
    # get indices of the folds
    k_nums <- kfold(targets = df$mancode, k = input$folds, repetition = input$repetitions)
    max_folds <- sum(unlist(lapply(k_nums, function(x) length(x))))


    if(input$do_parallel) {
      processor_cluster <- snow::makeCluster(input$parallel_cores, type="SOCK")
      on.exit(snow::stopCluster(processor_cluster))
      registerDoSNOW(processor_cluster)
    }

    "%doreco%" <- ifelse(input$do_parallel, `%dopar%`, `%do%`)

    evals <- list()
    for(rep in 1:length(k_nums)){
      all_idx <- unlist(k_nums[[rep]])

        # progress bar
        # setProgress(detail = paste0("Fold ", as.character(fold), " (t = ", as.character(rep), ")"))
        # get the indexes for the testset (trainset = data - testset)
      input_clustering <- isolate(input$clustering)
      input_clustnum <- isolate(input$clustnum)
      input_distance <- isolate(input$distance)
      v_clust_info <- isolate(v$clust_info)
      #freq <- isolate(v$freq)
        ## FOREARCH #####-------
      eval <- foreach(fold = 1:length(k_nums[[rep]]), .combine = rbind, .packages = 'shinyReCoR') %doreco% {

        test_idx <- k_nums[[rep]][[fold]]
        train_idx <- all_idx[!all_idx %in% k_nums[[rep]][[fold]]]
        # split the dataset
        train <- df[train_idx,]
        test <- df[test_idx,]
        train_dist <- as.dist(d_mat[train_idx, train_idx])

        # cluster the data
        if(input_clustering == 'kmeans'){
          cluster <- kmeans(train_dist, nrow(v_clust_info))$cluster
          # if hclust
        } else if(input_clustering == 'hclust'){
          cluster <- cutree(hclust(train_dist, method = "ward.D2", members = NULL), nrow(v_clust_info))
          # if knn (saves every datapoint as cluster)
        } else if(input_clustering == 'knn'){
          cluster <- c(1:nrow(train))
        }
        train$cluster <- cluster

        # create probability dataframe
        pred <- matrix(NA, nrow = nrow(test), ncol = length(empty_answers))
        colnames(pred) <- names(empty_answers)

        # # IF KNN
         if(input_clustering == 'knn'){
          knn_mat <- d_mat[test_idx, train_idx]
          colnames(knn_mat) <- train_idx
          # get the k answers (columns) of the k nearest distances
          for(i in 1:nrow(knn_mat)){
            idx_knn <- as.numeric(names(knn_mat[i,][order(knn_mat[i,], decreasing = F)][1:input_clustnum]))

            distr_clust <- table(df$mancode[idx_knn])
            distr_others <- table(df$mancode[-idx_knn])
            # add empty answers
            if(length(empty_answers) > 0){
              for(name in names(empty_answers)){
                distr_others[name] <-  distr_others[name] +  empty_answers[name]
              }
            }

          pred[i,] <- as.numeric(distr_clust/sum(distr_clust))
          }

          # IF HCLUST OR KMEANS
        } else {
          # CLUSTER INFORMATION
          # TODO: faster
          # create cluster centroids matrix
          # train_clust_info = data.frame(matrix(vector(), length(unique(cluster)),  length(grep('vsm_d', names(train))),
          #                                      dimnames=list(c(), paste0('center_d',1:length(grep('vsm_d', names(train)))) )),
          #                               stringsAsFactors=F)
          # # calculate cluster centers
          # train_clust_info$cluster <- unique(cluster)
          # for(i in 1:nrow(train_clust_info)){
          #   print(i)
          #   idx <- which(train$cluster == train_clust_info$cluster[i])
          #   center <- colMeans(train[idx, grep('vsm_d', names(train))])
          #   train_clust_info[i, grep('center_d', names(train_clust_info))] <- center
          # }

          # add dynamic cluster distribution
          # for(mancode in unique(train$mancode)){
          #   train_clust_info[, paste0('class_', as.character(mancode))] <- sapply(train_clust_info$cluster, function(clustnum)
          #     nrow(train[which(train$cluster == clustnum & train$mancode == mancode),]))
          # }
          #d <- as.matrix(proxy::dist(train_clust_info[,grep('center_', names(train_clust_info))], method = input_distance))

          # faster ----
          train_clust_info <- data.frame(t(sapply(unique(cluster), function(c) { #calc centroids
            colMeans(train[which(train$cluster == c), grep('vsm_d', names(train))])
          })))

          colnames(train_clust_info) <- gsub(pattern = 'vsm', 'center', colnames(train_clust_info))
          train_clust_info$cluster <- unique(cluster)

          classdist <- t(sapply(unique(cluster), function(c) {
            table(train$mancode[which(cluster == c)])
          }))

          dominant <- apply(classdist, 1, function(distr) {
            dom_id <- which(distr == max(distr))
            if(length(dom_id) > 1)
              dom_id <- sample(dom_id, 1)
            colnames(classdist)[dom_id]
          })

          prob <- apply(classdist, 1, function(distr) max(distr)/sum(distr))
          colnames(classdist) <- paste0('class_', colnames(classdist))# sapply(colnames(classdist), function(n) paste0('class_', n))
          train_clust_info <- cbind(train_clust_info, classdist, dominant, prob)

          # combine test- and clustercentroid to calc a distance matrix (fast)
          both <- rbind(as.matrix(df[test_idx, grep('vsm_', colnames(df))]),
                        as.matrix(train_clust_info[,grep('center_', colnames(train_clust_info))])
                        )
          both <- as.matrix(proxy::dist(both, method = input_distance))
          rownames(both) <- colnames(both) <- c(rep(-77, nrow(test)), train_clust_info$cluster)
          both <- both[which(rownames(both) == '-77'), which(colnames(both) != '-77')]

          # calc the nearest cluster to every response
          nearest <- lapply(1:nrow(both), function(i) as.numeric(colnames(both)[which(both[i,] == min(both[i,]))]))
          nearest <- unlist(lapply(nearest, function(x) ifelse(length(x) > 1, sample(x,1), x)))
          # match the dominant class of the nearest cluster

          pred <- train_clust_info[match(nearest, train_clust_info$cluster), grep('class', names(train_clust_info))]
          pred <- pred/rowSums(pred)
          colnames(pred) <- sapply(colnames(pred), function(n) strsplit(n, '_')[[1]][2])

          # get the class distribution of the nearest cluster and calculates the posterior probability for every test data point
          # for(i in 1:length(clusters)){
          #   nearest <- clusters[i]
          #   # if knn
          #   # get the class distributions
          #   distr_clust <- train_clust_info[nearest, grep('class', names(train_clust_info))]
          #   names(distr_clust) <- sapply(names(distr_clust), function(x) strsplit(x, '_')[[1]][2])
          #   distr_others <- apply(train_clust_info[-nearest, grep('class', names(train_clust_info))],2,sum)
          #   names(distr_others) <- sapply(names(distr_others), function(x) strsplit(x, '_')[[1]][2])
          #   # add empty answers
          #   if(length(empty_answers) > 0){
          #     for(name in names(empty_answers)){
          #       distr_others[name] <-  distr_others[name] +  empty_answers[name]
          #     }
          #   }
          #
          #   pred[i,] <- as.numeric(distr_clust/sum(distr_clust))
          #   # pred[i,] <- get_bayes(rbind(distr_clust, distr_others))
          # }
        }

        evaluation_multi(pred, as.character(df$mancode[test_idx]))$results
      }
      evals[[rep]] <- eval
    }
    if(input$do_parallel)
      unregister()
    evals <- do.call(rbind.data.frame, evals)
    saveRDS(evals, file = paste0("projects/", v$current_project, "/kfold_cv.RDS"))
    v$eval_res <- evals
    })
  })

  # Fold Results
  output$kfold_table <- DT::renderDataTable({
    if(nrow(v$eval_res)==0)
      return()

    show("cv_res_h4")
    show("cv_res_hr")
    DT::datatable(round(v$eval_res[v$eval_res$levels == input$label_select_results, -length(v$eval_res)],3),
                  rownames = F,
                  colnames = c('Expected ACC', 'Observed ACC', 'Kappa', 'Weighted Kappa', 'TPR', 'TNR', 'PPV', 'F1'),
                  options = list(dom = "liptr")
    )
  })

  # output$kfold_plot <- plotly::renderPlotly({
  #   list <- lapply(v$eval_res, function(x) unlist(x[4:length(x)]))
  #   df <- data.frame(matrix(unlist(list), nrow=length(list), byrow=T))
  #   names(df) <- names(list[[1]])
  #
  #   fig <- plot_ly(df, x = ~c(1:nrow(df)), y = ~expectedAcc, name = 'expected Acc', type = 'scatter', mode = 'lines')
  #   fig <- fig %>% add_trace(y = ~observedAcc, name = 'observed Acc', mode = 'lines')
  #   fig <- fig %>% add_trace(y = ~Kappa, name = 'Kappa', mode = 'lines')
  #   fig <- fig %>% add_trace(y = ~AUC, name = 'AUC', mode = 'lines')
  #   fig
  # })

  # Mean Fold Results
  output$kfold_results_UI <- renderUI({
    if(nrow(v$eval_res)==0)
      return()

    results <- round(colMeans(v$eval_res[v$eval_res$levels == input$label_select_results, -length(v$eval_res)]),3)
    # results <- round(colMeans(v$eval_res[v$eval_res$levels == '1', -length(v$eval_res)]),3)

    names(results) = c('Expected Accuracy', 'Observed Accuracy', 'Kappa', 'Quadratic Weighted Kappa', 'Sensitivity (TPR)', 'Specifity (TNR)',  'Precision' ,'F1')
    results <- round(results, 3)
    results <- ifelse(is.na(results) | is.nan(results), '-', results)

    text <- paste0(sapply(1:length(results), function(x) paste0('<tr><td><span class=\"highlighted\" style=\"font-style:italic; font-size:1.3em\">',
                                                         names(results)[x], ':   &nbsp',
                                                         '</span></td>',
                                                         '<td><span class=\"highlighted\" style=\"color:#000000; font-size:1.3em\">',
                                                         as.character(results[x]),
                                                         '</span></td></tr>'
                                                         )), collapse = ''
                   )
    text <- paste0('<table id="k-Folds">', text, '</table>')

    show("label_select_results")
    div(
      tags$h4("Aggregated Results"),
      HTML(text)
    )
  })

  # Build a Dialog Frame
  modalEvalDefs <- function(failed = F) {
    modalDialog(
      #renderTable(df),
      easyClose = T, footer = modalButton("Close"))
  }
  # Activate Dialog
  observeEvent(input$do_modal_eval_defs, {
      showModal(modalEvalDefs())
  })

  # Prev Button
  observeEvent(input$prev_RA_7, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_6")
  })

  # Next Button
  observeEvent(input$next_RA_7, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_8")
  })

  # 8 Classification ----
  class_info <- eventReactive(input$do_classify, {
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data))) {
      warnDialogue("Please carry out earlier steps before (Step 6 required).")
      return()
    } else if(nchar(input$text_classifier)==0) {
      warnDialogue("Type a response first and then hit the button again. It's fun!")
      return()
    }
    v$classifier_input <- input$text_classifier

    text <- str_split(preprocess(texts = input$text_classifier,
                                 decapitalize = as.logical(v$pp_decapitalize),
                                 del_numbers = as.logical(v$pp_del_numbers),
                                 del_stopwords = as.logical(v$pp_del_stopwords),
                                 list_stopwords = as.character(v$stopwords),
                                 del_symbols = as.logical(v$pp_del_symbols),
                                 umlaut_normalization = as.logical(v$umlaut_normalization),
                                 stemming = as.logical(v$pp_stemming),
                                 lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming
                                 ), ' ')[[1]]

    ids <- which(!is.na(v$data$mancode) & !is.na(v$data$vsm_d1))
    info <- classify(text = text,
                     data = v$data[ids,],
                     clust_info = v$clust_info,
                     corpus_space = v$vsm,
                     dist_measure = proj_conf["distMeasure"],
                     clust_method = proj_conf["clustMethod"],
                     clustnum = proj_conf["clustNum"])
    if(!is.list(info) && !info) {
      warnDialogue("None of the tokens are covered in the semantic space. Therefore, the entered response cannot be assigned to any response type. For scoring, such a response would be considered incorrect. For coding, a default class would be assigned if there is one.")
      return(F)
    } else {
      info['text_orig'] <- input$text_classifier

      show("classifier_res_hr1")
      show("classifier_res_hr2")
      show("do_modal_classifier_answers")
      info
    }
  })


  # Return Class of the Explorer's Response
  output$resultClass <- renderUI({
    if(!input$do_classify || is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nchar(input$text_classifier)==0 || input$text_classifier != v$classifier_input) {
      return()
    }

    c <- class_info()
    if(!is.list(c) && !c)
      return()

    c <- class_info()$code
    color <- v$codes$colors[v$codes$mancode == c]
    l <- v$codes$names[which(v$codes$mancode == c)]

    text <- paste0('<span class=\"highlighted\">Automatic Code:</span><br/><span class=\"highlighted\" style=\"font-size:1.5em; font-weight: bold; color:',color,'\">',
                   ifelse(nchar(l)>0, l, c),
                   '</span>')
    HTML(text)
  })

  # P
 # print(class_info()$prob)
  # output$certainty <- renderUI({
  #   if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nchar(input$text_classifier)==0) {
  #     return()
  #   }
  #   text <- paste0('<span class=\"highlighted\" style=\"font-style:italic; color:#000000\">',
  #                  'P',
  #                  '</span>',
  #                  '<span class=\"highlighted\" style=\"color:#000000\">',
  #                  '(A|B) = ',
  #                  '</span>',
  #                  '<span class=\"highlighted\" style=\"font-size:1.1em\">',
  #                  as.character(round(as.numeric(class_info()$prob[class_info()$prob == max(class_info()$prob)][1])*100, 2)),
  #                  '</span>'
  #                  )
  #   HTML(text)
  # })

  # Preprocessing for Explorer: Identify Affected Tokens
  beforeafter <- reactive({
    # process text to get raw and pp

    beforeafter <- sapply(trimws(strsplit(input$text_classifier,' ')[[1]]), function(t) preprocess(texts = t,
                                                                                                   decapitalize = as.logical(v$pp_decapitalize),
                                                                                                   del_numbers = as.logical(v$pp_del_numbers),
                                                                                                   del_stopwords = as.logical(v$pp_del_stopwords),
                                                                                                   list_stopwords = as.character(v$stopwords),
                                                                                                   del_symbols = as.logical(v$pp_del_symbols),
                                                                                                   umlaut_normalization = as.logical(v$umlaut_normalization),
                                                                                                   stemming = as.logical(v$pp_stemming),
                                                                                                   lang_stemming = lang_codes[[proj_conf["lang"]]]$stemming

    ))

    # Flag tokens which are not covered by the corpus
    beforeafter <- setNames(ifelse(!beforeafter %in% rownames(v$vsm) & beforeafter != '', 'n.a.vsm', beforeafter), names(beforeafter))
    beforeafter
  })

  # Preprocessing Display in Explorer
  output$original_text <- renderUI({
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nchar(input$text_classifier)==0) {
      return()
    }
    # c <- class_info()
    # if(!is.list(c) && !c)
    #   return()
    # text <- c$text_orig

    beforeafter <- beforeafter()
    string <- '<span class=\"highlighted\">Preprocessing:<br/></span>'
    span_line <- '<span class=\"highlighted\" style=\"color:#949494; text-decoration: line-through\">'
    span_line_red <- '<span class=\"highlighted\" style=\"color:#FF6600; text-decoration: line-through\">'
    span_normal <- '<span class=\"highlighted\" style=\"color:#000000\">'
    span_end <- '</span>'

    string <- paste0(sapply(1:length(beforeafter), function(i) {
      raw <- names(beforeafter[i])
      pp <- as.character(beforeafter[i])
      paste0(sapply(1:nchar(raw), function(j) {
        if(pp == 'n.a.vsm') {
          paste0(span_line_red, substr(raw, j, j), span_end)
        } else if(pp == '') {
          paste0(span_line, substr(raw, j, j), span_end)
        } else {
          if(grepl('[[:punct:]]', substr(raw, j, j)) && v$pp_del_symbols) {
            paste0(span_line, substr(raw, j, j), span_end)
          } else if(grepl('[[:digit:]]', substr(raw, j, j)) && v$pp_del_numbers) {
            paste0(span_line, substr(raw, j, j), span_end)
          } else {
            paste0(span_normal, substr(raw, j, j), span_end)
          }
        }
      }), collapse = '')
    }), collapse = ' ')

    # string <- paste0(string, paste0(sapply(1:length(beforeafter), function(i) {
    #   if(beforeafter[i] == '') {
    #     paste0(span_line, names(beforeafter[i]), span_end, ' ')
    #   } else if(beforeafter[i] == 'n.a.vsm') {
    #     paste0(span_line_red, names(beforeafter[i]), span_end, ' ')
    #   } else {
    #     paste0(span_normal, names(beforeafter[i]), span_end, ' ')
    #   }
    # }), collapse = ''))

    HTML(string)
  })

  # Explorer: Preprocessed Text
  output$processed_text <- renderUI({
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nchar(input$text_classifier)==0) {
      return()
    }

    beforeafter <- beforeafter()
    beforeafter <- beforeafter[!beforeafter %in% c('', 'n.a.vsm')]
    text <- paste0('<span class=\"highlighted\" style=\"color:#646464\">',
                   '[', paste0(beforeafter, collapse = '], ['), ']', # if processed input is a vector
                   '</span>'
    )
    HTML(text)
  })

  # Explorer: Cluster Infos
  output$cluster_info <- renderUI({
    # show and hide lines

    if(nchar(input$text_classifier)==0) {
      hide("classifier_res_hr1")
      hide("classifier_res_hr2")
    } else {
      show("classifier_res_hr1")
      show("classifier_res_hr2")
    }

    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nrow(v$clust_info) == 0 || is.null(v$clust_info) || nchar(input$text_classifier)==0) {
      return()
    }

    clust_method <- proj_conf['clustMethod']
    sent <- ifelse(clust_method == 'knn', 'Similar Responses: ', 'Prototypical Responses of the Cluster: ')
    df <- class_info()
    if(!is.list(df) && !df || input$text_classifier != v$classifier_input)
      return()
    df <- df$simil_answers

    HTML(paste0('<span class=\"highlighted\ style=\"font-weight: bold; font-size:0.9em\">', sent,'</span>',
                '<span class=\"highlighted\" style=\"color:#646464\">','(','</span>',
                '<span class=\"highlighted\" style=\"color:#646464; font-style:italic; font-size:0.9em\">', 'n', '</span>',
                '<span class=\"highlighted\" style=\"color:#646464; font-size:0.9em\">',' = ', as.character(nrow(df)),
                ', <span style="font-weight:bold;font-style:italic">', 'P','</span>',
                '<span style="font-weight:bold;">(', class_info()$code ,'|</span>',
                '<span class=\"highlighted\" style=\"color:#646464; font-weight:bold; font-style:italic; font-size:1.0em\">C</span>',
                '<span style="font-weight:bold;">)</span>',' = ',
                as.character(round(class_info()$prob[class_info()$prob == max(class_info()$prob)][1]*100, 2)),
                '%)'
    ))
  })

  # Explorer: Print the (3) Most Similar Responses
  # TODO: actual prints NAs too
  output$similar_sents <- renderUI({
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nrow(v$clust_info) == 0 || is.null(v$clust_info) || nchar(input$text_classifier)==0 || input$text_classifier != v$classifier_input) {
      return()
    }

    df <- class_info()
    if(!is.list(df) && !df)
      return()
    df <- df$simil_answers
    codes <- df$mancode[1:min(3, nrow(df))]
    if(all(nchar(v$codes$names) > 0)) {
      codes <- v$codes$names[match(codes, v$codes$mancode)]
    }

    HTML(paste0('<span class=\"highlighted\" style=\"color:#000000; font-size:0.9em\">',
                df$text[1:min(3, nrow(df))],
                '</span>',
                '<span class=\"highlighted\" style=\"color:#000000; font-size:0.9em;font-weight:bold\">',
                ' (',as.character(codes), ') ',
                '</span>',
                '<br/>', collapse = ''))
  })

  # Explorer: Cluster Detail Infos in Modal Dialog
  modalResponseAnswers <- function(failed = F) {
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nrow(v$clust_info) == 0 || is.null(v$clust_info) || nchar(input$text_classifier)==0) {
      return()
    }

    modalDialog(
      h3(paste0("Details on Cluster ", ifelse(length(class_info()$clustnr) == 1, class_info()$clustnr, ''), ": Responses' Distance to the Cluster Centroid"), style = "margin-bottom:-5px;"),
      tags$div(renderPlot({
        suggest <- class_info()
        if(!is.list(suggest) && !suggest)
          return()

        par("mar"=c(2, 4.1, 1, 2.1)) # Remove Region for Title
        x <- as.numeric(na.omit(class_info()$simil_answers$dist_to_center))
        x <- circular::circular(x, type = "angles", units = "radians", modulo = "asis", zero = pi, rotation = "clock")
        circular::rose.diag(x, shrink = 1, bins = 100, ticks = F, border = par("bg"), col = "#646464", tol = 0, prop = 1, axes = T, main = "")  # Adjust optional parameters to fit

        for(mc in v$codes$mancode[rev(order(sapply(v$codes$mancode, function(mcc) sum(v$data$cluster == input$cluster_detail_nr & v$data$mancode == mcc, na.rm = T))))]) {
          #print(mc)
          circs <- as.numeric(na.omit(class_info()$simil_answers$dist_to_center[class_info()$simil_answers$mancode == mc]))
          #circs <- as.numeric(na.omit(suggest$simil_answers$dist_to_center[suggest$simil_answers$mancode == mc]))
          if(is.vector(circs) & length(circs) > 0){
            circs <- circular::circular(circs, type = "angles", units = "radians", modulo = "asis", zero = pi, rotation = "clock")
            graphics::points(circs, col = paste0(v$codes$colors[v$codes$mancode == mc], "AF"), cex = 0.7, sep = .05, pch = 15+mc, stack = T, bins = 100)
          }
        }

        grid::grid.rect(x = unit(0.55, "npc"), y = unit(0.30, "npc"),
                        width = unit(1, "npc"), height = unit(0.4, "npc"),
                        gp=gpar(col="white"))

      }),
      ),
      tags$div(HTML("&nbsp;"), style = "margin-top:-100px;"),
      tags$div(DT::renderDataTable({DT::datatable(class_info()$simil_answers[,c(1:4)] %>%
                                               select(text, text_processed, mancode, dist_to_center),
                                             colnames = c("Response", "Preprocessed Response", "Manual Code", "Distance to Centroid"),
                                             rownames = F,
                                             filter = "top",
                                             options = list(
                                               columnDefs = list(
                                                 list(width = "38%", targets = 0),
                                                 list(width = "27%", targets = 1),
                                                 list(width = "15%", targets = 2),
                                                 list(width = "20%", targets = 3)
                                               ),
                                               dom = "lfiptr"
                                             )
      ) %>%
        DT::formatRound(columns = c('dist_to_center'), digits = 3)}),
        style = "margin-top:-100px;"),
      tags$div("", style = "clear:both;"),
      easyClose = T, size = "l", footer = modalButton("Close")
    )
  }

  # Explorer: Activate Dialog for Cluster Details
  observeEvent(input$do_modal_classifier_answers, {
    if(is.null(v$data) || is.null(v$vsm) || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nrow(v$clust_info) == 0 || is.null(v$clust_info))
      return()

    showModal(modalResponseAnswers())
  })

  # Dimensionality Reduction Technique for Plot (Explorer?!)
  observeEvent(input$dimred_technique_classifier, {
    if(input$dimred_technique_classifier=="tsne") {
      show("tsne_sat_classifier")
    } else {
      hide("tsne_sat_classifier")
    }
  })

  # Principal Component Analysis (Explorer)
  pca_classifier <- eventReactive(class_info()$text_orig,{
    # load answers
    if(input$dimensions_respvectors_classifier == 1)
      return()
    ids <- which(!is.na(v$data$mancode) & !is.na(v$data$vsm_d1))
    df <- v$data[ids,]
    # df <- na.omit(v$data)
    # add new (manual) answer
    df_ans <- as.data.frame(c(text = class_info()$text_orig,
                              mancode = '-7',
                              text_processed = paste0(class_info()$text_processed, collapse = ' '),
                              class_info()$centroid_answer,
                              cluster = NA,
                              dist_to_center = 0))

    # rename columns for rbind
    names(df_ans)[grep('center_d', names(df_ans))] <- paste0('vsm_d', 1:length(names(df_ans)[grep('center_d', names(df_ans))]))
    df <- df[,names(df) %in% names(df_ans)]
    df <- rbind(df, df_ans)

    # manipulate text/ tokens
    df$text_and_tokens <- sapply(df$text_processed, function(x)
      paste0('[', paste0(strsplit(x, ' ')[[1]], collapse = '], ['), ']')
    )

    df$text_and_tokens <- paste0(sapply(df$text, function(x)  add_br_to_string(x, 45)),
                              '<br>',
                              sapply(df$text_and_tokens, function(x)  add_br_to_string(x, 45)))
    # pca
    pca <- prcomp(df[,grepl('vsm_',names(df))])

    if(all(nchar(v$codes$names) > 0)) {
      codes <- v$codes$names[match(df$mancode, v$codes$mancode)]
      } else {
        codes <- df$mancode
      }

    pca <- data.frame('dim1' = pca$x[,1], 'dim2' = pca$x[,2], 'dim3' = pca$x[,3],
                      'mancode'= as.character(codes), 'text' = df$text_and_tokens)
    # add infos
    pca$mancode <- as.character(df$mancode)
    pca$text_and_tokens <- as.character(df$text_and_tokens)
    pca$cluster <- ifelse(as.character(df$cluster) %in% as.character(class_info()$clustnr), 'Similar Answer', 'Dissimilar Answer')
    pca$mancode[nrow(pca)] <- 'Your Response'
    pca$cluster[nrow(pca)] <- 'Your Response'
    pca
  })

  # T-SNE (Explorer)
  # tsne_classifier <- eventReactive(class_info()$text_orig,{
  tsne_classifier <- reactive({
    # print('TSNE')
    if(input$dimensions_respvectors_classifier == 1 || length(grep('vsm_d', names(v$data))) == 0 || class_info()$text_orig == '')
      return()

    ids <- which(!is.na(v$data$mancode) & !is.na(v$data$vsm_d1))
    df <- v$data[ids,]
    # add new (manual) answer
    df_ans <- as.data.frame(c(text = class_info()$text_orig,
                              mancode = '-7',
                              text_processed = paste0(class_info()$text_processed, collapse = ' '),
                              class_info()$centroid_answer,
                              cluster = NA,
                              dist_to_center = 0))

    # rename columns for rbind
    names(df_ans)[grep('center_d', names(df_ans))] <- paste0('vsm_d', 1:length(names(df_ans)[grep('center_d', names(df_ans))]))
    df <- df[,names(df) %in% names(df_ans)]
    df <- rbind(df, df_ans)

    # manipulate text/ tokens
    df$text_and_tokens <- sapply(df$text_processed, function(x)
      paste0('[', paste0(strsplit(x, ' ')[[1]], collapse = '], ['), ']')
    )

    df$text_and_tokens <- paste0(sapply(df$text, function(x)  add_br_to_string(x, 45)),
                                 '<br>',
                                 sapply(df$text_and_tokens, function(x)  add_br_to_string(x, 45)))
    # tsne
    dups <- which(duplicated(df[,grepl('vsm_d',names(df))]))
    reddim <- tryCatch({Rtsne::Rtsne(df[-dups,grepl('vsm_d',names(df))], perplexity = min(((nrow(df)-length(dups)-1)/3), 30), dims = as.numeric(input$dimensions_respvectors_classifier), theta = input$tsne_sat_classifier, check_duplicates = F)$Y},
                       error = function(e) { stop("Too few samples for tSNE-reduction. Please switch to *PCA* instead.\n\nOriginal message: ", e) })
    colnames(reddim) <- paste0('dim', 1:ncol(reddim))

    # deconstruct (consider all, not only duplicates)
    pat <- sapply(1:nrow(df), function(i) paste0(df[i,grepl('vsm_d',names(df))], collapse = ' '))
    reddim <- reddim[match(pat, pat[-dups]),]

    if(all(nchar(v$codes$names) > 0)) {
      codes <- v$codes$names[match(df$mancode, v$codes$mancode)]
    } else {
      codes <- df$mancode
    }

    tsne <- data.frame(reddim,
                      'mancode'= as.character(codes),
                      'text' = df$text_and_tokens)
    # add infos
    tsne$mancode <- as.character(df$mancode)
    tsne$text_and_tokens <- as.character(df$text_and_tokens)
    tsne$cluster <- ifelse(as.character(df$cluster) %in% as.character(class_info()$clustnr), 'Similar Answer', 'Dissimilar Answer')
    tsne$mancode[nrow(tsne)] <- 'Your Response'
    tsne$cluster[nrow(tsne)] <- 'Your Response'
    tsne
  })

  # Explorer: Plot
  #TO DO: Plot dynamisch an das Layout anpassen, keine fixen Werte!
  output$classifier_plotly <- renderPlotly({
    if(!input$do_classify || input$dimensions_respvectors_classifier == 1 || any(!c('vsm_d1', 'cluster') %in% names(v$data)) || nrow(v$clust_info) == 0 || is.null(v$clust_info) || input$text_classifier != v$classifier_input) {
      if(input$do_classify & nrow(v$clust_info) == 0)
        warnDialogue("Please carry out previous steps before (Step 6 required).")
      return()
    }

    if(!is.list(class_info()))
      return()
    # method
    if(input$dimred_technique_classifier == 'pca') {
      df <- pca_classifier()
    } else if(input$dimred_technique_classifier == 'tsne') {
      df <- tsne_classifier()
    }
    df[nrow(df)+1,] <- df[nrow(df),] # need to duplicate "Your Response" bc of hovertemplate-bug in plotly

    # normal plot
    codes <- df$mancode
    colors <- c(v$codes$colors[match(codes[-(length(codes):(length(codes)-1))], v$codes$mancode)], rep_len('#669900', length.out = 2))
    # rename codes if names are given
    if(all(nchar(v$codes$names)>0)) {
      codes <- c(v$codes$names[match(df$mancode[-(length(df$mancode):(length(df$mancode)-1))], as.character(v$codes$mancode))], rep_len("Your Response", length.out = 2))
    }

    # integrate cluster specific information
    if(input$show_cluster){
      codes <- ifelse(df$cluster == 'Dissimilar Answer', 'Unrelated Answer', codes)
      colors <- as.character(sapply(codes, function(c)
        if(c %in% as.character(v$codes$mancode)) {
        v$codes$colors[v$codes$mancode == c]
         } else if(c%in% as.character(v$codes$names)) {
           v$codes$colors[v$codes$names == c]
        } else if(c == 'Dissimilar Answer') {
          '#FFFFFF'
        } else if(c == 'Unrelated Answer') {
          '#ebebeb'
        } else if(c == 'Your Response') {
          '#669900'
        }))
    }
    palette <- unique(colors[-length(colors)])[order(unique(codes[-length(codes)]))] #c(unique(colors[-length(colors)])[order(unique(codes[-length(codes)]))],colors[length(colors)])
    # dims
    if(input$dimensions_respvectors_classifier == 2){
      plot_ly(x = df[,1], y = df[,2],
              mode ="markers", color = codes, colors = palette, text = df$text_and_tokens,
              width = 700, height = 600, hovertemplate = paste('%{text}<extra></extra>'), marker = list(size = ifelse(codes=="Your Response", 12, 5), opacity = ifelse(codes=="Your Response", 1, .6)))
    } else {
      plot_ly(x=df[,1], y=df[,2], z=df[,3], type="scatter3d",
              mode="markers", color=codes, colors = palette, size = ifelse(codes=="Your Response", 2, 1), sizes = c(50, 120), text = df$text_and_tokens,
              width = 700, height = 600,  hovertemplate = paste('%{text}<extra></extra>'), opacity = ifelse(codes=="Your Response", 1, .6))
    }
  })

  # Prev Button
  observeEvent(input$prev_RA_8, {
    updateTabsetPanel(session, "tabs_RA",
                      selected = "tabp_RA_7")
  })

  # $ Interface Management ----
  # Save Button
  observeEvent(input$btn_save, {
    showModal(
      modalDialog(
        h5(img(src = "www/alert-octagon.svg"), "This function is not available in the current app's version.", class = "reco_warn"),
        footer = modalButton("Sure, I'm okay. #wasJustBeingCurious"),
        easyClose = T,
        size = "m"
      )
    )
  })

  # Expert Button
  observeEvent(input$btn_xpert, {
    showModal(
      modalDialog(
        h5(img(src = "www/alert-octagon.svg"), "This function is not available in the current app's version", class = "reco_warn"),
        footer = modalButton("Sure, I'm okay. #wasJustBeingCurious"),
        easyClose = T,
        size = "m"
      )
    )
  })

  # Warn Dialogue
  warnDialogue <- function(text) {
    showModal(
      modalDialog(
        h5(img(src = "www/alert-octagon.svg", style = "height:1em;float:left;margin-right:10px;margin-bottom:10px;"), text, class = "reco_warn", style = "font-weight:normal;"),
        footer = modalButton("OK"),
        easyClose = T,
        size = "m"
      )
    )
  }

  # $ Info Texts ----
  output$txt_RA_info <- renderText({
    switch(input$tabs_RA,
           "tabp_RA_1" = "<h6>Step 1: Data Import</h6><p>In this step, you import your response data. If available, your data file should also contain response codes <img src=\"www/info.svg\" alt=\"info\" title=\"Response codes are, depending on the context, sometimes called classes, labels, or annotations.\" style=\"height:1em;margin-bottom:4px;\"> (e.g., a variable X E {0, 1} for <span style=\"font-style:italic;\">correct</span> and <span style=\"font-style:italic;\">incorrect</span>). Only then, you can carry out automatic coding beyond mere response categorization.</p><p>The data set must contain a text column and (currently) a column with response codes. Soon, we will allow to import text data only for unsupervised response categorization, too. Other columns are just carried along, not affecting the workflow.<p class='highlighted'>In the current app version, only CSV files are allowed that need to contain two columns. If the columns are not named <span style=\"font-style:italic;\">text</span> and <span style=\"font-style:italic;\">mancode</span> (manual code), you can use the column selector to choose the columns that contain your text and manual code data.</p>",
           "tabp_RA_2.1" = "<h6>Step 2.1: Preprocessing</h6><p>In this step, you specify the techniques that are applied for preprocessing your response texts.</p><p>The goal of this step is to reduce some (hopefully irrelevant) variance from the responses in order to make them more comparable. For example, the phrases <span style=\"font-style:italic;\">The doggs howled</span> and <span style=\"font-style:italic;\">dogs howling</span> can be considered identical in responses for most assessments and can thus be normalized (i.e., preprocessed) to <span style=\"font-style:italic;\">[dog] [howl]</span>.</p>",
           "tabp_RA_2.2" = "<h6>Step 2.2: Word Predictiveness</h6><p>This step is optional and only informs you about which words occur how often in your data set and how predictive single words are for their response code. For example, if responses containing the word <span style=\"font-style:italic;\">president</span> are mostly correct, this word is considered predictive. In this step, you should take notice of words that your text corpus should contain as they are predictive and occur frequently. Here, chi-square values serve as indicators for predictiveness (the higher, the more specific for one particular code).</p><p>Note that you should also check word predictiveness without removing stop words in order to check whether you would need to adapt your list of stop words.</p><p>Please also note that this step actually displays preprocessed tokens (indicated by [brackets]), not words. We use the term <span style=\"font-style:italic;\">words</span> here for better comprehensibility.</p>",
           "tabp_RA_3" = "<h6>Step 3: Text Corpus</h6><p>In the completed app, you will be able to create your own text corpora based on text collections such as Wikipedia. For this, you can select the text collection, and for example for the Wikipedia, enter the title of a page. Additionally, ReCo will make suggestions based on your response data. Starting from the selected page(s), ReCo crawls all additional pages (of a Wikipedia dump) which the page links to and, optionally, which link to the page. For gathering larger corpora, you can select varying depths for the crawling procedure. With a depth of 2, for example, ReCo does not only crawl pages based on in- and outgoing links of the page, but also crawls pages based on in- and outgoing links from the pages in the first crawling step.</p><p>For now, however, you cannot create your own text corpora and build semantic spaces but only select one we have built for you and which are ready to use. Make sure to select one that fits your project settings (test language, preprocessing pipeline, semantics). Additional prebuilt spaces are available for download in the Setup & Corpus Tab.</p><p>After having selected a corpus, the table on the bottom right shows all response tokens that are <span class=\"highlighted\">not covered</span> by the corpus. These will be ignored in subsequent steps. If you recognize relevant tokens (see also <span style=\"font-style:italic;\">2.2 Word Predictiveness</span>), please adjust the corpus.</p>",
           "tabp_RA_4" = "<h6>Step 4: Semantic Space</h6><p>In this step, you use the text corpus built or selected in the previous step for computing the semantic space. This means, a Latent Semantic Analysis (Deerwester, Dumais, Furnas, Landauer, & Harshman, 1990) or some other technique (like word2vec) computes a (typically) 300-dimensional vector for each token (i.e., preprocessed word) in the text corpus. If you selected a prebuilt text corpus, it already comes with the pre-computed semantic space.</p><p>Vectors of words that are considered semantically similar have a low angle. This is typically reported using cosine similarity (values close to 0 indicate no similarity, higher values higher similarity).</p><p>Here, you can explore the text corpus's semantic space visually in the plot on the right and quantitatively by computing words' semantic similarity. Please note that the plot only displays words that occur in the imported response data.</p><p class=\"highlighted\">In the current version of the app, it is only possible to use prebuilt text corpora and pre-computed vector spaces.</p><p style=\"font-size:0.8em;\">Deerwester, S., Dumais, S. T., Furnas, G. W., Landauer, T. K. & Harshman, R. (1990). Indexing by latent semantic analysis. Journal of the American Society for Information Science, 41 (6), 391–407. <a href=\"https://doi.org/10.1002/(SICI)1097-4571(199009)41:6<391::AID-ASI1>3.0.CO;2-9\">[doi]</a><br /><span style=\"font-style:italic;\">Further reading into LSA:</span> Landauer, T. K., McNamara, D. S., Dennis, S. & Kintsch, W. (Eds.). (2011). Handbook of latent semantic analysis. Erlbaum: Mahwah.</p>",
           "tabp_RA_5" = "<h6>Step 5: Response Semantics</h6><p>In this step, the preprocessed text responses are projected into the semantic space computed or selected in the previous step. For this, all tokens' vectors in a response are aggregated, resulting in one centroid vector for each response.</p><p class=\"highlighted\">Please note that the figure shows a substantial reduction of the semantic space that actually comprises (typically) 300 dimensions. This is why it can happen that the actual similarity between two responses differs from what is visualized in the figure.</p>",
           "tabp_RA_6" = "<h6>Step 6: Response Types</h6><p>At this point, responses are clustered into response types according to their representation in the semantic space. To determine the similarity of the response vectors, different distance measures (Euclidean and cosine) can be selected.</p><p class=\"highlighted\">Typically, cosine is used in this context.</p><p>After the analysis, you can investigate single clusters by clicking on them in the table. <p class=\"highlighted\">Use the button <span style=\"font-style:italic;\">Download Data</span> to download your classified and clustered data.</p>",
           "tabp_RA_7" = "<h6>Step 7: Evaluation</h6><p>In this step, you can evaluate the just learned cluster model through cross-validation. For this, the data is randomly split into <span style=\"font-style:italic;\">k</span> equal parts (folds), taking the code distribution into account. The classifier will be trained with <span style=\"font-style:italic;\">k</span>-1 folds and evaluated on the remaining fold. This procedure is repeated <span style=\"font-style:italic;\">k</span> times so that each fold is evaluated once. In order to keep the influence of randomization as small as possible, the whole process can be repeated several times. You can manually determine the number of folds and the number of repetitions.</p><p class=\"highlighted\">The process may take some time, while the progress can be monitored using the progress bar (bottom right). </p>",
           "tabp_RA_8" = "<h6>Step 8: Classifier</h6><p class=\"highlighted\" style=\"text-align:center;\">You're done! Try it yourself and type some response.</div></p><p>In this last step, you can explore the trained classifier. Your response is projected into the semantic space, and the code distribution in the most similar response type determines whether your response is correct or not.</p><p>In a later app version, the app will allow you to export the classifier so that you can use it elsewhere (for example, in the runtime environment of your test).</p>"
    )
  })

  # $ Welcome to App Screen ----
  observeEvent(input$tabs_RA, {
    if(!is.na(proj_conf["remove_app_popup"])) # Only display popup once
      return()

    set_confvar("remove_app_popup", 1, v$current_project) # Set config variable cookie to make sure we do not bother users with the popup over and over again
    update_proj_conf(v$current_project)

    app_info_popup()
    }
  )

  # Show App Info Popup every time the user navigates to the INFO menu
  observeEvent(input$ReCo, {
    if(input$ReCo=="INFO")
      app_info_popup()
  })

  # App Info Popup
  app_info_popup <- function() {
    showModal(
      modalDialog(
        div(
          img(src = "www/ReCo-Logo.png", height = 55, style = "float:left;margin:14px 10px 0px 15px;"),
          h2("Thrilled to See You in the ReCo Shiny-App!", style = "text-align:right;padding-left:10px;margin-right:15px;padding-top:39px;padding-bottom:15px;"),
          style = "position:relative;top:-41px;left:-20px;width:calc(100% + 40px);background-color:#006699;color:#FFFFFF;min-height:80px;margin-bottom:-30px;"
        ),
        img(src = "www/file-text.svg", width = 150, style = "clear:both;float:left;margin-right:15px;margin-bottom:20px;margin-left:-15px;"),
        p(HTML("The <span class=\"highlighted\">Automatic Text Response Coder (ReCo)</span> codes (e.g., scores) your text response data; for example:")),
        p(HTML("&quot;The story is about a girl falling into and wandering through a fantasy world.&quot;"), class = "highlighted", style = "text-align:center;font-style:italic;font-size:1.2em;"),
        p(HTML("Is this a similar response like another one? Is it a correct response? ReCo can answer those questions by, first, <span class=\"highlighted\">categorizing all responses</span> in your data into response types by their <span class=\"highlighted\">semantics</span>, and then, second, by <span class=\"highlighted\">training the model</span> to assign codes to each response type. ReCo is specialized in short text responses from assessments such as PISA, but can be employed in other contexts as well.")),
        p("Fancy a Test Drive?", class = "reco_warn", style = "font-style:italic;font-size:2em;border-top:2px solid #e5eff4;margin-top:7px;margin-bottom:0px;padding-top:7px;"),
        p(HTML("If you first want to get an idea of what ReCo has in store for you, just press the <span class=\"highlighted\" style=\"font-style:italic;\">Import Demo Data Set</span> button on the first page and walk through the workflow. Otherwise, go ahead and <span class=\"highlighted\">import your data</span>."), style = "padding-bottom:10px;border-bottom:2px solid #e5eff4;"),
        div(HTML("The app's current version is <span class=\"highlighted\">ready for operational use</span>. It will, however, receive more features soon. Currently, you can only use prebuilt text corpora and corresponding semantic spaces that are delivered with the setup. <div class = \"reco_warn\", style = \"float:right;font-style:italic;font-size:2em;margin:5px 10px 5px 10px;\">More to Come!</div> In the next release, you will be able to craft tailored text corpora from Wikipedia and download additional prebuilt spaces from our website. Please also note that the provided demo data contains either English or German text data that has been constructed from real empirical data (but can only be considered simulated data itself), which was only possible for text responses with pretty low language diversity. ReCo's methodology, however, works for much more complex response semantics and a bunch of other languages as well.<br/>And, by the way, under certain conditions it does <span style=\"font-style:italic;\">not</span> work. <span class=\"highlighted\" style=\"font-weight:bold;font-style:italic;\">Let empirics be your guide! And use ReCo ethically.</span>")),
        h2(HTML("Stay tuned on<br/>"), a("www.reco.science", href="http://www.reco.science", target="_blank"), class = "highlighted", style = "text-align:center;font-style:italic;font-size:2em;margin-top:30px;"),
        size = "l", footer = div(a(img(src = "www/TBA-Logo-complete.png", style = "width:150px;float:right;margin-left:15px;"), href = "https://tba.dipf.de/en/", target = "_blank"), HTML("ReCo is currently being developed<br/>at the <a href=\"https://tba.dipf.de/en/\" target=\"_blank\">Centre for Technology-Based Assessment</a>, <a href=\"https://www.dipf.de/en/\" target=\"_blank\">DIPF | Leibniz Institute for Research and Information in Education</a> in Germany<br/><hr>"), modalButton("Get Rolling"), style = "font-size:0.7em;"), easyClose = T
      )
    )
  }
}

shinyApp(ui, server, onStart = function() { setwd(system.file(package = pkgname)) })

